Unit NDateCtrls;

{$I DELPHIAREA.INC}

Interface

Uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls, Forms, StdCtrls,
  ExtCtrls, NCalJalali, Dialogs,
  NCalendar, UxTheme, NCalendars, NCtrls;

Type
  TDateTimeList = Class(TPersistent)
  Private
    FList: Array Of TDateTime;
    FCount: Integer;
    FUpdateCount: Integer;
    FOnChanging: TNotifyEvent;
    FOnChange: TNotifyEvent;
    Function GetCapacity: Integer; Inline;
    Procedure SetCapacity(Value: Integer);
    Function GetItems(Index: Integer): TDate;
  Protected
    Procedure Changing; Virtual;
    Procedure Change; Virtual;
  Public
    Destructor Destroy; Override;
    Procedure Assign(Source: TPersistent); Override;
    Procedure BeginUpdate;
    Procedure EndUpdate;
    Procedure Clear; Inline;
    Function Add(Const ADateTime: TDateTime): Integer;
    Function Remove(Const ADateTime: TDateTime): Integer;
    Procedure Delete(Index: Integer);
    Function Exists(Const ADateTime: TDateTime): Boolean; Inline;
    Function IndexOf(Const ADateTime: TDateTime): Integer; Inline;
    Function Find(Const ADateTime: TDateTime; Out Index: Integer): Boolean;
    Function First: TDateTime;
    Function Last: TDateTime;
    Property Count: Integer Read FCount;
    Property Capacity: Integer Read GetCapacity Write SetCapacity;
    Property Items[Index: Integer]: TDate Read GetItems; Default;
    Property OnChange: TNotifyEvent Read FOnChange Write FOnChange;
    Property OnChanging: TNotifyEvent Read FOnChanging Write FOnChanging;
  End;

{$IFDEF COMPILER_XE2_UP}
  [ComponentPlatformsAttribute(PidWin32 Or PidWin64)]
{$ENDIF}
  TIntlCalSelectionStyle = (SsSingle, SsRange, SsFree);
  TIntlCalViewMode = (VmDay, VmMonth, VmYear);

  TIntlCalHitTest = (ChtNowhere, ChtArrowLeft, ChtArrowRight, ChtHeaderText,
    ChtFooter, ChtWeekNumber, ChtDayOfWeek, ChtDay, ChtMonth, ChtYear);
  TIntlCalDrawState = (CdsHeader, CdsFooter, CdsFixedRow, CdsFixedCol, CdsToday,
    CdsHotLight, CdsSelected, CdsFocused, CdsGrayed, CdsBlank);
  TIntlCalDrawStates = Set Of TIntlCalDrawState;
  TDateHintEvent = Procedure(Sender: TObject; Const ADate: TDate;
    Var HintStr: String) Of Object;
  TDateSelectableEvent = Procedure(Sender: TObject; Const ADate: TDate;
    Var CanSelect: Boolean) Of Object;
  TCustomDrawTextEvent = Procedure(Sender: TObject; Canvas: TCanvas;
    Const ARect: TRect; State: TIntlCalDrawStates; Const AText: String;
    Var DefaultDraw: Boolean) Of Object;
  TCustomDrawDayEvent = Procedure(Sender: TObject; Canvas: TCanvas;
    Const ARect: TRect; State: TIntlCalDrawStates; Const ADate: TDate;
    ADay: Integer; Var DefaultDraw: Boolean) Of Object;
  TCustomDrawDayOfWeekEvent = Procedure(Sender: TObject; Canvas: TCanvas;
    Const ARect: TRect; State: TIntlCalDrawStates; ADayOfWeek: TDayOfWeek;
    Var DefaultDraw: Boolean) Of Object;
  TCustomDrawWeekNumberEvent = Procedure(Sender: TObject; Canvas: TCanvas;
    Const ARect: TRect; State: TIntlCalDrawStates; AWeek: Integer;
    Var DefaultDraw: Boolean) Of Object;
  TCustomDrawMonthEvent = Procedure(Sender: TObject; Canvas: TCanvas;
    Const ARect: TRect; State: TIntlCalDrawStates; AMonth: Integer;
    Var DefaultDraw: Boolean) Of Object;
  TCustomDrawYearEvent = Procedure(Sender: TObject; Canvas: TCanvas;
    Const ARect: TRect; State: TIntlCalDrawStates; AYear: Integer;
    Var DefaultDraw: Boolean) Of Object;

  TEventItem = Class(TCollectionItem)
  Private
    FDate: TDate;
    FDateStr: String;
    FComment: String;
    Procedure SetDate(Const Value: TDate);
    Procedure SetDateStr(Const Value: String);
    Procedure SetComment(Const Value: String);
  Protected
    Procedure Changed(AllItems: Boolean);
    Procedure SetIndex(Value: Integer); Override;
    Function GetDisplayName: String; Override;
  Public
    Constructor Create(Collection: TCollection); Override;
    Destructor Destroy; Override;
    Procedure Assign(Source: TPersistent); Override;
  Published
    Property Date: TDate Read FDate Write SetDate;
    Property DateStr: String Read FDateStr Write SetDateStr;
    Property Comment: String Read FComment Write SetComment;
  End;

  TEventItems = Class(TOwnedCollection)
  Private
  Protected
    Function GetItem(Index: Integer): TEventItem;
    Procedure SetItem(Index: Integer; Value: TEventItem);
    Procedure Update(Item: TCollectionItem); Override;
  Public
    Constructor Create(AOwner: TPersistent);
    Procedure AddDate(Date: TDate); Overload;
    Procedure AddDate(Date: String); Overload;
    Function FindItem(Date: TDate): Integer;
    Function GetItemComment(Date: TDate): String;
    Property Items[Index: Integer]: TEventItem Read GetItem Write SetItem;
  End;

  TCustomIntlMonthCalendar = Class(TCustomControl)
  Protected Type
    TIntlCalSection = (CsHeader, CsHeaderText, CsArrowLeft, CsArrowRight,
      CsFooter, CsWeekSpace, CsDaysOfWeek, CsWeekNumbers, CsYMD);
    TIntlCalSections = Set Of TIntlCalSection;
    TIntlCalNameKind = (NkLong, NkShort, NkShortest);

    TIntlCalMetrics = Record
      NameKind: TIntlCalNameKind;
      LongFormat: Boolean;
      Bounds: TRect;
      HeaderHeight: Integer;
      FooterHeight: Integer;
      FixedRowHeight: Integer;
      FixedColWidth: Integer;
      RowHeight: Integer;
      ColWidth: Integer;
      RowCount: Integer;
      ColCount: Integer;
      StretchedRows: Integer;
      StretchedCols: Integer;
      PrefferedWidth: Integer;
      PrefferedHeight: Integer;
      Sections: Array [TIntlCalSection] Of TRect;
    End;
  Private
    FDate: TDate;
    FDateMin: TDate;
    FDateMax: TDate;
    FCalendarType: TCalendarClass;
    FCalendar: TCalendar;
    FWeekNumbers: Boolean;
    FCurrentView: TIntlCalViewMode;
    FSelectionStyle: TIntlCalSelectionStyle;
    FSelectionAnchor: TDate;
    FSelection: TDateTimeList;
    FShowHeader: Boolean;
    FShowFooter: Boolean;
    FFontHeader: TFont;
    FFontFooter: TFont;
    FFontDayOfWeek: TFont;
    FFontWeekNumbers: TFont;
    FEra, FYear, FMonth, FDay: Integer;
    FMonthsInYear, FDaysInYear, FDaysInMonth: Integer;
    FMonthFirstDate: TDate;
    FMonthFirstDayOfWeek: TDayOfWeek;
    FMonthStartCell: Integer;
    FFirstCellDayNumber: Integer;
    FFirstCellYear: Integer;
    FActualDateMin: TDate;
    FActualDateMax: TDate;
    FHotCell: Integer;
    FHotInfo: TIntlCalHitTest;
    FDraggingCell: Integer;
    FDraggingShiftState: TShiftState;
    FDraggingOffPage: Boolean;
    FNavigationTimer: Integer;
    FHintDate: TDate;
    FMetrics: TIntlCalMetrics;
    FPrefferedSize: TSize;
    FPrefferedSizeReady: Boolean;
    FPrefferedSizeSmallest: Boolean;
    FUpdateMetricsCount: Integer;
    FColorDivider: TColor;
    FColorToday: TColor;
    FTheme: HTHEME;
    FAnimate: Boolean;
    FSavedScreen: TBitmap;
    FSavedView: TIntlCalViewMode;
    FSavedDate: TDate;
    FSavedCellRect: TRect;
    FSaveFontHeader: Boolean;
    FSaveFontFooter: Boolean;
    FSaveFontDayOfWeek: Boolean;
    FSaveFontWeekNumbers: Boolean;
    FOnDateChange: TNotifyEvent;
    FOnPageChange: TNotifyEvent;
    FOnCalendarChange: TNotifyEvent;
    FOnSelectionChange: TNotifyEvent;
    FOnDateHint: TDateHintEvent;
    FOnDateSelectable: TDateSelectableEvent;
    FOnCustomDrawDayOfWeek: TCustomDrawDayOfWeekEvent;
    FOnCustomDrawWeekNumber: TCustomDrawWeekNumberEvent;
    FOnCustomDrawDay: TCustomDrawDayEvent;
    FOnCustomDrawMonth: TCustomDrawMonthEvent;
    FOnCustomDrawYear: TCustomDrawYearEvent;
    FOnCustomDrawHeaderText: TCustomDrawTextEvent;
    FOnCustomDrawFooterText: TCustomDrawTextEvent;
    FEventItems: TEventItems;
    FShowEvents: Boolean;
    FAutoToday: Boolean;
    Procedure SetDate(Const Value: TDate); Overload;
    Procedure SetDateMin(Const Value: TDate);
    Procedure SetDateMax(Const Value: TDate);
    Procedure SetCalendarType(Value: TCalendarClass);
    Procedure SetWeekNumbers(Value: Boolean);
    Procedure SetSelectionStyle(Value: TIntlCalSelectionStyle);
    Procedure SetShowHeader(Value: Boolean);
    Procedure SetShowFooter(Value: Boolean);
    Procedure SetFontHeader(Value: TFont);
    Procedure SetFontFooter(Value: TFont);
    Procedure SetFontDayOfWeek(Value: TFont);
    Procedure SetFontWeekNumbers(Value: TFont);
    Procedure SetCurrentView(Value: TIntlCalViewMode);
    Procedure SetColorDivider(Value: TColor);
    Procedure SetColorToday(Value: TColor);
    Function GetSelectionStart: TDate;
    Function GetSelectionEnd: TDate;
    Function GetSelectionCount: Integer;
    Function GetSelection(Index: Integer): TDate;
    Function GetMonthLastDate: TDate; Inline;
    Function GetFirstCellDate: TDate; Inline;
    Function GetLastCellDate: TDate; Inline;
    Function GetLastCellYear: Integer; Inline;
    Function GetFirstYearOnPage: Integer;
    Function GetLastYearOnPage: Integer;
    Function GetColCount: Integer; Inline;
    Function GetRowCount: Integer; Inline;
    Function GetCellCount: Integer; Inline;
    Function GetFocusedCell: Integer;
    Function GetTheme: HTHEME;
    Function IsDateStored: Boolean;
    Function IsDateMinStored: Boolean;
    Function IsDateMaxStored: Boolean;
    Procedure ReadCalendarType(Reader: TReader);
    Procedure WriteCalendarType(Writer: TWriter);
    Procedure FontHeaderChanged(Sender: TObject);
    Procedure FontFooterChanged(Sender: TObject);
    Procedure FontDayOfWeekChanged(Sender: TObject);
    Procedure FontWeekNumbersChanged(Sender: TObject);
    Procedure CalendarChanged(Sender: TObject);
    Procedure WMTimer(Var Message: TWMTimer); Message WM_TIMER;
    Procedure WMEraseBkgnd(Var Message: TWMEraseBkgnd); Message WM_ERASEBKGND;
    Procedure WMGetDlgCode(Var Message: TWMGetDlgCode); Message WM_GETDLGCODE;
    Procedure WMMouseActivate(Var Message: TWMMouseActivate);
      Message WM_MOUSEACTIVATE;
    Procedure CMMouseLeave(Var Message: TMessage); Message CM_MOUSELEAVE;
    Procedure CMHintShow(Var Message: TCMHintShow); Message CM_HINTSHOW;
    Procedure CMFontChanged(Var Message: TMessage); Message CM_FONTCHANGED;
    Procedure ExtractDateInfo(Const ADate: TDate);
    Procedure RenderEventMarker(Canvas: TCanvas; Var Rect: TRect);
    Procedure SetEventItems(Const Value: TEventItems);
    Procedure SetShowEvents(Const Value: Boolean);
    Function GetSettings: TCalendarSettings;
    Procedure SetSettings(Const Value: TCalendarSettings);
    Function GetDatesAsText: String;
    Procedure SetAutoToday(Const Value: Boolean);
  Protected Const
    AnimateDuration = 400; { ms }

  Const
    AutoNavInterval = AnimateDuration + 100; { ms }

  Const
    AutoNavTimerForward = 100;

  Const
    AutoNavTimerBackward = 200;
  Protected
    Procedure Notification(AComponent: TComponent;
      Operation: TOperation); Override;
    Procedure DefineProperties(Filer: TFiler); Override;
    Procedure Resize; Override;
    Procedure Paint; Override;
    Procedure Loaded; Override;
    Function CanAutoSize(Var NewWidth: Integer; Var NewHeight: Integer)
      : Boolean; Override;
    Procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); Override;
    Procedure MouseMove(Shift: TShiftState; X, Y: Integer); Override;
    Procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); Override;
    Function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint)
      : Boolean; Override;
    Function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint)
      : Boolean; Override;
    Procedure KeyDown(Var Key: Word; Shift: TShiftState); Override;
    Procedure DoEnter; Override;
    Procedure DoExit; Override;
    Procedure DoDateChange; Virtual;
    Procedure DoPageChange; Virtual;
    Procedure DoCalendarChange; Virtual;
    Procedure DoSelectionChange; Virtual;
    Procedure DoDateHint(Const ADate: TDate; Var HintStr: String); Virtual;
    Function DoDateSelectable(Const ADate: TDate): Boolean; Virtual;
    Function DoCustomDrawDayOfWeek(Canvas: TCanvas; Const ARect: TRect;
      State: TIntlCalDrawStates; ADayOfWeek: TDayOfWeek): Boolean; Virtual;
    Function DoCustomDrawWeekNumber(Canvas: TCanvas; Const ARect: TRect;
      State: TIntlCalDrawStates; AWeek: Integer): Boolean; Virtual;
    Function DoCustomDrawDay(Canvas: TCanvas; Const ARect: TRect;
      State: TIntlCalDrawStates; Const ADate: TDate; ADay: Integer)
      : Boolean; Virtual;
    Function DoCustomDrawMonth(Canvas: TCanvas; Const ARect: TRect;
      State: TIntlCalDrawStates; AMonth: Integer): Boolean; Virtual;
    Function DoCustomDrawYear(Canvas: TCanvas; Const ARect: TRect;
      State: TIntlCalDrawStates; AYear: Integer): Boolean; Virtual;
    Function DoCustomDrawHeaderText(Canvas: TCanvas; Const ARect: TRect;
      State: TIntlCalDrawStates; Const AText: String): Boolean; Virtual;
    Function DoCustomDrawFooterText(Canvas: TCanvas; Const ARect: TRect;
      State: TIntlCalDrawStates; Const AText: String): Boolean; Virtual;
    Procedure NormalizeDigits(Var Str: String); Inline;
    Function GetDayName(DayOfWeek: TDayOfWeek; NameKind: TIntlCalNameKind)
      : String; Virtual;
    Function GetMonthName(Month: Integer; NameKind: TIntlCalNameKind)
      : String; Virtual;
    Function GetYearText(AYear: Integer): String; Virtual;
    Function GetDayText(ADay: Integer): String; Virtual;
    Function GetWeekText(AWeek: Integer): String; Virtual;
    Function GetHeaderText(ViewMode: TIntlCalViewMode): String; Virtual;
    Function GetFooterText(LongFormat: Boolean): String; Virtual;
    Function GetHitTestInfoAt(X, Y: Integer; Out Cell: Integer)
      : TIntlCalHitTest; Virtual;
    Function GetDateAt(X, Y: Integer; Out ADate: TDate): Boolean; Virtual;
    Function GetDateAtCell(Index: Integer; Out ADate: TDate): Boolean; Virtual;
    Function GetMonthAt(X, Y: Integer; Out AMonth: Integer): Boolean; Virtual;
    Function GetMonthAtCell(Index: Integer; Out AMonth: Integer)
      : Boolean; Virtual;
    Function GetYearAt(X, Y: Integer; Out AYear: Integer): Boolean; Virtual;
    Function GetYearAtCell(Index: Integer; Out AYear: Integer)
      : Boolean; Virtual;
    Function GetCalendarClientRect: TRect; Virtual;
    Function GetSectionRect(Section: TIntlCalSection): TRect; Inline;
    Function GetSectionWidth(Section: TIntlCalSection): Integer; Inline;
    Function GetSectionHeight(Section: TIntlCalSection): Integer; Inline;
    Function GetCellRect(Row, Col: Integer): TRect; Overload;
    Function GetCellRect(Index: Integer): TRect; Overload; Inline;
    Function GetRowAt(Y: Integer): Integer;
    Function GetColAt(X: Integer): Integer;
    Function GetAdjacentPageDate(Backward: Boolean): TDate;
    Function GetPreferredClientSize(Out AWidth, AHeight: Integer)
      : Boolean; Virtual;
    Function GetPreferredControlSize(Out AWidth, AHeight: Integer)
      : Boolean; Virtual;
    Procedure CalcMetrics(ViewMode: TIntlCalViewMode;
      Out Metrics: TIntlCalMetrics); Virtual;
    Procedure CalcMetricsSections(Var Metrics: TIntlCalMetrics); Virtual;
    Procedure UpdateMetrics; Virtual;
    Procedure UpdateMinMaxRanges; Virtual;
    Procedure ValidateCalendar; Virtual;
    Function IsThemed: Boolean;
    Function IsYearOnPage(AYear: Integer): Boolean; Inline;
    Procedure AnimateBegin; Virtual;
    Procedure AnimatePlay; Virtual;
    Procedure AnimateEnd; Virtual;
    Function TakeSnapshot: TBitmap;
    Procedure RenderBackground(Canvas: TCanvas; Var Rect: TRect;
      State: TIntlCalDrawStates); Virtual;
    Procedure RenderText(Canvas: TCanvas; Var Rect: TRect;
      State: TIntlCalDrawStates; Const Text: String;
      Alignment: TAlignment); Virtual;
    Procedure RenderArrow(Canvas: TCanvas; Var Rect: TRect;
      State: TIntlCalDrawStates; Backward: Boolean); Virtual;
    Procedure RenderTodayMarker(Canvas: TCanvas; Var Rect: TRect); Virtual;
    Procedure DrawCalendar(Canvas: TCanvas); Virtual;
    Procedure DrawHeader(Canvas: TCanvas); Virtual;
    Procedure DrawFooter(Canvas: TCanvas); Virtual;
    Procedure DrawWeekSpace(Canvas: TCanvas); Virtual;
    Procedure DrawDaysOfWeek(Canvas: TCanvas); Virtual;
    Procedure DrawWeekNumbers(Canvas: TCanvas); Virtual;
    Procedure DrawDays(Canvas: TCanvas); Virtual;
    Procedure DrawMonths(Canvas: TCanvas); Virtual;
    Procedure DrawYears(Canvas: TCanvas); Virtual;
    Procedure InvalidateSections(Sections: TIntlCalSections); Virtual;
    Procedure InvalidateDateDependentSections; Inline;
    Procedure InvalidateCell(Index: Integer); Virtual;
    Procedure InvalidateDate(Const ADate: TDate); Virtual;
    Procedure InvalidateDates(ADates: TDateTimeList); Virtual;
    Procedure InvalidateDateRange(Const ADate1, ADate2: TDate); Virtual;
    Procedure InvalidateHotLight; Virtual;
    Procedure InvalidateSelection; Virtual;
    Procedure SelectDate(Const ADate: TDate; ShiftState: TShiftState = []);
    Procedure FocusDate(Const ADate: TDate);
    Property Theme: HTHEME Read GetTheme;
    Property ColCount: Integer Read GetColCount;
    Property RowCount: Integer Read GetRowCount;
    Property CellCount: Integer Read GetCellCount;
    Property FirstCellDayNumber: Integer Read FFirstCellDayNumber;
    Property FirstCellDate: TDate Read GetFirstCellDate;
    Property LastCellDate: TDate Read GetLastCellDate;
    Property FirstCellYear: Integer Read FFirstCellYear;
    Property LastCellYear: Integer Read GetLastCellYear;
    Property FirstYearOnPage: Integer Read GetFirstYearOnPage;
    Property LastYearOnPage: Integer Read GetLastYearOnPage;
    Property MonthStartCell: Integer Read FMonthStartCell;
    Property FocusedCell: Integer Read GetFocusedCell;
    Property HotCell: Integer Read FHotCell;
    Property HotInfo: TIntlCalHitTest Read FHotInfo;
    Property DraggingCell: Integer Read FDraggingCell;
    Property DraggingOffPage: Boolean Read FDraggingOffPage;
    Property NavigationTimer: Integer Read FNavigationTimer;
  Protected
    Property Date: TDate Read FDate Write SetDate Stored IsDateStored;
    Property DateMin: TDate Read FDateMin Write SetDateMin
      Stored IsDateMinStored;
    Property DateMax: TDate Read FDateMax Write SetDateMax
      Stored IsDateMaxStored;
    Property CalendarType: TCalendarClass Read FCalendarType
      Write SetCalendarType;
    Property SelectionStyle: TIntlCalSelectionStyle Read FSelectionStyle
      Write SetSelectionStyle Default SsSingle;
    Property WeekNumbers: Boolean Read FWeekNumbers Write SetWeekNumbers
      Default False;
    Property ShowHeader: Boolean Read FShowHeader Write SetShowHeader
      Default True;
    Property ShowFooter: Boolean Read FShowFooter Write SetShowFooter
      Default True;
    Property FontHeader: TFont Read FFontHeader Write SetFontHeader
      Stored FSaveFontHeader;
    Property FontFooter: TFont Read FFontFooter Write SetFontFooter
      Stored FSaveFontFooter;
    Property FontDayOfWeek: TFont Read FFontDayOfWeek Write SetFontDayOfWeek
      Stored FSaveFontDayOfWeek;
    Property FontWeekNumbers: TFont Read FFontWeekNumbers
      Write SetFontWeekNumbers Stored FSaveFontWeekNumbers;
    Property CurrentView: TIntlCalViewMode Read FCurrentView
      Write SetCurrentView;
    Property ColorDivider: TColor Read FColorDivider Write SetColorDivider
      Default ClBtnFace;
    Property ColorToday: TColor Read FColorToday Write SetColorToday
      Default ClRed;
    Property Animate: Boolean Read FAnimate Write FAnimate Default True;
    Property OnDateChange: TNotifyEvent Read FOnDateChange Write FOnDateChange;
    Property OnPageChange: TNotifyEvent Read FOnPageChange Write FOnPageChange;
    Property OnCalendarChange: TNotifyEvent Read FOnCalendarChange
      Write FOnCalendarChange;
    Property OnSelectionChange: TNotifyEvent Read FOnSelectionChange
      Write FOnSelectionChange;
    Property OnDateHint: TDateHintEvent Read FOnDateHint Write FOnDateHint;
    Property OnDateSelectable: TDateSelectableEvent Read FOnDateSelectable
      Write FOnDateSelectable;
    Property OnCustomDrawDayOfWeek: TCustomDrawDayOfWeekEvent
      Read FOnCustomDrawDayOfWeek Write FOnCustomDrawDayOfWeek;
    Property OnCustomDrawWeekNumber: TCustomDrawWeekNumberEvent
      Read FOnCustomDrawWeekNumber Write FOnCustomDrawWeekNumber;
    Property OnCustomDrawDay: TCustomDrawDayEvent Read FOnCustomDrawDay
      Write FOnCustomDrawDay;
    Property OnCustomDrawMonth: TCustomDrawMonthEvent Read FOnCustomDrawMonth
      Write FOnCustomDrawMonth;
    Property OnCustomDrawYear: TCustomDrawYearEvent Read FOnCustomDrawYear
      Write FOnCustomDrawYear;
    Property OnCustomDrawHeaderText: TCustomDrawTextEvent
      Read FOnCustomDrawHeaderText Write FOnCustomDrawHeaderText;
    Property OnCustomDrawFooterText: TCustomDrawTextEvent
      Read FOnCustomDrawFooterText Write FOnCustomDrawFooterText;
  Public
    Constructor Create(AOwner: TComponent); Override;
    Destructor Destroy; Override;
    Function GetMonthEvents(Month: Integer): String;
    Function GetYearEvents(Year: Integer): String;
    Procedure SetToday;
    Procedure SetDate(DateStr: String; FmtStr: String = 'yyyy/mm/dd'); Overload;
    Procedure UpdateCalendar; Virtual;
    Procedure ClearSelection; Virtual;
    Procedure Select(Const ADate: TDate; ShiftState: TShiftState = []); Virtual;
    Function IsSelected(Const ADate: TDate): Boolean; Virtual;
    Function IsYearInRange(AEra, AYear: Integer): Boolean; Overload;
    Function IsYearInRange(AYear: Integer): Boolean; Overload; Inline;
    Function IsMonthInRange(AEra, AYear, AMonth: Integer): Boolean; Overload;
    Function IsMonthInRange(AYear, AMonth: Integer): Boolean; Overload; Inline;
    Function IsDateInRange(Const ADate: TDate): Boolean; Virtual;
    Function EnsureDateInRange(Const ADate: TDate): TDate; Virtual;
    Function Format(Const FmtStr: String; Const ADate: TDateTime): String;
    Function Parse(Const FmtStr: String; Const DateTimeStr: String): TDateTime;
    Property Calendar: TCalendar Read FCalendar;
    Property Era: Integer Read FEra;
    Property Year: Integer Read FYear;
    Property Month: Integer Read FMonth;
    Property Day: Integer Read FDay;
    Property DaysInMonth: Integer Read FDaysInMonth;
    Property MonthFirstDate: TDate Read FMonthFirstDate;
    Property MonthLastDate: TDate Read GetMonthLastDate;
    Property MonthFirstDayOfWeek: TDayOfWeek Read FMonthFirstDayOfWeek;
    Property MonthsInYear: Integer Read FMonthsInYear;
    Property DaysInYear: Integer Read FDaysInYear;
    Property ActualDateMin: TDate Read FActualDateMin;
    Property ActualDateMax: TDate Read FActualDateMax;
    Property SelectionStart: TDate Read GetSelectionStart;
    Property SelectionEnd: TDate Read GetSelectionEnd;
    Property SelectionCount: Integer Read GetSelectionCount;
    Property Selection[Index: Integer]: TDate Read GetSelection;
    Property EventItems: TEventItems Read FEventItems Write SetEventItems;
    Property ShowEvents: Boolean Read FShowEvents Write SetShowEvents;
  Published
    Property Settings: TCalendarSettings Read GetSettings Write SetSettings;
    Property DatesAsText: String Read GetDatesAsText;
    Property AutoToday: Boolean Read FAutoToday Write SetAutoToday;
  End;

{$IFDEF COMPILER_XE2_UP}
  [ComponentPlatformsAttribute(PidWin32 Or PidWin64)]
{$ENDIF}

  TIntlMonthCalendar = Class(TCustomIntlMonthCalendar)
  Public
  Published
    Property CurrentView;
    Property Align;
    Property AlignWithMargins;
    Property Anchors;
    Property Animate;
    Property AutoSize;
    Property BevelEdges;
    Property BevelInner;
    Property BevelKind Default BkTile;
    Property BevelOuter;
    Property BorderWidth Default 2;
    Property CalendarType;
    Property Color Default ClWindow;
    Property ColorDivider;
    Property ColorToday;
    Property Constraints;
    Property Ctl3D;
    Property Date;
    Property DateMin;
    Property DateMax;
    Property DoubleBuffered;
    Property DragCursor;
    Property DragKind;
    Property DragMode;
    Property Enabled;
    Property Font;
    Property FontHeader;
    Property FontFooter;
    Property FontDayOfWeek;
    Property FontWeekNumbers;
    Property Hint;
    Property ImeMode;
    Property ImeName;
    Property Margins;
    Property ParentColor Default False;
    Property ParentCtl3D;
    Property ParentDoubleBuffered;
    Property ParentFont;
    Property ParentShowHint;
    Property PopupMenu;
    Property SelectionStyle;
    Property ShowHeader;
    Property ShowFooter;
    Property ShowHint;
    Property TabOrder;
    Property TabStop Default True;
{$IFDEF COMPILER2010_UP}
    Property Touch;
{$ENDIF}
    Property Visible;
    Property WeekNumbers;
    Property OnCalendarChange;
    Property OnClick;
    Property OnContextPopup;
    Property OnCustomDrawHeaderText;
    Property OnCustomDrawDay;
    Property OnCustomDrawMonth;
    Property OnCustomDrawYear;
    Property OnCustomDrawDayOfWeek;
    Property OnCustomDrawWeekNumber;
    Property OnCustomDrawFooterText;
    Property OnDateChange;
    Property OnDateSelectable;
    Property OnDateHint;
    Property OnDblClick;
    Property OnDragDrop;
    Property OnDragOver;
    Property OnEndDock;
    Property OnEndDrag;
    Property OnEnter;
    Property OnExit;
{$IFDEF COMPILER2010_UP}
    Property OnGesture;
{$ENDIF}
    Property OnKeyDown;
    Property OnKeyPress;
    Property OnKeyUp;
    Property OnMouseActivate;
    Property OnMouseDown;
    Property OnMouseMove;
    Property OnMouseUp;
{$IFDEF COMPILER2010_UP}
    Property OnMouseEnter;
{$ENDIF}
{$IFDEF COMPILER2010_UP}
    Property OnMouseLeave;
{$ENDIF}
    Property OnMouseWheel;
    Property OnMouseWheelDown;
    Property OnMouseWheelUp;
    Property OnPageChange;
    Property OnResize;
    Property OnSelectionChange;
    Property OnStartDock;
    Property OnStartDrag;
    Property EventItems;
    Property ShowEvents;
    Property BiDiMode;
    Property ParentBiDiMode;
  End;

  TNasimMonthCalendar = Class(TIntlMonthCalendar);
  TPopupCloseUpEvent = Procedure(Sender: TObject; Accept: Boolean) Of Object;

  TIntlMonthCalendarPopup = Class(TCustomIntlMonthCalendar)
  Private
    FOnCloseUp: TPopupCloseUpEvent;
    FEventItems: TEventItems;
    Procedure WMMouseActivate(Var Message: TMessage); Message WM_MOUSEACTIVATE;
    Procedure CMVisibleChanged(Var Message: TMessage);
      Message CM_VISIBLECHANGED;
  Protected
    Procedure DblClick; Override;
    Procedure CreateParams(Var Params: TCreateParams); Override;
    Procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); Override;
    Procedure KeyDown(Var Key: Word; Shift: TShiftState); Override;
    Procedure DoCloseUp(Accept: Boolean); Virtual;
    Property OnCloseUp: TPopupCloseUpEvent Read FOnCloseUp Write FOnCloseUp;
  Public
    Constructor Create(AOwner: TWinControl); Reintroduce;
    Function Focused: Boolean; Override;
    Procedure Show(X, Y: Integer);
  Published
    Property Left Stored False;
    Property Top Stored False;
    Property Width; // Stored False;
    Property Height; // Stored False;
    Property Animate;
    Property Color Default ClWindow;
    Property ColorDivider;
    Property ColorToday;
    Property Font;
    Property FontHeader;
    Property FontFooter;
    Property FontDayOfWeek;
    Property FontWeekNumbers;
    Property ParentFont;
    Property ParentShowHint;
    Property ShowHeader;
    Property ShowFooter Default True;
    Property ShowHint;
    Property WeekNumbers;
    Property OnCustomDrawHeaderText;
    Property OnCustomDrawDay;
    Property OnCustomDrawMonth;
    Property OnCustomDrawYear;
    Property OnCustomDrawDayOfWeek;
    Property OnCustomDrawWeekNumber;
    Property OnCustomDrawFooterText;
    Property OnDateHint;
  End;

  TIntlPickerHitTest = (PhtNowhere, PhtDropDown, PhtCheckBox, PhtDate,
    PhtControl);

  TCustomIntlDatePicker = Class(TCustomControl)
  Private
    FDate: TDate;
    FDateFormat: String;
    FChecked: Boolean;
    FShowCheckBox: Boolean;
    FBorderStyle: TBorderStyle;
    FCalendarPopup: TIntlMonthCalendarPopup;
    FCheckBoxRect: TRect;
    FDropDownRect: TRect;
    FDateRect: TRect;
    FHotInfo: TIntlPickerHitTest;
    FTheme: HTHEME;
    FCheckPressed: Boolean;
    FOnChange: TNotifyEvent;
    FAutoToday: Boolean;
    Procedure SetDate(Const Value: TDate); Overload;
    Procedure SetDateFormat(Const Value: String);
    Procedure SetChecked(Value: Boolean);
    Procedure SetShowCheckBox(Value: Boolean);
    Procedure SetBorderStyle(Value: TBorderStyle);
    Function GetDateMin: TDate; Inline;
    Procedure SetDateMin(Const Value: TDate); Inline;
    Function GetDateMax: TDate; Inline;
    Procedure SetDateMax(Const Value: TDate); Inline;
    Function GetCalendarType: TCalendarClass; Inline;
    Procedure SetCalendarType(Value: TCalendarClass); Inline;
    Function GetCalendar: TCalendar; Inline;
    Function GetActualDateMin: TDate; Inline;
    Function GetActualDateMax: TDate; Inline;
    Function GetTheme: HTHEME;
    Function IsDateStored: Boolean;
    Function IsDateMinStored: Boolean;
    Function IsDateMaxStored: Boolean;
    Procedure ReadCalendarType(Reader: TReader);
    Procedure WriteCalendarType(Writer: TWriter);
    Procedure CalendarChanged(Sender: TObject);
    Procedure PopupCloseUp(Sender: TObject; Accept: Boolean);
    Procedure WMNCPaint(Var Message: TWMNCPaint); Message WM_NCPAINT;
    Procedure WMMouseActivate(Var Message: TWMMouseActivate);
      Message WM_MOUSEACTIVATE;
    Procedure WMSetFocus(Var Message: TWMSetFocus); Message WM_SETFOCUS;
    Procedure WMKillFocus(Var Message: TWMKillFocus); Message WM_KILLFOCUS;
    Procedure WMGetDlgCode(Var Message: TWMGetDlgCode); Message WM_GETDLGCODE;
    Procedure CMEnableChanged(Var Message: TMessage); Message CM_ENABLEDCHANGED;
    Procedure CMBiDiModeChanged(Var Message: TMessage);
      Message CM_BIDIMODECHANGED;
    Procedure CMBorderChanged(Var Message: TMessage); Message CM_BORDERCHANGED;
    Procedure CMCtl3DChanged(Var Message: TMessage); Message CM_CTL3DCHANGED;
    Procedure CMCancelMode(Var Message: TCMCancelMode); Message CM_CANCELMODE;
    Procedure CMMouseEnter(Var Message: TMessage); Message CM_MOUSEENTER;
    Procedure CMMouseLeave(Var Message: TMessage); Message CM_MOUSELEAVE;
    Function GetDateAsText: String;
    Function GetText: String;
    Procedure SetAutoToday(Const Value: Boolean);
  Protected
    Procedure DefineProperties(Filer: TFiler); Override;
    Procedure CreateParams(Var Params: TCreateParams); Override;
    Procedure Resize; Override;
    Procedure Paint; Override;
    Function CanAutoSize(Var NewWidth: Integer; Var NewHeight: Integer)
      : Boolean; Override;
    Procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); Override;
    Procedure MouseMove(Shift: TShiftState; X, Y: Integer); Override;
    Procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); Override;
    Function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint)
      : Boolean; Override;
    Function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint)
      : Boolean; Override;
    Procedure KeyDown(Var Key: Word; Shift: TShiftState); Override;
    Function CanModify: Boolean; Virtual;
    Procedure UpdateText; Virtual;
    Procedure UpdateRects; Virtual;
    Function GetMinHeight: Integer; Virtual;
    Function GetHitTestInfoAt(X, Y: Integer): TIntlPickerHitTest; Virtual;
    Procedure InvalidateCheckBox; Virtual;
    Procedure InvalidateDropDown; Virtual;
    Procedure InvalidateDate; Virtual;
    Procedure InvalidateFrame; Virtual;
    Procedure PopupDropDown; Virtual;
    Procedure Change; Virtual;
    Property Theme: HTHEME Read GetTheme;
    Property HotInfo: TIntlPickerHitTest Read FHotInfo;
    Property CheckPressed: Boolean Read FCheckPressed;
    Property CalendarPopup: TIntlMonthCalendarPopup Read FCalendarPopup;
    // write FCalendarPopup;
    Property Date: TDate Read FDate Write SetDate Stored IsDateStored;
    Property DateFormat: String Read FDateFormat Write SetDateFormat;
    Property DateMin: TDate Read GetDateMin Write SetDateMin
      Stored IsDateMinStored;
    Property DateMax: TDate Read GetDateMax Write SetDateMax
      Stored IsDateMaxStored;
    Property CalendarType: TCalendarClass Read GetCalendarType
      Write SetCalendarType;
    Property Checked: Boolean Read FChecked Write SetChecked Default True;
    Property ShowCheckBox: Boolean Read FShowCheckBox Write SetShowCheckBox
      Default False;
    Property BorderStyle: TBorderStyle Read FBorderStyle Write SetBorderStyle
      Default BsSingle;
    Property OnChange: TNotifyEvent Read FOnChange Write FOnChange;
  Public
    Constructor Create(AOwner: TComponent); Override;
    Destructor Destroy; Override;
    Procedure SetToday;
    Procedure SetDate(DateStr: String = '';
      FmtStr: String = 'yyyy/mm/dd'); Overload;
    Function IsYearInRange(AEra, AYear: Integer): Boolean; Overload; Inline;
    Function IsYearInRange(AYear: Integer): Boolean; Overload; Inline;
    Function IsMonthInRange(AEra, AYear, AMonth: Integer): Boolean;
      Overload; Inline;
    Function IsMonthInRange(AYear, AMonth: Integer): Boolean; Overload; Inline;
    Function IsDateInRange(Const ADate: TDate): Boolean; Inline;
    Function EnsureDateInRange(Const ADate: TDate): TDate; Inline;
    Property Calendar: TCalendar Read GetCalendar;
    Property ActualDateMin: TDate Read GetActualDateMin;
    Property ActualDateMax: TDate Read GetActualDateMax;
  Published
    Property DateAsText: String Read GetDateAsText;
    Property Text: String Read GetText;
    Property AutoToday: Boolean Read FAutoToday Write SetAutoToday;
  End;

{$IFDEF COMPILER_XE2_UP}
  [ComponentPlatformsAttribute(PidWin32 Or PidWin64)]
{$ENDIF}

  TIntlDatePicker = Class(TCustomIntlDatePicker)
  Published
    Property Align;
    Property AlignWithMargins;
    Property Anchors;
    Property AutoSize Default True;
    Property BevelEdges;
    Property BevelInner;
    Property BevelKind Default BkNone;
    Property BevelOuter;
    Property BiDiMode;
    Property BorderStyle;
    Property CalendarPopup;
    Property CalendarType;
    Property Checked;
    Property Color Default ClWindow;
    Property Constraints;
    Property Ctl3D;
    Property Date;
    Property DateFormat;
    Property DateMin;
    Property DateMax;
    Property DoubleBuffered;
    Property DragCursor;
    Property DragKind;
    Property DragMode;
    Property Enabled;
    Property Font;
    Property Hint;
    Property ImeMode;
    Property ImeName;
    Property Margins;
    Property ParentBiDiMode;
    Property ParentColor Default False;
    Property ParentCtl3D;
    Property ParentDoubleBuffered;
    Property ParentFont;
    Property ParentShowHint;
    Property PopupMenu;
    Property ShowCheckBox;
    Property ShowHint;
    Property TabOrder;
    Property TabStop Default True;
{$IFDEF COMPILER2010_UP}
    Property Touch;
{$ENDIF}
    Property Visible;
    Property OnChange;
    Property OnClick;
    Property OnContextPopup;
    Property OnDblClick;
    Property OnDragDrop;
    Property OnDragOver;
    Property OnEndDock;
    Property OnEndDrag;
    Property OnEnter;
    Property OnExit;
{$IFDEF COMPILER2010_UP}
    Property OnGesture;
{$ENDIF}
    Property OnKeyDown;
    Property OnKeyPress;
    Property OnKeyUp;
    Property OnMouseActivate;
    Property OnMouseDown;
    Property OnMouseMove;
    Property OnMouseUp;
{$IFDEF COMPILER2010_UP}
    Property OnMouseEnter;
{$ENDIF}
{$IFDEF COMPILER2010_UP}
    Property OnMouseLeave;
{$ENDIF}
    Property OnMouseWheel;
    Property OnMouseWheelDown;
    Property OnMouseWheelUp;
    Property OnResize;
    Property OnStartDock;
    Property OnStartDrag;
  End;

  TNasimDatePicker = Class(TIntlDatePicker);

Implementation

Uses
  RTLConsts, Math, Types, Variants, Themes, ComCtrls, NUnicode;

Resourcestring
  SDateOutOfRangeError = 'The date must be between %s and %s';

Type
  TControlHack = Class(TControl);

  { Helper Functions }

Procedure SwapScreenByScroll(DC: HDC; Const Rect: TRect; OfsX, OfsY: Integer;
  OldBmp, NewBmp: TBitmap; RightToLeft: Boolean; Duration: Cardinal);
Var
  Step, LastStep, Distance, X: Integer;
  StartTime: Cardinal;
Begin
  StartTime := GetTickCount;
  Distance := Rect.Right - Rect.Left;
  Step := 0;
  Repeat
    LastStep := Step;
    Repeat
      Sleep(1);
      Step := MulDiv(Distance, GetTickCount - StartTime, Duration);
    Until Step <> LastStep;
    X := Min(Rect.Left + Step, Rect.Right);
    If RightToLeft Then
    Begin
      BitBlt(DC, Rect.Left, Rect.Top, X - Rect.Left, Rect.Bottom - Rect.Top,
        NewBmp.Canvas.Handle, Rect.Right - X - OfsX, Rect.Top - OfsY, SRCCOPY);
      BitBlt(DC, X, Rect.Top, Rect.Right - X, Rect.Bottom - Rect.Top,
        OldBmp.Canvas.Handle, Rect.Left - OfsX, Rect.Top - OfsY, SRCCOPY);
    End
    Else
    Begin
      BitBlt(DC, Rect.Left, Rect.Top, Rect.Right - X, Rect.Bottom - Rect.Top,
        OldBmp.Canvas.Handle, X - OfsX, Rect.Top - OfsY, SRCCOPY);
      BitBlt(DC, Rect.Right - X, Rect.Top, X - Rect.Left,
        Rect.Bottom - Rect.Top, NewBmp.Canvas.Handle, Rect.Left - OfsX,
        Rect.Top - OfsY, SRCCOPY);
    End;
  Until X = Rect.Right;
End;

Procedure SwapScreenByZoom(DC: HDC; Const Rect: TRect; OfsX, OfsY: Integer;
  OldBmp, NewBmp: TBitmap; Const OldRect, NewRect: TRect; Duration: Cardinal);
Var
  Progress, LastProgress: Integer;
  StartTime: Cardinal;
  R: TRect;
Begin
  // StartTime := GetTickCount;
  // Progress := 0;
  // repeat
  // LastProgress := Progress;
  // repeat
  // Sleep(1);
  // Progress := MulDiv(200, GetTickCount - StartTime, Duration);
  // until Progress <> LastProgress;
  // if Progress > 100 then
  // Progress := 100;
  // R.Left := OldRect.Left + MulDiv(Rect.Left - OldRect.Left, 100 - Progress, 100);
  // R.Top := OldRect.Top + MulDiv(Rect.Top - OldRect.Top, 100 - Progress, 100);
  // R.Right := OldRect.Right + MulDiv(Rect.Right - OldRect.Right, 100 - Progress, 100);
  // R.Bottom := OldRect.Bottom + MulDiv(Rect.Bottom - OldRect.Bottom, 100 - Progress, 100);
  // StretchBlt(DC, Rect.Left, Rect.Top, Rect.Right - Rect.Left, Rect.Bottom - Rect.Top,
  // OldBmp.Canvas.Handle, R.Left - OfsX, R.Top - OfsY, R.Right - R.Left, R.Bottom - R.Top, SRCCOPY);
  // until Progress = 100;

  StartTime := GetTickCount;
  Progress := 0;
  Repeat
    LastProgress := Progress;
    Repeat
      Sleep(1);
      Progress := MulDiv(200, GetTickCount - StartTime, Duration);
    Until Progress <> LastProgress;
    If Progress > 100 Then
      Progress := 100;
    R.Left := NewRect.Left + MulDiv(Rect.Left - NewRect.Left, Progress, 100);
    R.Top := NewRect.Top + MulDiv(Rect.Top - NewRect.Top, Progress, 100);
    R.Right := NewRect.Right + MulDiv(Rect.Right - NewRect.Right,
      Progress, 100);
    R.Bottom := NewRect.Bottom + MulDiv(Rect.Bottom - NewRect.Bottom,
      Progress, 100);
    StretchBlt(DC, Rect.Left, Rect.Top, Rect.Right - Rect.Left,
      Rect.Bottom - Rect.Top, NewBmp.Canvas.Handle, R.Left - OfsX, R.Top - OfsY,
      R.Right - R.Left, R.Bottom - R.Top, SRCCOPY);
  Until Progress = 100;
End;

{ TDateTimeList }

Destructor TDateTimeList.Destroy;
Begin
  Capacity := 0;
  Inherited Destroy;
End;

Function TDateTimeList.GetCapacity: Integer;
Begin
  Result := Length(FList);
End;

Procedure TDateTimeList.SetCapacity(Value: Integer);
Begin
  If Capacity <> Value Then
  Begin
    SetLength(FList, Value);
    If Capacity < Count Then
    Begin
      Changing;
      FCount := Capacity;
      Change;
    End;
  End;
End;

Function TDateTimeList.GetItems(Index: Integer): TDate;
Begin
  If (Index < 0) Or (Index >= Count) Then
    Raise EListError.CreateFmt(SListIndexError, [Index]);
  Result := FList[Index];
End;

Procedure TDateTimeList.Changing;
Begin
  If (FUpdateCount = 0) And Assigned(OnChanging) Then
    OnChanging(Self);
End;

Procedure TDateTimeList.Change;
Begin
  If (FUpdateCount = 0) And Assigned(OnChange) Then
    OnChange(Self);
End;

Procedure TDateTimeList.Assign(Source: TPersistent);
Begin
  If Source Is TDateTimeList Then
  Begin
    Changing;
    FList := Copy(TDateTimeList(Source).FList, 0, TDateTimeList(Source).FCount);
    FCount := TDateTimeList(Source).FCount;
    Change;
  End
  Else
    Inherited Assign(Source);
End;

Procedure TDateTimeList.BeginUpdate;
Begin
  If FUpdateCount = 0 Then
    Changing;
  Inc(FUpdateCount);
End;

Procedure TDateTimeList.EndUpdate;
Begin
  Dec(FUpdateCount);
  If FUpdateCount = 0 Then
    Change;
End;

Procedure TDateTimeList.Clear;
Begin
  If Count <> 0 Then
  Begin
    Changing;
    FCount := 0;
    Change;
  End;
End;

Function TDateTimeList.Add(Const ADateTime: TDateTime): Integer;
Var
  Index: Integer;
Begin
  If Not Find(ADateTime, Index) Then
  Begin
    Changing;
    If Capacity = Count Then
      Capacity := Count + 4;
    If Index <> Count Then
      Move(FList[Index], FList[Index + 1], (Count - Index) * SizeOf(TDateTime));
    FList[Index] := ADateTime;
    Inc(FCount);
    Change;
  End;
  Result := Index;
End;

Function TDateTimeList.Remove(Const ADateTime: TDateTime): Integer;
Var
  Index: Integer;
Begin
  Result := -1;
  If Find(ADateTime, Index) Then
    Delete(Index);
End;

Procedure TDateTimeList.Delete(Index: Integer);
Begin
  If (Index < 0) Or (Index >= Count) Then
    Raise EListError.CreateFmt(SListIndexError, [Index]);
  Changing;
  Dec(FCount);
  If Index <> Count Then
    Move(FList[Index + 1], FList[Index], (Count - Index) * SizeOf(TDateTime));
  Change;
End;

Function TDateTimeList.Exists(Const ADateTime: TDateTime): Boolean;
Var
  Index: Integer;
Begin
  Result := Find(ADateTime, Index);
End;

Function TDateTimeList.IndexOf(Const ADateTime: TDateTime): Integer;
Begin
  If Not Find(ADateTime, Result) Then
    Result := -1;
End;

Function TDateTimeList.Find(Const ADateTime: TDateTime;
  Out Index: Integer): Boolean;
Var
  L, H, M: Integer;
  C: Integer;
Begin
  L := 0;
  H := Count - 1;
  While L <= H Do
  Begin
    M := (L + H) Shr 1;
    C := TCalendar.Compare(FList[M], ADateTime);
    If C < 0 Then
      L := M + 1
    Else
    Begin
      H := M - 1;
      If C = 0 Then
      Begin
        Result := True;
        Index := M;
        Exit;
      End;
    End;
  End;
  Index := L;
  Result := False;
End;

Function TDateTimeList.First: TDateTime;
Begin
  If Count <> 0 Then
    Result := Items[0]
  Else
    Result := TCalendar.NoDate;
End;

Function TDateTimeList.Last: TDateTime;
Begin
  If Count <> 0 Then
    Result := Items[Count - 1]
  Else
    Result := TCalendar.NoDate;
End;

{ TCustomIntlMonthCalendar }

Constructor TCustomIntlMonthCalendar.Create(AOwner: TComponent);
Begin
  Inherited Create(AOwner);
  ControlStyle := ControlStyle + [CsOpaque] - [CsSetCaption];
  FCalendar := DefaultCalendar.Clone;
  FCalendar.OnChange := CalendarChanged;
  FSelection := TDateTimeList.Create;
  FSelection.Clear;
  FFontHeader := TFont.Create;
  FFontHeader.OnChange := FontHeaderChanged;
  FFontFooter := TFont.Create;
  FFontFooter.OnChange := FontFooterChanged;
  FFontDayOfWeek := TFont.Create;
  FFontDayOfWeek.OnChange := FontDayOfWeekChanged;
  FFontWeekNumbers := TFont.Create;
  FFontWeekNumbers.OnChange := FontWeekNumbersChanged;
  FDate := TCalendar.Today;
  FDateMin := TCalendar.NoDate;
  FDateMax := TCalendar.NoDate;
  FColorDivider := ClBtnFace;
  FColorToday := ClRed;
  FShowHeader := True;
  FShowFooter := True;
  FAnimate := True;
  FHintDate := Calendar.NoDate;
  FDraggingCell := -1;
  FHotCell := -1;
  ParentColor := False;
  Color := ClWindow;
  BevelKind := BkTile;
  BorderWidth := 2;
  TabStop := True;
  UpdateCalendar;
  Width := 180;
  Height := 180;
  FEventItems := TEventItems.Create(Self);
  CalendarType := TJalaliCalendar;
  AutoToday := True;
End;

Destructor TCustomIntlMonthCalendar.Destroy;
Begin
  FSelection.Free;
  FFontHeader.Free;
  FFontFooter.Free;
  FFontDayOfWeek.Free;
  FFontWeekNumbers.Free;
  FCalendar.Free;
  If Assigned(FSavedScreen) Then
    FSavedScreen.Free;
  If FTheme <> 0 Then
    CloseThemeData(FTheme);
  Inherited Destroy;
End;

Procedure TCustomIntlMonthCalendar.SetDate(Const Value: TDate);
Begin
  If (CsDesigning In ComponentState) And Calendar.IsNoDate(Value) Then
  Begin
    Select(Calendar.Today);
    Exit;
  End;
  If Not IsDateInRange(Value) Then
  Begin
    Raise EDateTimeError.CreateResFmt(@SDateOutOfRangeError,
      [Calendar.Format('ddddd', ActualDateMin), Calendar.Format('ddddd',
      ActualDateMax)]);
  End;
  If CsLoading In ComponentState Then
  Begin
    FDate := Trunc(Value);
    FSelectionAnchor := FDate;
  End
  Else
    Select(Value);
End;

Procedure TCustomIntlMonthCalendar.SetDateMin(Const Value: TDate);
Begin
  If Calendar.IsNoDate(Value) Then
    FDateMin := Calendar.NoDate
  Else
  Begin
    FDateMin := Trunc(Value);
    If Not Calendar.IsNoDate(DateMax) And
      (Calendar.CompareDate(DateMin, DateMax) > 0) Then
      FDateMin := FDateMax;
  End;
  UpdateMinMaxRanges;
End;

Procedure TCustomIntlMonthCalendar.SetEventItems(Const Value: TEventItems);
Begin
  If FEventItems <> Value Then
  Begin
    FEventItems := Value;
    Invalidate;
  End;
End;

Procedure TCustomIntlMonthCalendar.SetDate(DateStr, FmtStr: String);
Begin
  Date := Calendar.Parse(FmtStr, DateStr);
End;

Procedure TCustomIntlMonthCalendar.SetDateMax(Const Value: TDate);
Begin
  If Calendar.IsNoDate(Value) Then
    FDateMax := Calendar.NoDate
  Else
  Begin
    FDateMax := Trunc(Value);
    If Not Calendar.IsNoDate(DateMin) And
      (Calendar.CompareDate(DateMin, DateMax) > 0) Then
      FDateMax := FDateMin;
  End;
  UpdateMinMaxRanges;
End;

Procedure TCustomIntlMonthCalendar.SetAutoToday(Const Value: Boolean);
Begin
  FAutoToday := Value;
  If FAutoToday Then
    SetToday;
End;

Procedure TCustomIntlMonthCalendar.SetCalendarType(Value: TCalendarClass);
Begin
  If CalendarType <> Value Then
  Begin
    FCalendarType := Value;
    ValidateCalendar;
  End;
End;

Procedure TCustomIntlMonthCalendar.SetWeekNumbers(Value: Boolean);
Begin
  If WeekNumbers <> Value Then
  Begin
    FWeekNumbers := Value;
    UpdateMetrics;
  End;
End;

Procedure TCustomIntlMonthCalendar.SetSelectionStyle
  (Value: TIntlCalSelectionStyle);
Var
  OldStyle: TIntlCalSelectionStyle;
  TheDate: TDate;
  I: Integer;
Begin
  If SelectionStyle <> Value Then
  Begin
    OldStyle := SelectionStyle;
    FSelectionStyle := Value;
    If Not(CsLoading In ComponentState) Then
    Begin
      Case OldStyle Of
        SsSingle:
          Begin
            If SelectionStyle = SsFree Then
              FSelection.Add(Date);
            FSelectionAnchor := Date;
          End;
        SsRange:
          Begin
            If SelectionStyle = SsFree Then
            Begin
              TheDate := Min(Date, FSelectionAnchor);
              For I := 0 To Calendar.DaysBetween(Date, FSelectionAnchor) Do
                FSelection.Add(Calendar.NextDay(TheDate, I));
            End
            Else
              InvalidateDateRange(FSelectionAnchor, Date);
            FSelectionAnchor := Date;
          End;
        SsFree:
          Begin
            If (SelectionStyle <> SsRange) Or (FSelection.Count = 0) Then
              FSelectionAnchor := Date
            Else If FSelection.First = Date Then
              FSelectionAnchor := FSelection.Last
            Else If FSelection.Last = Date Then
              FSelectionAnchor := FSelection.First
            Else
            Begin
              FSelectionAnchor := FSelection.First;
              FocusDate(FSelection.Last);
            End;
            FSelection.Clear;
            InvalidateDateRange(FSelectionAnchor, Date);
          End;
      End;
    End
  End;
End;

Procedure TCustomIntlMonthCalendar.SetSettings(Const Value: TCalendarSettings);
Begin
  FCalendar.Settings := Value;
  UpdateCalendar;
End;

Procedure TCustomIntlMonthCalendar.SetShowHeader(Value: Boolean);
Begin
  If ShowHeader <> Value Then
  Begin
    FShowHeader := Value;
    UpdateMetrics;
  End;
End;

Procedure TCustomIntlMonthCalendar.SetShowEvents(Const Value: Boolean);
Begin
  FShowEvents := Value;
  Invalidate;
End;

Procedure TCustomIntlMonthCalendar.SetShowFooter(Value: Boolean);
Begin
  If ShowFooter <> Value Then
  Begin
    FShowFooter := Value;
    UpdateMetrics;
  End;
End;

Procedure TCustomIntlMonthCalendar.SetColorDivider(Value: TColor);
Begin
  If ColorDivider <> Value Then
  Begin
    FColorDivider := Value;
    InvalidateSections([CsDaysOfWeek, CsWeekNumbers]);
  End;
End;

Procedure TCustomIntlMonthCalendar.SetColorToday(Value: TColor);
Begin
  If ColorToday <> Value Then
  Begin
    FColorToday := Value;
    InvalidateDate(Calendar.Today);
    InvalidateSections([CsFooter]);
  End;
End;

Procedure TCustomIntlMonthCalendar.SetFontHeader(Value: TFont);
Begin
  FFontHeader.Assign(Value);
End;

Procedure TCustomIntlMonthCalendar.SetFontFooter(Value: TFont);
Begin
  FFontFooter.Assign(Value);
End;

Procedure TCustomIntlMonthCalendar.SetFontDayOfWeek(Value: TFont);
Begin
  FFontDayOfWeek.Assign(Value);
End;

Procedure TCustomIntlMonthCalendar.SetFontWeekNumbers(Value: TFont);
Begin
  FFontWeekNumbers.Assign(Value);
End;

Procedure TCustomIntlMonthCalendar.SetCurrentView(Value: TIntlCalViewMode);
Begin
  If CurrentView <> Value Then
  Begin
    If Not(CsLoading In ComponentState) Then
    Begin
      FHintDate := Calendar.NoDate;
      FDraggingCell := -1;
      FHotCell := -1;
      FHotInfo := ChtNowhere;
      AnimateBegin;
      Try
        FCurrentView := Value;
        CalcMetrics(CurrentView, FMetrics);
        CalcMetricsSections(FMetrics);
        AnimatePlay;
      Finally
        AnimateEnd;
      End;
      DoPageChange;
      Invalidate;
    End
    Else
      FCurrentView := Value;
  End;
End;

Procedure TCustomIntlMonthCalendar.NormalizeDigits(Var Str: String);
Begin
  // If Assigned(Culture) Then
  // Str := Culture.FreezeDigits(Str, False);
End;

Function TCustomIntlMonthCalendar.GetDayName(DayOfWeek: TDayOfWeek;
  NameKind: TIntlCalNameKind): String;
Begin
  Case NameKind Of
    NkLong:
      Result := Calendar.Settings.DayNames[DayOfWeek];
    NkShort:
      Result := Calendar.Settings.ShortDayNames[DayOfWeek];
    NkShortest:
      Result := Calendar.Settings.ShortestDayNames[DayOfWeek];
  Else
    Result := '';
  End;
  NormalizeDigits(Result);
End;

Function TCustomIntlMonthCalendar.GetMonthName(Month: Integer;
  NameKind: TIntlCalNameKind): String;
Begin
  If NameKind = NkLong Then
    Result := Calendar.Settings.MonthNames[Month]
  Else
    Result := Calendar.Settings.ShortMonthNames[Month];
  NormalizeDigits(Result);
End;

Function TCustomIntlMonthCalendar.GetYearText(AYear: Integer): String;
Var
  YearStartDate: TDate;
Begin
  YearStartDate := Calendar.StartOfYear(Era, AYear);
  Result := Format(Calendar.Settings.YearFormat, YearStartDate);
End;

Procedure TCustomIntlMonthCalendar.SetToday;
Begin
  SetDate(Now);
End;

Function TCustomIntlMonthCalendar.GetDayText(ADay: Integer): String;
Begin
  Result := IntToStr(ADay);
  NormalizeDigits(Result);
End;

Function TCustomIntlMonthCalendar.GetWeekText(AWeek: Integer): String;
Begin
  Result := IntToStr(AWeek);
  NormalizeDigits(Result);
End;

Function TCustomIntlMonthCalendar.GetHeaderText
  (ViewMode: TIntlCalViewMode): String;
Begin
  If ViewMode = VmYear Then
  Begin
    If IsRightToLeft Then
      Result := GetYearText(LastYearOnPage) + ' - ' +
        GetYearText(FirstYearOnPage)
    Else
      Result := GetYearText(FirstYearOnPage) + ' - ' +
        GetYearText(LastYearOnPage);
  End
  Else
  Begin
    If ViewMode = VmMonth Then
      Result := Format(Calendar.Settings.YearFormat, Date)
    Else
      Result := Format(Calendar.Settings.YearMonthFormat, Date);
  End;
End;

Function TCustomIntlMonthCalendar.GetFooterText(LongFormat: Boolean): String;
Begin
  If LongFormat Then
    Result := Format(Calendar.Settings.LongDateFormat, Calendar.Today)
  Else
    Result := Format(Calendar.Settings.ShortDateFormat, Calendar.Today);
End;

Function TCustomIntlMonthCalendar.GetMonthLastDate: TDate;
Begin
  Result := MonthFirstDate + DaysInMonth - 1;
End;

Function TCustomIntlMonthCalendar.GetFirstCellDate: TDate;
Begin
  Result := MonthFirstDate - MonthStartCell;
End;

Function TCustomIntlMonthCalendar.GetLastCellDate: TDate;
Begin
  Result := FirstCellDate + CellCount - 1;
End;

Function TCustomIntlMonthCalendar.GetLastCellYear: Integer;
Begin
  Result := FirstCellYear + CellCount - 1;
End;

Function TCustomIntlMonthCalendar.GetFirstYearOnPage: Integer;
Begin
  Result := FirstCellYear;
  While (Result < Year) And Not(IsYearOnPage(Result) And IsYearInRange(Era,
    Result)) Do
    Result := Calendar.OffsetYear(Era, Result, +1);
End;

Function TCustomIntlMonthCalendar.GetLastYearOnPage: Integer;
Begin
  Result := LastCellYear;
  While (Result > Year) And Not(IsYearOnPage(Result) And IsYearInRange(Era,
    Result)) Do
    Result := Calendar.OffsetYear(Era, Result, -1);
End;

Function TCustomIntlMonthCalendar.GetColCount: Integer;
Begin
  Result := FMetrics.ColCount;
End;

Function TCustomIntlMonthCalendar.GetRowCount: Integer;
Begin
  Result := FMetrics.RowCount;
End;

Function TCustomIntlMonthCalendar.GetCellCount: Integer;
Begin
  Result := RowCount * ColCount;
End;

Function TCustomIntlMonthCalendar.GetFocusedCell: Integer;
Begin
  Case CurrentView Of
    VmDay:
      Result := MonthStartCell + Day - 1;
    VmMonth:
      Result := Month - 1;
    VmYear:
      Result := Year - FirstCellYear;
  Else
    Result := -1;
  End;
End;

Function TCustomIntlMonthCalendar.GetSelectionStart: TDate;
Begin
  Case SelectionStyle Of
    SsSingle:
      Result := Date;
    SsRange:
      Result := Min(Date, FSelectionAnchor);
  Else
    Result := FSelection.First;
  End;
End;

Function TCustomIntlMonthCalendar.GetSettings: TCalendarSettings;
Begin
  Result := FCalendar.Settings;
End;

Function TCustomIntlMonthCalendar.GetSelectionEnd: TDate;
Begin
  Case SelectionStyle Of
    SsSingle:
      Result := Date;
    SsRange:
      Result := Max(Date, FSelectionAnchor);
  Else
    Result := FSelection.Last;
  End;
End;

Function TCustomIntlMonthCalendar.GetSelectionCount: Integer;
Begin
  Case SelectionStyle Of
    SsSingle:
      Result := 1;
    SsRange:
      Result := Calendar.DaysBetween(FSelectionAnchor, Date) + 1;
  Else
    Result := FSelection.Count;
  End;
End;

Function TCustomIntlMonthCalendar.GetSelection(Index: Integer): TDate;
Begin
  Case SelectionStyle Of
    SsSingle:
      Begin
        If Index <> 0 Then
          Raise EListError.CreateFmt(SListIndexError, [Index]);
        Result := Date;
      End;
    SsRange:
      Begin
        If (Index < 0) Or (Index > Calendar.DaysBetween(FSelectionAnchor,
          Date)) Then
          Raise EListError.CreateFmt(SListIndexError, [Index]);
        Result := Min(Date, FSelectionAnchor) + Index;
      End
  Else
    Result := FSelection[Index];
  End;
End;

Function TCustomIntlMonthCalendar.GetAdjacentPageDate(Backward: Boolean): TDate;
Var
  Dir: Integer;
Begin
  If Backward Then
    Dir := -1
  Else
    Dir := +1;
  Result := Date;
  Case CurrentView Of
    VmDay:
      Calendar.IncMonth(Result, Dir);
    VmMonth:
      Calendar.IncYear(Result, Dir);
    VmYear:
      Calendar.IncYear(Result, Dir * 10);
  End;
End;

Function TCustomIntlMonthCalendar.GetCalendarClientRect: TRect;
Var
  TL, BR: Integer;
Begin
  If FUpdateMetricsCount = 0 Then
  Begin
    Result := FMetrics.Bounds;
    Exit;
  End;
  Result := ClientRect;
  If BevelKind In [BkTile, BkSoft] Then
  Begin
    TL := 0;
    BR := 0;
    If (BevelInner = BvLowered) Or
      ((BevelInner = BvRaised) And (BevelKind = BkSoft)) Then
      Inc(TL);
    If BevelOuter In [BvLowered, BvRaised] Then
      Inc(BR);
    If BeLeft In BevelEdges Then
      Inc(Result.Left, TL);
    If BeTop In BevelEdges Then
      Inc(Result.Top, TL);
    If BeRight In BevelEdges Then
      Dec(Result.Right, BR);
    If BeBottom In BevelEdges Then
      Dec(Result.Bottom, BR);
  End;
End;

Function TCustomIntlMonthCalendar.GetSectionRect
  (Section: TIntlCalSection): TRect;
Begin
  Result := FMetrics.Sections[Section];
End;

Function TCustomIntlMonthCalendar.GetSectionWidth
  (Section: TIntlCalSection): Integer;
Begin
  With FMetrics.Sections[Section] Do
    Result := Right - Left;
End;

Function TCustomIntlMonthCalendar.GetSectionHeight
  (Section: TIntlCalSection): Integer;
Begin
  With FMetrics.Sections[Section] Do
    Result := Bottom - Top;
End;

Function TCustomIntlMonthCalendar.GetCellRect(Row, Col: Integer): TRect;
Begin
  If (Row < 0) And (Col < 0) Then
    Result := FMetrics.Sections[CsWeekSpace]
  Else If Row < 0 Then
  Begin
    Result := FMetrics.Sections[CsDaysOfWeek];
    Dec(Result.Bottom, 2);
  End
  Else If Col < 0 Then
  Begin
    Result := FMetrics.Sections[CsWeekNumbers];
    If IsRightToLeft Then
      Inc(Result.Left, 2)
    Else
      Dec(Result.Right, 2);
  End
  Else
    Result := FMetrics.Sections[CsYMD];
  If Row >= 0 Then
  Begin
    Inc(Result.Top, Row * FMetrics.RowHeight + Min(Row,
      FMetrics.StretchedRows));
    Result.Bottom := Result.Top + FMetrics.RowHeight;
    If Row < FMetrics.StretchedRows Then
      Inc(Result.Bottom);
  End;
  If Col >= 0 Then
  Begin
    If IsRightToLeft And (CurrentView In [VmDay, VmMonth]) Then
    Begin
      Dec(Result.Right, Col * FMetrics.ColWidth + Min(Col,
        FMetrics.StretchedCols));
      Result.Left := Result.Right - FMetrics.ColWidth;
      If Col < FMetrics.StretchedCols Then
        Dec(Result.Left);
    End
    Else
    Begin
      Inc(Result.Left, Col * FMetrics.ColWidth + Min(Col,
        FMetrics.StretchedCols));
      Result.Right := Result.Left + FMetrics.ColWidth;
      If Col < FMetrics.StretchedCols Then
        Inc(Result.Right);
    End;
  End;
End;

Function TCustomIntlMonthCalendar.GetCellRect(Index: Integer): TRect;
Begin
  Result := GetCellRect(Index Div ColCount, Index Mod ColCount);
End;

Function TCustomIntlMonthCalendar.GetRowAt(Y: Integer): Integer;
Var
  D: Integer;
Begin
  Dec(Y, FMetrics.Sections[CsYMD].Top);
  D := (FMetrics.RowHeight + 1) * FMetrics.StretchedRows;
  If Y < D Then
    Result := Y Div (FMetrics.RowHeight + 1)
  Else
    Result := FMetrics.StretchedRows + (Y - D) Div FMetrics.RowHeight;
End;

Function TCustomIntlMonthCalendar.GetColAt(X: Integer): Integer;
Var
  D: Integer;
Begin
  Dec(X, FMetrics.Sections[CsYMD].Left);
  If IsRightToLeft And (CurrentView In [VmDay, VmMonth]) Then
    X := (FMetrics.Sections[CsYMD].Right - FMetrics.Sections[CsYMD]
      .Left) - X - 1;
  D := (FMetrics.ColWidth + 1) * FMetrics.StretchedCols;
  If X < D Then
    Result := X Div (FMetrics.ColWidth + 1)
  Else
    Result := FMetrics.StretchedCols + (X - D) Div FMetrics.ColWidth;
End;

Function TCustomIntlMonthCalendar.GetPreferredClientSize(Out AWidth,
  AHeight: Integer): Boolean;
Var
  Metrics: TIntlCalMetrics;
  Mode: TIntlCalViewMode;
Begin
  If Not FPrefferedSizeReady Then
  Begin
    FPrefferedSizeReady := True;
    FPrefferedSize.Cx := FMetrics.PrefferedWidth;
    FPrefferedSize.Cy := FMetrics.PrefferedHeight;
    FPrefferedSizeSmallest := (FMetrics.NameKind = NkShortest);
    For Mode := Low(TIntlCalViewMode) To High(TIntlCalViewMode) Do
    Begin
      CalcMetrics(Mode, Metrics);
      If FPrefferedSize.Cx < Metrics.PrefferedWidth Then
        FPrefferedSize.Cx := Metrics.PrefferedWidth;
      If FPrefferedSize.Cy < Metrics.PrefferedHeight Then
        FPrefferedSize.Cy := Metrics.PrefferedHeight;
      If Mode = VmDay Then
        FPrefferedSizeSmallest := (Metrics.NameKind = NkShortest);
    End;
  End;
  AWidth := FPrefferedSize.Cx;
  AHeight := FPrefferedSize.Cy;
  Result := FPrefferedSizeSmallest;
End;

Function TCustomIntlMonthCalendar.GetPreferredControlSize(Out AWidth,
  AHeight: Integer): Boolean;
Var
  EdgeSize: Integer;
Begin
  Result := GetPreferredClientSize(AWidth, AHeight);
  Inc(AWidth, 2 * BorderWidth);
  Inc(AHeight, 2 * BorderWidth);
  If BevelKind <> BkNone Then
  Begin
    EdgeSize := 0;
    If BevelInner <> BvNone Then
      Inc(EdgeSize, BevelWidth);
    If BevelOuter <> BvNone Then
      Inc(EdgeSize, BevelWidth);
    If BeLeft In BevelEdges Then
      Inc(AWidth, EdgeSize);
    If BeRight In BevelEdges Then
      Inc(AWidth, EdgeSize);
    If BeTop In BevelEdges Then
      Inc(AHeight, EdgeSize);
    If BeBottom In BevelEdges Then
      Inc(AHeight, EdgeSize);
  End;
End;

Function TCustomIntlMonthCalendar.GetHitTestInfoAt(X, Y: Integer;
  Out Cell: Integer): TIntlCalHitTest;
Var
  Pt: TPoint;
Begin
  Pt.X := X;
  Pt.Y := Y;
  Cell := -1;
  Result := ChtNowhere;
  If PtInRect(FMetrics.Bounds, Pt) Then
  Begin
    If PtInRect(FMetrics.Sections[CsHeader], Pt) Then
    Begin
      If PtInRect(FMetrics.Sections[CsHeaderText], Pt) Then
        Result := ChtHeaderText
      Else If PtInRect(FMetrics.Sections[CsArrowLeft], Pt) Then
        Result := ChtArrowLeft
      Else If PtInRect(FMetrics.Sections[CsArrowRight], Pt) Then
        Result := ChtArrowRight;
    End
    Else If PtInRect(FMetrics.Sections[CsFooter], Pt) Then
      Result := ChtFooter
    Else If (CurrentView = VmDay) And
      PtInRect(FMetrics.Sections[CsDaysOfWeek], Pt) Then
      Result := ChtDayOfWeek
    Else If (CurrentView = VmDay) And WeekNumbers And
      PtInRect(FMetrics.Sections[CsWeekNumbers], Pt) Then
      Result := ChtWeekNumber
    Else If PtInRect(FMetrics.Sections[CsYMD], Pt) Then
    Begin
      Cell := GetRowAt(Y) * FMetrics.ColCount + GetColAt(X);
      Case CurrentView Of
        VmDay:
          Result := ChtDay;
        VmMonth:
          Result := ChtMonth;
        VmYear:
          Result := ChtYear;
      End;
    End;
  End
  Else
End;

Function TCustomIntlMonthCalendar.GetTheme: HTHEME;
Var
  ListView: TListView;
Begin
  If FTheme = 0 Then
  Begin
    ListView := TListView.Create(Nil);
    Try
      ListView.Visible := False;
      ListView.ParentWindow := Handle;
      FTheme := OpenThemeData(ListView.Handle, 'listview');
    Finally
      ListView.Free;
    End;
  End;
  Result := FTheme;
End;

Function TCustomIntlMonthCalendar.IsDateStored: Boolean;
Begin
  Result := Not Calendar.IsToday(Date);
End;

Function TCustomIntlMonthCalendar.IsDateMinStored: Boolean;
Begin
  Result := Not Calendar.IsNoDate(DateMin);
End;

Function TCustomIntlMonthCalendar.IsDateMaxStored: Boolean;
Begin
  Result := Not Calendar.IsNoDate(DateMax);
End;

Procedure TCustomIntlMonthCalendar.WMEraseBkgnd(Var Message: TWMEraseBkgnd);
Begin
  Message.Result := 1;
End;

Procedure TCustomIntlMonthCalendar.WMGetDlgCode(Var Message: TWMGetDlgCode);
Begin
  Message.Result := DLGC_WANTARROWS;
End;

Procedure TCustomIntlMonthCalendar.WMMouseActivate
  (Var Message: TWMMouseActivate);
Begin
  Inherited;
  If CanFocus Then
    SetFocus;
End;

Procedure TCustomIntlMonthCalendar.WMTimer(Var Message: TWMTimer);
Var
  TheDate: TDate;
Begin
  If NavigationTimer = AutoNavTimerForward Then
    TheDate := GetAdjacentPageDate(False)
  Else If NavigationTimer = AutoNavTimerBackward Then
    TheDate := GetAdjacentPageDate(True)
  Else
  Begin
    Inherited;
    Exit;
  End;
  If IsDateInRange(TheDate) Then
    SelectDate(TheDate);
End;

Procedure TCustomIntlMonthCalendar.CMMouseLeave(Var Message: TMessage);
Begin
  Inherited;
  InvalidateHotLight;
  FHotCell := -1;
  FHotInfo := ChtNowhere;
End;

Procedure TCustomIntlMonthCalendar.CMHintShow(Var Message: TCMHintShow);
Var
  TheDate: TDate;
  TheCell: Integer;
Begin
  TheCell := -1;
  With Message.HintInfo^ Do
  Begin
    If (GetHitTestInfoAt(CursorPos.X, CursorPos.Y, TheCell) = ChtDay) And
      GetDateAtCell(TheCell, TheDate) Then
    Begin
      If TheDate <> FHintDate Then
      Begin
        FHintDate := TheDate;
        DoDateHint(TheDate, HintStr);
      End;
    End
    Else
    Begin
      FHintDate := Calendar.NoDate;
      Inherited;
    End;
    CursorRect := GetCellRect(TheCell);
  End;
End;

Procedure TCustomIntlMonthCalendar.CMFontChanged(Var Message: TMessage);
Begin
  Inherited;
  Inc(FUpdateMetricsCount);
  Try
    If Not FSaveFontHeader Then
      FontHeader.Assign(Font);
    If Not FSaveFontFooter Then
      FontFooter.Assign(Font);
    If Not FSaveFontDayOfWeek Then
      FontDayOfWeek.Assign(Font);
    If Not FSaveFontWeekNumbers Then
      FontWeekNumbers.Assign(Font);
  Finally
    Dec(FUpdateMetricsCount);
  End;
  UpdateMetrics;
End;

Procedure TCustomIntlMonthCalendar.DoDateChange;
Begin
  If Assigned(OnDateChange) Then
    OnDateChange(Self);
End;

Procedure TCustomIntlMonthCalendar.DoPageChange;
Begin
  If Assigned(OnPageChange) Then
    OnPageChange(Self);
End;

Procedure TCustomIntlMonthCalendar.DoCalendarChange;
Begin
  If Assigned(OnCalendarChange) Then
    OnCalendarChange(Self);
End;

Procedure TCustomIntlMonthCalendar.DoSelectionChange;
Begin
  If Assigned(OnSelectionChange) Then
    OnSelectionChange(Self);
End;

Procedure TCustomIntlMonthCalendar.DoDateHint(Const ADate: TDate;
  Var HintStr: String);
Begin
  HintStr := Hint;
  If Assigned(OnDateHint) Then
    OnDateHint(Self, ADate, HintStr);
End;

Function TCustomIntlMonthCalendar.DoDateSelectable(Const ADate: TDate): Boolean;
Begin
  Result := True;
  If Assigned(OnDateSelectable) Then
    OnDateSelectable(Self, ADate, Result);
End;

Function TCustomIntlMonthCalendar.DoCustomDrawDayOfWeek(Canvas: TCanvas;
  Const ARect: TRect; State: TIntlCalDrawStates;
  ADayOfWeek: TDayOfWeek): Boolean;
Begin
  Result := True;
  If Assigned(OnCustomDrawDayOfWeek) Then
    OnCustomDrawDayOfWeek(Self, Canvas, ARect, State, ADayOfWeek, Result);
End;

Function TCustomIntlMonthCalendar.DoCustomDrawWeekNumber(Canvas: TCanvas;
  Const ARect: TRect; State: TIntlCalDrawStates; AWeek: Integer): Boolean;
Begin
  Result := True;
  If Assigned(OnCustomDrawWeekNumber) Then
    OnCustomDrawWeekNumber(Self, Canvas, ARect, State, AWeek, Result);
End;

Function TCustomIntlMonthCalendar.DoCustomDrawDay(Canvas: TCanvas;
  Const ARect: TRect; State: TIntlCalDrawStates; Const ADate: TDate;
  ADay: Integer): Boolean;
Begin
  Result := True;
  If Assigned(OnCustomDrawDay) Then
    OnCustomDrawDay(Self, Canvas, ARect, State, ADate, ADay, Result);
End;

Function TCustomIntlMonthCalendar.DoCustomDrawMonth(Canvas: TCanvas;
  Const ARect: TRect; State: TIntlCalDrawStates; AMonth: Integer): Boolean;
Begin
  Result := True;
  If Assigned(OnCustomDrawMonth) Then
    OnCustomDrawMonth(Self, Canvas, ARect, State, AMonth, Result);
End;

Function TCustomIntlMonthCalendar.DoCustomDrawYear(Canvas: TCanvas;
  Const ARect: TRect; State: TIntlCalDrawStates; AYear: Integer): Boolean;
Begin
  Result := True;
  If Assigned(OnCustomDrawYear) Then
    OnCustomDrawYear(Self, Canvas, ARect, State, AYear, Result);
End;

Function TCustomIntlMonthCalendar.DoCustomDrawHeaderText(Canvas: TCanvas;
  Const ARect: TRect; State: TIntlCalDrawStates; Const AText: String): Boolean;
Begin
  Result := True;
  If Assigned(OnCustomDrawHeaderText) Then
    OnCustomDrawHeaderText(Self, Canvas, ARect, State, AText, Result);
End;

Function TCustomIntlMonthCalendar.DoCustomDrawFooterText(Canvas: TCanvas;
  Const ARect: TRect; State: TIntlCalDrawStates; Const AText: String): Boolean;
Begin
  Result := True;
  If Assigned(OnCustomDrawFooterText) Then
    OnCustomDrawFooterText(Self, Canvas, ARect, State, AText, Result);
End;

Procedure TCustomIntlMonthCalendar.FontHeaderChanged(Sender: TObject);
Begin
  FSaveFontHeader := (FontHeader.Handle <> Font.Handle);
  UpdateMetrics;
End;

Procedure TCustomIntlMonthCalendar.FontFooterChanged(Sender: TObject);
Begin
  FSaveFontFooter := (FontFooter.Handle <> Font.Handle);
  UpdateMetrics;
End;

Procedure TCustomIntlMonthCalendar.FontDayOfWeekChanged(Sender: TObject);
Begin
  FSaveFontDayOfWeek := (FontDayOfWeek.Handle <> Font.Handle);
  UpdateMetrics;
End;

Procedure TCustomIntlMonthCalendar.FontWeekNumbersChanged(Sender: TObject);
Begin
  FSaveFontWeekNumbers := (FontWeekNumbers.Handle <> Font.Handle);
  If WeekNumbers Then
    UpdateMetrics;
End;

Procedure TCustomIntlMonthCalendar.CalendarChanged(Sender: TObject);
Begin
  UpdateCalendar;
  DoCalendarChange;
End;

Procedure TCustomIntlMonthCalendar.ReadCalendarType(Reader: TReader);
Begin
  CalendarType := CalendarTypes.ByName(Reader.ReadString);
End;

Procedure TCustomIntlMonthCalendar.WriteCalendarType(Writer: TWriter);
Begin
  Writer.WriteString(CalendarType.CalendarName);
End;

Procedure TCustomIntlMonthCalendar.DefineProperties(Filer: TFiler);
Begin
  Inherited DefineProperties(Filer);
  Filer.DefineProperty('CalendarClass', ReadCalendarType, WriteCalendarType,
    Assigned(CalendarType));
End;

Procedure TCustomIntlMonthCalendar.Notification(AComponent: TComponent;
  Operation: TOperation);
Begin
  Inherited Notification(AComponent, Operation);
End;

Procedure TCustomIntlMonthCalendar.Resize;
Begin
  Inherited Resize;
  UpdateMetrics;
End;

Procedure TCustomIntlMonthCalendar.Paint;
Var
  Rect: TRect;
Begin
  DrawCalendar(Canvas);
  Rect := ClientRect;
  If Not EqualRect(FMetrics.Bounds, Rect) Then
  Begin
    With FMetrics.Bounds Do
      ExcludeClipRect(Canvas.Handle, Left, Top, Right, Bottom);
    Canvas.Brush.Color := Color;
    Canvas.FillRect(Rect);
  End;
End;

Function TCustomIntlMonthCalendar.Parse(Const FmtStr, DateTimeStr: String)
  : TDateTime;
Begin
  Result := Calendar.Parse(FmtStr, DateTimeStr);
End;

Procedure TCustomIntlMonthCalendar.Loaded;
Begin
  Inherited Loaded;
  If (Assigned(CalendarType) { Or Assigned(Culture) } ) Then
    ValidateCalendar
  Else
  Begin
    UpdateCalendar;
    DoCalendarChange;
  End;
  DoPageChange;
  DoDateChange;
  DoSelectionChange;
End;

Function TCustomIntlMonthCalendar.CanAutoSize(Var NewWidth,
  NewHeight: Integer): Boolean;
Begin
  Result := True;
  If AutoSize And Not(CsLoading In ComponentState) Then
    GetPreferredControlSize(NewWidth, NewHeight);
End;

Procedure TCustomIntlMonthCalendar.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
Var
  TheDate: TDate;
  TheYear, TheMonth, TheDay: Integer;
  TheCell: Integer;
Begin
  FDraggingCell := -1;
  If (Button = MbLeft) And Not(SsDouble In Shift) Then
  Begin
    Case GetHitTestInfoAt(X, Y, TheCell) Of
      ChtArrowLeft:
        Begin
          TheDate := GetAdjacentPageDate(True);
          If IsDateInRange(TheDate) Then
          Begin
            FNavigationTimer := SetTimer(Handle, AutoNavTimerBackward,
              AutoNavInterval, Nil);
            SelectDate(TheDate);
          End;
        End;
      ChtArrowRight:
        Begin
          TheDate := GetAdjacentPageDate(False);
          If IsDateInRange(TheDate) Then
          Begin
            FNavigationTimer := SetTimer(Handle, AutoNavTimerForward,
              AutoNavInterval, Nil);
            SelectDate(TheDate);
          End;
        End;
      ChtHeaderText:
        Case CurrentView Of
          VmDay:
            CurrentView := VmMonth;
          VmMonth:
            CurrentView := VmYear;
        End;
      ChtFooter:
        Begin
          CurrentView := VmDay;
          Select(Calendar.Today);
          InvalidateSections([CsFooter]);
        End;
      ChtDay:
        If GetDateAtCell(TheCell, TheDate) Then
        Begin
          SelectDate(TheDate, Shift);
          FDraggingCell := TheCell;
        End;
      ChtMonth:
        If GetMonthAtCell(TheCell, TheMonth) Then
        Begin
          TheDay := Min(Day, Calendar.DaysInMonth(Era, Year, TheMonth));
          Select(Calendar.EncodeDate(Era, Year, TheMonth, TheDay));
          FDraggingCell := TheCell;
        End;
      ChtYear:
        If GetYearAtCell(TheCell, TheYear) Then
        Begin
          FDraggingOffPage := Not IsYearOnPage(TheYear);
          TheMonth := Min(Month, Calendar.MonthsInYear(Era, TheYear));
          TheDay := Min(Day, Calendar.DaysInMonth(Era, TheYear, TheMonth));
          Select(Calendar.EncodeDate(Era, TheYear, TheMonth, TheDay));
          FDraggingCell := TheCell;
        End;
    End;
    FDraggingShiftState := Shift + [SsShift];
  End;
  Inherited MouseDown(Button, Shift, X, Y);
End;

Procedure TCustomIntlMonthCalendar.MouseMove(Shift: TShiftState; X, Y: Integer);
Var
  TheDate: TDate;
  TheYear, TheMonth, TheDay: Integer;
  TheCell: Integer;
  TheHitTest: TIntlCalHitTest;
Begin
  If NavigationTimer <> 0 Then
    Exit;
  TheHitTest := GetHitTestInfoAt(X, Y, TheCell);
  If (TheHitTest <> HotInfo) Or (TheCell <> HotCell) Then
  Begin
    InvalidateHotLight;
    FHotInfo := TheHitTest;
    FHotCell := TheCell;
    InvalidateHotLight;
  End;
  If (DraggingCell >= 0) And (DraggingCell <> TheCell) And (TheCell >= 0) Then
  Begin
    Case HotInfo Of
      ChtDay:
        If GetDateAtCell(TheCell, TheDate) Then
        Begin
          FDraggingCell := TheCell;
          SelectDate(TheDate, FDraggingShiftState);
        End;
      ChtMonth:
        If GetMonthAtCell(TheCell, TheMonth) Then
        Begin
          FDraggingCell := TheCell;
          TheDay := Min(Day, Calendar.DaysInMonth(Era, Year, TheMonth));
          Select(Calendar.EncodeDate(Era, Year, TheMonth, TheDay));
        End;
      ChtYear:
        If GetYearAtCell(TheCell, TheYear) Then
        Begin
          FDraggingCell := TheCell;
          FDraggingOffPage := Not IsYearOnPage(TheYear);
          TheMonth := Min(Month, Calendar.MonthsInYear(Era, TheYear));
          TheDay := Min(Day, Calendar.DaysInMonth(Era, TheYear, TheMonth));
          Select(Calendar.EncodeDate(Era, TheYear, TheMonth, TheDay));
        End;
    End;
  End;
  Inherited MouseMove(Shift, X, Y);
End;

Procedure TCustomIntlMonthCalendar.MouseUp(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
Var
  TheMonth, TheYear: Integer;
Begin
  If NavigationTimer <> 0 Then
  Begin
    KillTimer(Handle, NavigationTimer);
    FNavigationTimer := 0;
    InvalidateHotLight;
  End;
  If DraggingCell >= 0 Then
  Begin
    FDraggingCell := -1;
    If (Button = MbLeft) And (HotCell >= 0) Then
      Case HotInfo Of
        ChtMonth:
          If GetMonthAtCell(HotCell, TheMonth) Then
            CurrentView := VmDay;
        ChtYear:
          If Not DraggingOffPage And GetYearAtCell(HotCell, TheYear) Then
            CurrentView := VmMonth;
      End;
  End;
  Inherited MouseUp(Button, Shift, X, Y);
End;

Function TCustomIntlMonthCalendar.DoMouseWheelDown(Shift: TShiftState;
  MousePos: TPoint): Boolean;
Begin
  Result := Inherited DoMouseWheelDown(Shift, MousePos);
  If Not Result Then
  Begin
    Case CurrentView Of
      VmDay:
        Select(Calendar.NextDay(Date), Shift);
      VmMonth:
        Select(Calendar.NextMonth(Date));
      VmYear:
        Select(Calendar.NextYear(Date));
    End;
    Result := True;
  End;
End;

Function TCustomIntlMonthCalendar.DoMouseWheelUp(Shift: TShiftState;
  MousePos: TPoint): Boolean;
Begin
  Result := Inherited DoMouseWheelUp(Shift, MousePos);
  If Not Result Then
  Begin
    Case CurrentView Of
      VmDay:
        Select(Calendar.PrevDay(Date), Shift);
      VmMonth:
        Select(Calendar.PrevMonth(Date));
      VmYear:
        Select(Calendar.PrevYear(Date));
    End;
    Result := True;
  End;
End;

Procedure TCustomIntlMonthCalendar.KeyDown(Var Key: Word; Shift: TShiftState);
Begin
  Inherited KeyDown(Key, Shift);
  If (SelectionStyle <> SsFree) Or (SsShift In Shift) Or
    Not(SsCtrl In Shift) Then
    Include(Shift, SsLeft);
  Case Key Of
    VK_RIGHT:
      Begin
        Case CurrentView Of
          VmDay:
            If IsRightToLeft Then
              SelectDate(Calendar.PrevDay(Date), Shift)
            Else
              SelectDate(Calendar.NextDay(Date), Shift);
          VmMonth:
            If IsRightToLeft Then
              Select(Calendar.PrevMonth(Date))
            Else
              Select(Calendar.NextMonth(Date));
          VmYear:
            Select(Calendar.NextYear(Date));
        End;
        Key := 0;
      End;
    VK_LEFT:
      Begin
        Case CurrentView Of
          VmDay:
            If IsRightToLeft Then
              SelectDate(Calendar.NextDay(Date), Shift)
            Else
              SelectDate(Calendar.PrevDay(Date), Shift);
          VmMonth:
            If IsRightToLeft Then
              Select(Calendar.NextMonth(Date))
            Else
              Select(Calendar.PrevMonth(Date));
          VmYear:
            Select(Calendar.PrevYear(Date));
        End;
        Key := 0;
      End;
    VK_DOWN:
      Begin
        Case CurrentView Of
          VmDay:
            SelectDate(Calendar.NextDay(Date, ColCount), Shift);
          VmMonth:
            Select(Calendar.NextMonth(Date, ColCount));
          VmYear:
            Select(Calendar.NextYear(Date, ColCount));
        End;
        Key := 0;
      End;
    VK_UP:
      Begin
        Case CurrentView Of
          VmDay:
            SelectDate(Calendar.PrevDay(Date, ColCount), Shift);
          VmMonth:
            Select(Calendar.PrevMonth(Date, ColCount));
          VmYear:
            Select(Calendar.PrevYear(Date, ColCount));
        End;
        Key := 0;
      End;
    VK_NEXT:
      Begin
        Case CurrentView Of
          VmDay:
            SelectDate(Calendar.NextMonth(Date), Shift);
          VmMonth:
            Select(Calendar.NextYear(Date));
          VmYear:
            Select(Calendar.NextYear(Date, 10));
        End;
        Key := 0;
      End;
    VK_PRIOR:
      Begin
        Case CurrentView Of
          VmDay:
            SelectDate(Calendar.PrevMonth(Date), Shift);
          VmMonth:
            Select(Calendar.PrevYear(Date));
          VmYear:
            Select(Calendar.PrevYear(Date, 10));
        End;
        Key := 0;
      End;
    VK_HOME:
      Begin
        Case CurrentView Of
          VmDay:
            SelectDate(MonthFirstDate, Shift);
          VmMonth:
            Select(Calendar.StartOfYear(Date));
          VmYear:
            Select(Calendar.StartOfYear(Era, FirstYearOnPage));
        End;
        Key := 0;
      End;
    VK_END:
      Begin
        Case CurrentView Of
          VmDay:
            SelectDate(MonthLastDate, Shift);
          VmMonth:
            Select(Calendar.EndOfYear(Date));
          VmYear:
            Select(Calendar.EndOfYear(Era, LastYearOnPage));
        End;
        Key := 0;
      End;
    VK_SPACE:
      Begin
        If CurrentView = VmDay Then
          SelectDate(Date, [SsLeft, SsCtrl])
        Else If CurrentView = VmYear Then
          CurrentView := VmMonth
        Else
          CurrentView := VmDay;
        Key := 0;
      End;
    VK_BACK:
      Begin
        If CurrentView = VmDay Then
          CurrentView := VmMonth
        Else If CurrentView = VmMonth Then
          CurrentView := VmYear;
        Key := 0;
      End;
  End;
End;

Procedure TCustomIntlMonthCalendar.DoEnter;
Begin
  Inherited DoEnter;
  InvalidateSelection;
End;

Procedure TCustomIntlMonthCalendar.DoExit;
Begin
  Inherited DoExit;
  InvalidateSelection;
End;

Function TCustomIntlMonthCalendar.IsThemed: Boolean;
Begin
  Result := (Win32MajorVersion > 5) And ThemeControl(Self) And (Theme <> 0);
End;

Function TCustomIntlMonthCalendar.IsYearOnPage(AYear: Integer): Boolean;
Begin
  Result := (AYear = Year) Or
    ((Calendar.EraOf(Calendar.StartOfYear(Era, AYear)) = Era) And
    (AYear Div 10 = Year Div 10));
End;

Procedure TCustomIntlMonthCalendar.AnimateBegin;
Begin
  If Assigned(FSavedScreen) Then
    FreeAndNil(FSavedScreen);
  If Animate And Showing Then
  Begin
    FSavedScreen := TakeSnapshot;
    FSavedView := CurrentView;
    FSavedDate := Date;
    FSavedCellRect := GetCellRect(FocusedCell);
  End;
End;

Procedure TCustomIntlMonthCalendar.AnimateEnd;
Begin
  If Assigned(FSavedScreen) Then
    FreeAndNil(FSavedScreen);
End;

Procedure TCustomIntlMonthCalendar.AnimatePlay;
Var
  NewScreen: TBitmap;
  Rect: TRect;
  DC: HDC;
Begin
  If Animate And Showing And Assigned(FSavedScreen) Then
  Begin
    NewScreen := TakeSnapshot;
    Try
      DC := GetDC(Handle);
      Try
        Rect := FMetrics.Bounds;
        If ShowHeader Then
        Begin
          BitBlt(DC, Rect.Left, Rect.Top, Rect.Right - Rect.Left,
            FMetrics.HeaderHeight, NewScreen.Canvas.Handle, 0, 0, SRCCOPY);
          Inc(Rect.Top, FMetrics.HeaderHeight);
        End;
        If ShowFooter Then
          Dec(Rect.Bottom, FMetrics.FooterHeight);
        If CurrentView = FSavedView Then
        Begin
          SwapScreenByScroll(DC, Rect, FMetrics.Bounds.Left,
            FMetrics.Bounds.Top, FSavedScreen, NewScreen, Date < FSavedDate,
            AnimateDuration);
        End
        Else
        Begin
          SwapScreenByZoom(DC, Rect, FMetrics.Bounds.Left, FMetrics.Bounds.Top,
            FSavedScreen, NewScreen, FSavedCellRect, GetCellRect(FocusedCell),
            AnimateDuration);
        End;
      Finally
        ReleaseDC(Handle, DC);
      End;
    Finally
      NewScreen.Free;
    End;
  End;
End;

Function TCustomIntlMonthCalendar.TakeSnapshot: TBitmap;
Begin
  Result := TBitmap.Create;
  With FMetrics.Bounds Do
  Begin
    Result.Width := Right - Left;
    Result.Height := Bottom - Top;
    SetWindowOrgEx(Result.Canvas.Handle, -Left, -Top, Nil);
  End;
  DrawCalendar(Result.Canvas);
  SetWindowOrgEx(Result.Canvas.Handle, 0, 0, Nil);
End;

Procedure TCustomIntlMonthCalendar.RenderBackground(Canvas: TCanvas;
  Var Rect: TRect; State: TIntlCalDrawStates);
Const
  CP_CALITEM = LVP_LISTITEM;
Var
  StateId: Integer;
  ForeColor: Cardinal;
Begin
  If ((CdsSelected In State) Or (CdsHotLight In State)) And IsThemed Then
  Begin
    If CdsGrayed In State Then
      StateId := LIS_DISABLED
    Else If CdsSelected In State Then
    Begin
      If Focused Then
        If CdsHotLight In State Then
          StateId := LISS_HOTSELECTED
        Else
          StateId := LIS_SELECTED
      Else
        StateId := LIS_SELECTEDNOTFOCUS
    End
    Else If CdsHotLight In State Then
      StateId := LIS_HOT
    Else
      StateId := LIS_NORMAL;
    If CdsGrayed In State Then
      ForeColor := ColorToRGB(ClGrayText)
    Else If CdsHeader In State Then
      ForeColor := ColorToRGB(FontHeader.Color)
    Else If CdsFixedRow In State Then
      ForeColor := ColorToRGB(FontDayOfWeek.Color)
    Else If CdsFixedCol In State Then
      ForeColor := ColorToRGB(FontWeekNumbers.Color)
    Else
      ForeColor := ColorToRGB(Font.Color);
    Canvas.Brush.Color := Color;
    Canvas.FillRect(Rect);
    DrawThemeBackground(Theme, Canvas.Handle, CP_CALITEM, StateId, Rect, @Rect);
    GetThemeColor(Theme, CP_CALITEM, StateId, TMT_TEXTCOLOR, ForeColor);
    Canvas.Font.Color := ForeColor;
  End
  Else
  Begin
    If CdsGrayed In State Then
      Canvas.Font.Color := ClGrayText
    Else If CdsSelected In State Then
      If CdsFocused In State Then
        Canvas.Font.Color := ClHighlightText
      Else
        Canvas.Font.Color := ClBtnText
    Else If CdsHotLight In State Then
      Canvas.Font.Color := ClHotLight
    Else If CdsHeader In State Then
      Canvas.Font.Color := FontHeader.Color
    Else If CdsFooter In State Then
      Canvas.Font.Color := FontFooter.Color
    Else If CdsFixedRow In State Then
      Canvas.Font.Color := FontDayOfWeek.Color
    Else If CdsFixedCol In State Then
      Canvas.Font.Color := FontWeekNumbers.Color
    Else
      Canvas.Font.Color := Font.Color;
    If CdsSelected In State Then
      If CdsFocused In State Then
        Canvas.Brush.Color := ClHighlight
      Else
        Canvas.Brush.Color := ClBtnFace
    Else
      Canvas.Brush.Color := Color;
    Canvas.FillRect(Rect);
  End;
  InflateRect(Rect, -1, -1);
  SetBkMode(Canvas.Handle, TRANSPARENT);
End;

Procedure TCustomIntlMonthCalendar.RenderText(Canvas: TCanvas; Var Rect: TRect;
  State: TIntlCalDrawStates; Const Text: String; Alignment: TAlignment);
Const
  TextAlignments: Array [TAlignment] Of Cardinal = (DT_LEFT, DT_RIGHT,
    DT_CENTER);
Var
  DrawFlags: Cardinal;
Begin
  DrawFlags := DrawTextBiDiModeFlags(TextAlignments[Alignment] Or DT_VCENTER Or
    DT_NOPREFIX Or DT_SINGLELINE);
  Windows.DrawText(Canvas.Handle, PChar(Text), Length(Text), Rect, DrawFlags);
End;

Procedure TCustomIntlMonthCalendar.RenderArrow(Canvas: TCanvas; Var Rect: TRect;
  State: TIntlCalDrawStates; Backward: Boolean);
Var
  Points: Array [0 .. 2] Of TPoint;
  ArrowWidth: Integer;
  W, H: Integer;
Begin
  W := Rect.Right - Rect.Left;
  H := Rect.Bottom - Rect.Top;
  If W < H Then
    ArrowWidth := W Div 6
  Else
    ArrowWidth := H Div 6;
  If Backward Then
  Begin
    Points[0].X := Rect.Left + (W Div 2 - ArrowWidth + 1);
    Points[1].X := Points[0].X + ArrowWidth;
  End
  Else
  Begin
    Points[0].X := Rect.Right - (W Div 2 - ArrowWidth + 1);
    Points[1].X := Points[0].X - ArrowWidth;
  End;
  Points[2].X := Points[1].X;
  Points[0].Y := (Rect.Top + Rect.Bottom) Div 2;
  Points[1].Y := Points[0].Y - ArrowWidth;
  Points[2].Y := Points[0].Y + ArrowWidth;
  Canvas.Brush.Color := Canvas.Font.Color;
  Canvas.Pen.Color := Canvas.Font.Color;
  Canvas.Polygon(Points);
End;

Procedure TCustomIntlMonthCalendar.RenderTodayMarker(Canvas: TCanvas;
  Var Rect: TRect);
Begin
  Canvas.Brush.Style := BsClear;
  Canvas.Pen.Width := 0;
  Canvas.Pen.Color := ColorToday;
  // Canvas.RoundRect(Rect.Left + 1, Rect.Top + 1, Rect.Right - 1, Rect.Bottom - 1, 4, 4);
  Canvas.RoundRect(Rect.Left + 1, Rect.Top + 1, Rect.Right - 1,
    Rect.Bottom - 1, 1, 1);
  Canvas.Brush.Style := BsSolid;
End;

Procedure TCustomIntlMonthCalendar.RenderEventMarker(Canvas: TCanvas;
  Var Rect: TRect);
Var
  Wi, Hi: Integer;
Begin
  Canvas.Brush.Style := BsClear;
  Canvas.Brush.Color := ColorToday;
  Canvas.Pen.Color := ColorToday;
  Wi := (Rect.Right - Rect.Left) Div 4;
  Hi := (Rect.Bottom - Rect.Top) Div 3;
  Wi := Min(Hi, Wi);
  Canvas.Polygon([Point(Rect.Right - Wi, Rect.Bottom - 2), Point(Rect.Right - 2,
    Rect.Bottom - 2), Point(Rect.Right - 2, Rect.Bottom - Wi)]);
  Frame3d(Canvas, Rect, ClGray, ClGray, 1);
End;

Procedure TCustomIntlMonthCalendar.DrawCalendar(Canvas: TCanvas);
Begin
  If RectVisible(Canvas.Handle, FMetrics.Bounds) Then
  Begin
    If ShowHeader Then
      DrawHeader(Canvas);
    Case CurrentView Of
      VmDay:
        Begin
          If WeekNumbers Then
          Begin
            DrawWeekSpace(Canvas);
            DrawWeekNumbers(Canvas);
          End;
          DrawDaysOfWeek(Canvas);
          DrawDays(Canvas);
        End;
      VmMonth:
        DrawMonths(Canvas);
      VmYear:
        DrawYears(Canvas);
    End;
    If ShowFooter Then
      DrawFooter(Canvas);
  End;
End;

Procedure TCustomIntlMonthCalendar.DrawHeader(Canvas: TCanvas);
Var
  Rect: TRect;
  State: TIntlCalDrawStates;
  HeaderText: String;
Begin
  Rect := GetSectionRect(CsHeader);
  If RectVisible(Canvas.Handle, Rect) Then
  Begin
    Canvas.Font := FontHeader;
    // Left Arrow
    Rect := GetSectionRect(CsArrowLeft);
    If RectVisible(Canvas.Handle, Rect) Then
    Begin
      State := [CdsHeader];
      If Not IsDateInRange(GetAdjacentPageDate(True)) Then
        Include(State, CdsGrayed)
      Else If HotInfo = ChtArrowLeft Then
      Begin
        Include(State, CdsHotLight);
        If NavigationTimer <> 0 Then
          Include(State, CdsSelected);
      End;
      RenderBackground(Canvas, Rect, State);
      RenderArrow(Canvas, Rect, State, True);
    End;
    // Right Arrow
    Rect := GetSectionRect(CsArrowRight);
    If RectVisible(Canvas.Handle, Rect) Then
    Begin
      State := [CdsHeader];
      If Not IsDateInRange(GetAdjacentPageDate(False)) Then
        Include(State, CdsGrayed)
      Else If HotInfo = ChtArrowRight Then
      Begin
        Include(State, CdsHotLight);
        If NavigationTimer <> 0 Then
          Include(State, CdsSelected);
      End;
      RenderBackground(Canvas, Rect, State);
      RenderArrow(Canvas, Rect, State, False);
    End;
    // Header text
    Rect := GetSectionRect(CsHeaderText);
    If RectVisible(Canvas.Handle, Rect) Then
    Begin
      State := [CdsHeader];
      If (HotInfo = ChtHeaderText) And (CurrentView <> VmYear) Then
        Include(State, CdsHotLight);
      RenderBackground(Canvas, Rect, State);
      HeaderText := GetHeaderText(CurrentView);
      If DoCustomDrawHeaderText(Canvas, Rect, State, HeaderText) Then
        RenderText(Canvas, Rect, State, HeaderText, TaCenter);
    End;
  End;
End;

Procedure TCustomIntlMonthCalendar.DrawFooter(Canvas: TCanvas);
Const
  Spacing = 2;
Var
  Rect, R: TRect;
  State: TIntlCalDrawStates;
  FooterText: String;
Begin
  Rect := GetSectionRect(CsFooter);
  If RectVisible(Canvas.Handle, Rect) Then
  Begin
    Canvas.Pen.Width := 0;
    Canvas.Pen.Color := Color;
    Canvas.MoveTo(Rect.Left, Rect.Top);
    Canvas.LineTo(Rect.Right, Rect.Top);
    Inc(Rect.Top);
    Canvas.MoveTo(Rect.Left, Rect.Top);
    Canvas.LineTo(Rect.Right, Rect.Top);
    Inc(Rect.Top);
    Canvas.Font := FontFooter;
    State := [CdsFooter];
    If (HotInfo = ChtFooter) And Not Calendar.IsToday(Date) Then
      Include(State, CdsHotLight);
    RenderBackground(Canvas, Rect, State);
    FooterText := GetFooterText(FMetrics.LongFormat);
    If DoCustomDrawFooterText(Canvas, Rect, State, FooterText) Then
    Begin
      InflateRect(Rect, -Spacing, -Spacing);
      R.Top := Rect.Top;
      R.Bottom := Rect.Bottom;
      If IsRightToLeft Then
      Begin
        R.Right := Rect.Right;
        R.Left := R.Right - FMetrics.FooterHeight;
        Dec(Rect.Right, FMetrics.FooterHeight + Spacing);
      End
      Else
      Begin
        R.Left := Rect.Left;
        R.Right := R.Left + FMetrics.FooterHeight;
        Inc(Rect.Left, FMetrics.FooterHeight + Spacing);
      End;
      RenderText(Canvas, Rect, State, FooterText, TaLeftJustify);
      RenderTodayMarker(Canvas, R);
    End;
  End;
End;

Procedure TCustomIntlMonthCalendar.DrawWeekSpace(Canvas: TCanvas);
Var
  Rect: TRect;
Begin
  Rect := GetSectionRect(CsWeekSpace);
  If RectVisible(Canvas.Handle, Rect) Then
    RenderBackground(Canvas, Rect, [CdsFixedRow, CdsFixedCol, CdsBlank]);
End;

Procedure TCustomIntlMonthCalendar.DrawDaysOfWeek(Canvas: TCanvas);
Var
  Rect: TRect;
  Col: Integer;
  DayOfWeek: TDayOfWeek;
  State: TIntlCalDrawStates;
Begin
  Rect := GetSectionRect(CsDaysOfWeek);
  If RectVisible(Canvas.Handle, Rect) Then
  Begin
    Canvas.Pen.Width := 0;
    Dec(Rect.Bottom);
    Canvas.Pen.Color := Color;
    Canvas.MoveTo(Rect.Left, Rect.Bottom);
    Canvas.LineTo(Rect.Right, Rect.Bottom);
    Dec(Rect.Bottom);
    Canvas.Pen.Color := ColorDivider;
    Canvas.MoveTo(Rect.Left, Rect.Bottom);
    Canvas.LineTo(Rect.Right, Rect.Bottom);
    Canvas.Font := FontDayOfWeek;
    For Col := 0 To ColCount - 1 Do
    Begin
      DayOfWeek := TDayOfWeek(((Ord(Calendar.Settings.FirstDayOfWeek) + Col - 1)
        Mod Calendar.DaysPerWeek) + 1);
      Rect := GetCellRect(-1, Col);
      If RectVisible(Canvas.Handle, Rect) Then
      Begin
        State := [CdsFixedRow];
        If Calendar.IsToday(DayOfWeek) And Calendar.IsThisMonth(Era, Year,
          Month) Then
          Include(State, CdsToday);
        RenderBackground(Canvas, Rect, State);
        If DoCustomDrawDayOfWeek(Canvas, Rect, State, DayOfWeek) Then
          RenderText(Canvas, Rect, State, GetDayName(DayOfWeek,
            FMetrics.NameKind), TaCenter);
      End;
    End;
  End;
End;

Procedure TCustomIntlMonthCalendar.DrawWeekNumbers(Canvas: TCanvas);
Var
  Rect: TRect;
  Row, TheWeek: Integer;
  StartRange, EndRange: Integer;
  WeekStartDate: TDateTime;
  State: TIntlCalDrawStates;
Begin
  Rect := GetSectionRect(CsWeekNumbers);
  If RectVisible(Canvas.Handle, Rect) Then
  Begin
    Canvas.Pen.Width := 0;
    If IsRightToLeft Then
    Begin
      Canvas.Pen.Color := Color;
      Canvas.MoveTo(Rect.Left, Rect.Top);
      Canvas.LineTo(Rect.Left, Rect.Bottom);
      Inc(Rect.Left);
      Canvas.Pen.Color := ColorDivider;
      Canvas.MoveTo(Rect.Left, Rect.Top);
      Canvas.LineTo(Rect.Left, Rect.Bottom);
      Inc(Rect.Left);
    End
    Else
    Begin
      Dec(Rect.Right);
      Canvas.Pen.Color := Color;
      Canvas.MoveTo(Rect.Right, Rect.Top);
      Canvas.LineTo(Rect.Right, Rect.Bottom);
      Dec(Rect.Right);
      Canvas.Pen.Color := ColorDivider;
      Canvas.MoveTo(Rect.Right, Rect.Top);
      Canvas.LineTo(Rect.Right, Rect.Bottom);
    End;
    StartRange := MonthStartCell Div ColCount;
    EndRange := (MonthStartCell + DaysInMonth - 1) Div ColCount;
    Canvas.Font := FontWeekNumbers;
    For Row := 0 To RowCount - 1 Do
    Begin
      Rect := GetCellRect(Row, -1);
      If RectVisible(Canvas.Handle, Rect) Then
      Begin
        WeekStartDate := Calendar.NextWeek(FirstCellDate, Row);
        State := [CdsFixedCol];
        If (Row = (MonthStartCell + Day - 1) Div ColCount) Or
          IsDateInRange(WeekStartDate) Or
          IsDateInRange(Calendar.EndOfWeek(WeekStartDate)) Then
        Begin
          If (Row < StartRange) Or (Row > EndRange) Then
            Include(State, CdsGrayed);
          If Calendar.IsThisWeek(WeekStartDate) Then
            Include(State, CdsToday);
          If Row = StartRange Then
            TheWeek := Calendar.WeekOfYear(MonthFirstDate)
          Else
            TheWeek := Calendar.WeekOfYear(WeekStartDate);
        End
        Else
        Begin
          Include(State, CdsBlank);
          TheWeek := 0;
        End;
        RenderBackground(Canvas, Rect, State);
        If Not(CdsBlank In State) And DoCustomDrawWeekNumber(Canvas, Rect,
          State, TheWeek) Then
          RenderText(Canvas, Rect, State, GetWeekText(TheWeek), TaCenter);
      End;
    End;
  End;
End;

Procedure TCustomIntlMonthCalendar.DrawDays(Canvas: TCanvas);
Var
  Rect: TRect;
  Cell: Integer;
  TheDate: TDate;
  TheDay: Integer;
  State: TIntlCalDrawStates;
Begin
  Rect := GetSectionRect(CsYMD);
  If RectVisible(Canvas.Handle, Rect) Then
  Begin
    Canvas.Font := Font;
    For Cell := 0 To CellCount - 1 Do
    Begin
      Rect := GetCellRect(Cell);
      If RectVisible(Canvas.Handle, Rect) Then
      Begin
        TheDate := FirstCellDate + Cell;
        State := [];
        If Cell < MonthStartCell Then
        Begin
          TheDay := FirstCellDayNumber + Cell;
          Include(State, CdsGrayed);
        End
        Else If Cell >= MonthStartCell + DaysInMonth Then
        Begin
          TheDay := Cell - MonthStartCell - DaysInMonth + 1;
          Include(State, CdsGrayed);
        End
        Else
          TheDay := Cell - MonthStartCell + 1;
        If IsDateInRange(TheDate) Then
        Begin
          If IsSelected(TheDate) Then
            Include(State, CdsSelected);
          If (Cell = FocusedCell) And Focused Then
            Include(State, CdsFocused);
          If Cell = HotCell Then
            Include(State, CdsHotLight);
          If Calendar.IsToday(TheDate) Then
            Include(State, CdsToday);
        End
        Else
          Include(State, CdsBlank);
        RenderBackground(Canvas, Rect, State);
        If Not(CdsBlank In State) And DoCustomDrawDay(Canvas, Rect, State,
          TheDate, TheDay) Then
        Begin
          RenderText(Canvas, Rect, State, GetDayText(TheDay), TaCenter);
          If CdsToday In State Then
            RenderTodayMarker(Canvas, Rect);
          If (FShowEvents) And (FEventItems.FindItem(TheDate) >= 0) Then
            RenderEventMarker(Canvas, Rect);
          If CdsFocused In State Then
            Canvas.DrawFocusRect(Rect);
        End;
      End;
    End;
  End;
End;

Procedure TCustomIntlMonthCalendar.DrawMonths(Canvas: TCanvas);
Var
  Rect: TRect;
  Cell: Integer;
  TheMonth: Integer;
  State: TIntlCalDrawStates;
Begin
  Rect := GetSectionRect(CsYMD);
  If RectVisible(Canvas.Handle, Rect) Then
  Begin
    Canvas.Font := Font;
    For Cell := 0 To CellCount - 1 Do
    Begin
      Rect := GetCellRect(Cell);
      If RectVisible(Canvas.Handle, Rect) Then
      Begin
        TheMonth := Cell + 1;
        State := [];
        If IsMonthInRange(Era, Year, TheMonth) Then
        Begin
          If Cell = FocusedCell Then
          Begin
            Include(State, CdsSelected);
            If Focused Then
              Include(State, CdsFocused);
          End;
          If Cell = HotCell Then
            Include(State, CdsHotLight);
          If Calendar.IsThisMonth(Era, Year, TheMonth) Then
            Include(State, CdsToday);
        End
        Else
          Include(State, CdsBlank);
        RenderBackground(Canvas, Rect, State);
        If Not(CdsBlank In State) And DoCustomDrawMonth(Canvas, Rect, State,
          TheMonth) Then
        Begin
          RenderText(Canvas, Rect, State, GetMonthName(TheMonth,
            FMetrics.NameKind), TaCenter);
          If ShowEvents Then
          Begin
            Canvas.Font.Color := ClRed;
            Canvas.Font.Style := [FsBold];
            RenderText(Canvas, Rect, State, GetMonthEvents(TheMonth),
              TaRightJustify);
            Canvas.Font := Font;
          End;
          If CdsFocused In State Then
            Canvas.DrawFocusRect(Rect);
        End;
      End;
    End;
  End;
End;

Function TCustomIntlMonthCalendar.GetMonthEvents(Month: Integer): String;
Var
  I: Integer;
  Y, M, D, Rs: Integer;
Begin
  Rs := 0;
  For I := 0 To EventItems.Count - 1 Do
  Begin
    FCalendar.DecodeDate(EventItems.Items[I].Date, Y, M, D);
    If (Y = Year) And (M = Month) Then
      Inc(Rs);
  End;
  If Rs = 0 Then
    Result := ''
  Else
    Result := ' ' + IntToStr(Rs) + ' ';
End;

Procedure TCustomIntlMonthCalendar.DrawYears(Canvas: TCanvas);
Var
  Rect: TRect;
  Cell: Integer;
  TheYear: Integer;
  State: TIntlCalDrawStates;
Begin
  Rect := GetSectionRect(CsYMD);
  If RectVisible(Canvas.Handle, Rect) Then
  Begin
    Canvas.Font := Font;
    For Cell := 0 To CellCount - 1 Do
    Begin
      Rect := GetCellRect(Cell);
      If RectVisible(Canvas.Handle, Rect) Then
      Begin
        TheYear := Calendar.OffsetYear(Era, FirstCellYear, Cell);
        State := [];
        If IsYearInRange(Era, TheYear) Then
        Begin
          If Cell = FocusedCell Then
          Begin
            Include(State, CdsSelected);
            If Focused Then
              Include(State, CdsFocused);
          End
          Else If Not IsYearOnPage(TheYear) Then
            Include(State, CdsGrayed);
          If Cell = HotCell Then
            Include(State, CdsHotLight);
          If Calendar.IsThisYear(Era, TheYear) Then
            Include(State, CdsToday);
        End
        Else
          Include(State, CdsBlank);
        RenderBackground(Canvas, Rect, State);
        If Not(CdsBlank In State) And DoCustomDrawYear(Canvas, Rect, State,
          TheYear) Then
        Begin
          RenderText(Canvas, Rect, State, GetYearText(TheYear), TaCenter);
          If ShowEvents Then
          Begin
            Canvas.Font.Color := ClRed;
            Canvas.Font.Style := [FsBold];
            RenderText(Canvas, Rect, State, GetYearEvents(TheYear),
              TaRightJustify);
            Canvas.Font := Font;
          End;
          If CdsFocused In State Then
            Canvas.DrawFocusRect(Rect);
        End;
      End;
    End;
  End;
End;

Function TCustomIntlMonthCalendar.GetYearEvents(Year: Integer): String;
Var
  I: Integer;
  Y, M, D, Rs: Integer;
Begin
  Rs := 0;
  For I := 0 To EventItems.Count - 1 Do
  Begin
    FCalendar.DecodeDate(EventItems.Items[I].Date, Y, M, D);
    If (Y = Year) Then
      Inc(Rs);
  End;
  If Rs = 0 Then
    Result := ''
  Else
    Result := ' ' + IntToStr(Rs) + ' ';
End;

Procedure TCustomIntlMonthCalendar.InvalidateSections
  (Sections: TIntlCalSections);
Var
  Section: TIntlCalSection;
Begin
  If WindowHandle <> 0 Then
  Begin
    If (CurrentView <> VmDay) Or Not WeekNumbers Then
    Begin
      Exclude(Sections, CsWeekSpace);
      Exclude(Sections, CsWeekNumbers);
      If CurrentView <> VmDay Then
        Exclude(Sections, CsDaysOfWeek);
    End;
    If Not ShowHeader Then
    Begin
      Exclude(Sections, CsHeader);
      Exclude(Sections, CsHeaderText);
      Exclude(Sections, CsArrowLeft);
      Exclude(Sections, CsArrowRight);
    End;
    If Not ShowFooter Then
      Exclude(Sections, CsFooter);
    For Section := Low(TIntlCalSection) To High(TIntlCalSection) Do
      If Section In Sections Then
        InvalidateRect(WindowHandle, GetSectionRect(Section), False);
  End;
End;

Procedure TCustomIntlMonthCalendar.InvalidateDateDependentSections;
Begin
  InvalidateSections([CsHeader, CsWeekNumbers, CsYMD, CsFooter]);
End;

Procedure TCustomIntlMonthCalendar.InvalidateCell(Index: Integer);
Begin
  If (WindowHandle <> 0) And (Index >= 0) And (Index < CellCount) Then
    InvalidateRect(WindowHandle, GetCellRect(Index), False);
End;

Procedure TCustomIntlMonthCalendar.InvalidateDate(Const ADate: TDate);
Var
  TheYear: Integer;
Begin
  If WindowHandle <> 0 Then
    Case CurrentView Of
      VmDay:
        InvalidateCell(Trunc(ADate) - Trunc(FirstCellDate));
      VmMonth:
        InvalidateCell(Calendar.MonthOfYear(ADate) - 1);
      VmYear:
        Begin
          TheYear := Calendar.ConvertYear(Calendar.YearOf(ADate),
            Calendar.DefaultEra, Era);
          InvalidateCell(TheYear - FirstCellYear);
        End;
    End;
End;

Procedure TCustomIntlMonthCalendar.InvalidateDates(ADates: TDateTimeList);
Var
  I, Index: Integer;
Begin
  If (WindowHandle <> 0) And (CurrentView = VmDay) And (ADates.Count <> 0) Then
  Begin
    ADates.Find(FirstCellDate, I);
    For I := I To FSelection.Count - 1 Do
    Begin
      Index := Trunc(ADates[I]) - Trunc(FirstCellDate);
      If Index >= CellCount Then
        Exit;
      InvalidateCell(Index);
    End;
  End;
End;

Procedure TCustomIntlMonthCalendar.InvalidateDateRange(Const ADate1,
  ADate2: TDate);
Var
  Index: Integer;
  IStart, IEnd: Integer;
Begin
  If (WindowHandle <> 0) And (CurrentView = VmDay) Then
  Begin
    IStart := Trunc(ADate1) - Trunc(FirstCellDate);
    IEnd := Trunc(ADate2) - Trunc(FirstCellDate);
    If IStart > IEnd Then
    Begin
      Index := IStart;
      IStart := IEnd;
      IEnd := Index;
    End;
    If IStart < 0 Then
      IStart := 0;
    If IEnd >= CellCount Then
      IEnd := CellCount - 1;
    For Index := IStart To IEnd Do
      InvalidateCell(Index);
  End;
End;

Procedure TCustomIntlMonthCalendar.InvalidateSelection;
Begin
  Case CurrentView Of
    VmDay:
      Begin
        Case SelectionStyle Of
          SsSingle:
            InvalidateCell(MonthStartCell + Day - 1);
          SsRange:
            InvalidateDateRange(Date, FSelectionAnchor);
          SsFree:
            Begin
              InvalidateDates(FSelection);
              If Not FSelection.Exists(Date) Then
                InvalidateCell(MonthStartCell + Day - 1);
            End;
        End;
      End;
    VmMonth:
      InvalidateCell(Month - 1);
    VmYear:
      InvalidateCell(Year - FirstCellYear);
  End;
End;

Procedure TCustomIntlMonthCalendar.InvalidateHotLight;
Begin
  Case HotInfo Of
    ChtYear, ChtMonth, ChtDay:
      InvalidateCell(HotCell);
    ChtHeaderText:
      InvalidateSections([CsHeaderText]);
    ChtArrowLeft:
      InvalidateSections([CsArrowLeft]);
    ChtArrowRight:
      InvalidateSections([CsArrowRight]);
    ChtFooter:
      InvalidateSections([CsFooter]);
  End;
End;

Procedure TCustomIntlMonthCalendar.CalcMetrics(ViewMode: TIntlCalViewMode;
  Out Metrics: TIntlCalMetrics);
Const
  HSpacing = 4;
  VSpacing = 2;
Var
  DC: HDC;
  OldFont: HFONT;
  NameKind: TIntlCalNameKind;
  DoW: TDayOfWeek;
  Extends: Array [TIntlCalNameKind] Of TSize;
  Extend: TSize;
  W, H, M: Integer;
  S: String;
Begin
  If (CsLoading In ComponentState) Or (WindowHandle = 0) Or
    (FUpdateMetricsCount <> 0) Then
    Exit;
  Inc(FUpdateMetricsCount);
  DC := CreateCompatibleDC(0);
  Try
    OldFont := SelectObject(DC, Font.Handle);
    Try
      Metrics.Bounds := GetCalendarClientRect;
      Metrics.HeaderHeight := 0;
      Metrics.FooterHeight := 0;
      Metrics.FixedRowHeight := 0;
      Metrics.FixedColWidth := 0;
      W := Metrics.Bounds.Right - Metrics.Bounds.Left;
      H := Metrics.Bounds.Bottom - Metrics.Bounds.Top;
      If ShowHeader Then
      Begin
        S := GetHeaderText(ViewMode);
        SelectObject(DC, FontHeader.Handle);
        GetTextExtentPoint32(DC, PChar(S), Length(S), Extend);
        SelectObject(DC, Font.Handle);
        Metrics.HeaderHeight := Max(3 * Extend.Cy Div 2, 13) + VSpacing;
        If Not Odd(Metrics.HeaderHeight) Then
          Inc(Metrics.HeaderHeight);
        Dec(H, Metrics.HeaderHeight);
      End;
      If ShowFooter Then
      Begin
        Metrics.LongFormat := True;
        SelectObject(DC, FontFooter.Handle);
        S := GetFooterText(Metrics.LongFormat);
        GetTextExtentPoint32(DC, PChar(S), Length(S), Extend);
        If Extend.Cx > 75 * W Div 100 Then
        Begin
          Metrics.LongFormat := False;
          S := GetFooterText(Metrics.LongFormat);
          GetTextExtentPoint32(DC, PChar(S), Length(S), Extend);
        End;
        SelectObject(DC, Font.Handle);
        Metrics.FooterHeight := Max(3 * Extend.Cy Div 2, 13) + VSpacing;
        If Not Odd(Metrics.FooterHeight) Then
          Inc(Metrics.FooterHeight);
        Dec(H, Metrics.FooterHeight);
      End;
      Case ViewMode Of
        VmDay:
          Begin
            Metrics.ColCount := Calendar.DaysPerWeek;
            Metrics.RowCount :=
              Ceil(Ceil(DaysInYear / MonthsInYear) / Metrics.ColCount) + 1;
            If WeekNumbers Then
            Begin
              S := GetWeekText(Calendar.WeeksInYear(Year));
              SelectObject(DC, FontWeekNumbers.Handle);
              GetTextExtentPoint32(DC, PChar(S), Length(S), Extend);
              SelectObject(DC, Font.Handle);
              Metrics.FixedColWidth := Extend.Cx + 3 * HSpacing Div 2 + 2;
              Dec(W, Metrics.FixedColWidth);
            End;
            S := GetDayText(Ceil(DaysInYear / MonthsInYear));
            GetTextExtentPoint32(DC, PChar(S), Length(S), Extend);
            For NameKind := Low(TIntlCalNameKind) To High(TIntlCalNameKind) Do
              Extends[NameKind] := Extend;
            SelectObject(DC, FontDayOfWeek.Handle);
            For DoW := Low(TDayOfWeek) To High(TDayOfWeek) Do
            Begin
              For NameKind := NkLong To NkShortest Do
              Begin
                S := GetDayName(DoW, NameKind);
                GetTextExtentPoint32(DC, PChar(S), Length(S), Extend);
                If Extend.Cx > Extends[NameKind].Cx Then
                  Extends[NameKind].Cx := Extend.Cx;
                If Extend.Cy > Extends[NameKind].Cy Then
                  Extends[NameKind].Cy := Extend.Cy;
              End;
            End;
            SelectObject(DC, Font.Handle);
            Metrics.NameKind := NkShortest;
            For NameKind := NkLong To NkShortest Do
            Begin
              If (Extends[NameKind].Cx + HSpacing) * Metrics.ColCount <= W Then
              Begin
                Metrics.NameKind := NameKind;
                Break;
              End;
            End;
            Metrics.FixedRowHeight := Extends[Metrics.NameKind].Cy +
              VSpacing + 2;
            Dec(H, Metrics.FixedRowHeight);
            Metrics.ColWidth := W Div Metrics.ColCount;
            Metrics.RowHeight := H Div Metrics.RowCount;
            Metrics.StretchedCols := W - Metrics.ColWidth * Metrics.ColCount;
            Metrics.StretchedRows := H - Metrics.RowHeight * Metrics.RowCount;
            Metrics.PrefferedWidth := Metrics.FixedColWidth +
              (Extends[Metrics.NameKind].Cx + HSpacing) * Metrics.ColCount;
            Metrics.PrefferedHeight := Metrics.FixedRowHeight +
              (Extends[Metrics.NameKind].Cy + 2 * VSpacing) * Metrics.RowCount;
          End;
        VmMonth:
          Begin
            Metrics.ColCount := 3;
            Metrics.RowCount := Ceil(MonthsInYear / Metrics.ColCount);
            Metrics.ColWidth := W Div Metrics.ColCount;
            Metrics.RowHeight := H Div Metrics.RowCount;
            Metrics.StretchedRows := H - (Metrics.RowHeight * Metrics.RowCount);
            Metrics.StretchedCols := W - (Metrics.ColWidth * Metrics.ColCount);
            FillChar(Extends[NkLong], SizeOf(Extend) * 2, 0);
            For M := 1 To Calendar.MaxMonthsPerYear Do
              For NameKind := NkLong To NkShort Do
              Begin
                S := GetMonthName(M, NameKind);
                GetTextExtentPoint32(DC, PChar(S), Length(S), Extend);
                If Extend.Cx > Extends[NameKind].Cx Then
                  Extends[NameKind].Cx := Extend.Cx;
                If Extend.Cy > Extends[NameKind].Cy Then
                  Extends[NameKind].Cy := Extend.Cy;
              End;
            Metrics.NameKind := NkShort;
            For NameKind := NkLong To NkShort Do
            Begin
              If (Extends[NameKind].Cx + HSpacing) * Metrics.ColCount <= W Then
              Begin
                Metrics.NameKind := NameKind;
                Break;
              End;
            End;
            Metrics.PrefferedWidth := (Extends[Metrics.NameKind].Cx + HSpacing)
              * Metrics.ColCount;
            Metrics.PrefferedHeight := (Extends[Metrics.NameKind].Cy + VSpacing)
              * Metrics.RowCount;
          End;
        VmYear:
          Begin
            Metrics.ColCount := 4;
            Metrics.RowCount := 3;
            Metrics.ColWidth := W Div Metrics.ColCount;
            Metrics.RowHeight := H Div Metrics.RowCount;
            Metrics.StretchedRows := H - (Metrics.RowHeight * Metrics.RowCount);
            Metrics.StretchedCols := W - (Metrics.ColWidth * Metrics.ColCount);
            S := GetYearText(Calendar.YearOf(Calendar.MaxSupportedDateTime));
            GetTextExtentPoint32(DC, PChar(S), Length(S), Extend);
            Metrics.PrefferedWidth := (Extend.Cx + HSpacing) * Metrics.ColCount;
            Metrics.PrefferedHeight := (Extend.Cy + VSpacing) *
              Metrics.RowCount;
          End;
      End;
      Inc(Metrics.PrefferedHeight, Metrics.HeaderHeight);
      Inc(Metrics.PrefferedHeight, Metrics.FooterHeight);
    Finally
      SelectObject(DC, OldFont);
    End;
  Finally
    DeleteObject(DC);
    Dec(FUpdateMetricsCount);
  End;
End;

Procedure TCustomIntlMonthCalendar.CalcMetricsSections
  (Var Metrics: TIntlCalMetrics);
Var
  Sec: TIntlCalSection;
Begin
  If (CsLoading In ComponentState) Or (WindowHandle = 0) Or
    (FUpdateMetricsCount <> 0) Then
    Exit;
  For Sec := Low(TIntlCalSection) To High(TIntlCalSection) Do
    With FMetrics, Sections[Sec] Do
    Begin
      Sections[Sec] := Bounds;
      Case Sec Of
        CsHeader:
          Begin
            Bottom := Top + HeaderHeight;
          End;
        CsHeaderText:
          Begin
            Inc(Left, HeaderHeight);
            Dec(Right, HeaderHeight);
            Bottom := Top + HeaderHeight;
          End;
        CsArrowLeft:
          Begin
            Right := Left + HeaderHeight;
            Bottom := Top + HeaderHeight;
          End;
        CsArrowRight:
          Begin
            Left := Right - HeaderHeight;
            Bottom := Top + HeaderHeight;
          End;
        CsFooter:
          Begin
            Top := Bottom - FooterHeight;
          End;
        CsWeekSpace:
          Begin
            Inc(Top, HeaderHeight);
            If IsRightToLeft Then
              Left := Right - FixedColWidth
            Else
              Right := Left + FixedColWidth;
            Bottom := Top + FixedRowHeight;
          End;
        CsDaysOfWeek:
          Begin
            Inc(Top, HeaderHeight);
            If IsRightToLeft Then
              Dec(Right, FixedColWidth)
            Else
              Inc(Left, FixedColWidth);
            Bottom := Top + FixedRowHeight;
          End;
        CsWeekNumbers:
          Begin
            Inc(Top, HeaderHeight);
            Inc(Top, FixedRowHeight);
            If IsRightToLeft Then
              Left := Right - FixedColWidth
            Else
              Right := Left + FixedColWidth;
            Dec(Bottom, FooterHeight);
          End;
        CsYMD:
          Begin
            Inc(Top, HeaderHeight);
            Inc(Top, FixedRowHeight);
            If IsRightToLeft Then
              Dec(Right, FixedColWidth)
            Else
              Inc(Left, FixedColWidth);
            Dec(Bottom, FooterHeight);
          End;
      End;
    End;
End;

Procedure TCustomIntlMonthCalendar.UpdateMetrics;
Begin
  If FUpdateMetricsCount = 0 Then
  Begin
    FPrefferedSizeReady := False;
    CalcMetrics(CurrentView, FMetrics);
    CalcMetricsSections(FMetrics);
    If Not(CsLoading In ComponentState) Then
      AdjustSize;
    Invalidate;
  End;
End;

Procedure TCustomIntlMonthCalendar.UpdateMinMaxRanges;
Var
  I: Integer;
Begin
  If Calendar.IsNoDate(DateMin) Or
    (Calendar.CompareDate(DateMin, Calendar.MinSupportedDateTime) < 0) Then
    FActualDateMin := Trunc(Calendar.MinSupportedDateTime)
  Else
    FActualDateMin := DateMin;
  If Calendar.IsNoDate(DateMax) Or
    (Calendar.CompareDate(DateMax, Calendar.MaxSupportedDateTime) > 0) Then
    FActualDateMax := Trunc(2958465.99999)
    // Trunc(Calendar.MaxSupportedDateTime)
  Else
    FActualDateMax := DateMax;
  If Not(CsLoading In ComponentState) Then
  Begin
    Case SelectionStyle Of
      SsRange:
        FSelectionAnchor := EnsureDateInRange(FSelectionAnchor);
      SsFree:
        Begin
          FSelection.Find(ActualDateMin, I);
          If I = FSelection.Count Then
            FSelection.Clear
          Else
          Begin
            For I := I - 1 Downto 0 Do
              FSelection.Delete(I);
            If FSelection.Find(ActualDateMax, I) Then
              Inc(I);
            If I = 0 Then
              FSelection.Clear
            Else
            Begin
              For I := FSelection.Count - 1 Downto I Do
                FSelection.Delete(I);
            End;
          End;
        End;
    End;
    FocusDate(EnsureDateInRange(Date));
    InvalidateDateDependentSections;
  End;
End;

Procedure TCustomIntlMonthCalendar.UpdateCalendar;
Begin
  Inc(FUpdateMetricsCount);
  Try
    ParentBiDiMode := True;
    UpdateMinMaxRanges;
    ExtractDateInfo(EnsureDateInRange(Date));
  Finally
    Dec(FUpdateMetricsCount);
  End;
  UpdateMetrics;
End;

Procedure TCustomIntlMonthCalendar.ValidateCalendar;

  Procedure InitNewCalendar;
  Begin
    FCalendar.OnChange := CalendarChanged;
    UpdateCalendar;
    DoCalendarChange;
  End;

Var
  Reference: TCalendar;
Begin
  If CsLoading In ComponentState Then
    Exit;
  If Not Assigned(CalendarType) Then
  Begin
    Reference := DefaultCalendar;
    If Reference.ClassType <> Calendar.ClassType Then
    Begin
      FreeAndNil(FCalendar);
      FCalendar := Reference.Clone;
      InitNewCalendar;
    End
    Else
      FCalendar.Assign(Reference);
  End
  Else If CalendarType <> Calendar.ClassType Then
  Begin
    FreeAndNil(FCalendar);
    FCalendar := CalendarType.Create(LOCALE_USER_DEFAULT);
    InitNewCalendar;
  End
  Else
    FCalendar.Settings.Prepare(LOCALE_USER_DEFAULT);
End;

Procedure TCustomIntlMonthCalendar.ExtractDateInfo(Const ADate: TDate);
Begin
  If CsLoading In ComponentState Then
    Exit;
  FDate := ADate;
  Calendar.DecodeDate(Date, FEra, FYear, FMonth, FDay);
  FMonthsInYear := Calendar.MonthsInYear(Era, Year);
  FDaysInYear := Calendar.DaysInYear(Era, Year);
  FDaysInMonth := Calendar.DaysInMonth(Era, Year, Month);
  FMonthFirstDate := Date - Day + 1;
  FMonthFirstDayOfWeek := Calendar.DayOfWeek(MonthFirstDate);
  FMonthStartCell := Calendar.DaysOfWeekBetween(MonthFirstDayOfWeek,
    Calendar.Settings.FirstDayOfWeek);
  If FMonthStartCell = 0 Then
    Inc(FMonthStartCell, Calendar.DaysPerWeek);
  FFirstCellDayNumber := Calendar.DayOfMonth(FirstCellDate);
  FFirstCellYear := Calendar.OffsetYear(Era, Year, -(Year Mod 10) - 1);
End;

Procedure TCustomIntlMonthCalendar.FocusDate(Const ADate: TDate);
Var
  OldEra, OldYear, OldMonth, OldDay: Integer;
  OldMonthsInYear: Integer;
Begin
  If (CsLoading In ComponentState) Or (ADate = Date) Then
    Exit;
  OldDay := Day;
  FDay := Trunc(ADate) - Trunc(MonthFirstDate) + 1;
  If (Day >= 1) And (Day <= DaysInMonth) Then
  Begin
    FDate := ADate;
    If CurrentView = VmDay Then
    Begin
      InvalidateCell(MonthStartCell + OldDay - 1);
      InvalidateCell(MonthStartCell + Day - 1);
    End;
  End
  Else
  Begin
    OldEra := Era;
    OldYear := Year;
    OldMonth := Month;
    FDay := OldDay;
    OldMonthsInYear := MonthsInYear;
    AnimateBegin;
    Try
      ExtractDateInfo(ADate);
      Case CurrentView Of
        VmDay:
          Begin
            AnimatePlay;
            DoPageChange;
            InvalidateDateDependentSections;
          End;
        VmMonth:
          Begin
            If (Era <> OldEra) Or (Year <> OldYear) Then
            Begin
              If OldMonthsInYear <> MonthsInYear Then
                CalcMetrics(CurrentView, FMetrics);
              AnimatePlay;
              DoPageChange;
              InvalidateDateDependentSections;
            End
            Else
            Begin
              InvalidateCell(OldMonth - 1);
              InvalidateCell(Month - 1);
            End;
          End;
        VmYear:
          Begin
            If (Era <> OldEra) Or ((OldYear Div 10) <> (Year Div 10)) Then
            Begin
              AnimatePlay;
              DoPageChange;
              InvalidateDateDependentSections;
            End
            Else
            Begin
              InvalidateCell(OldYear - FirstCellYear);
              InvalidateCell(Year - FirstCellYear);
            End;
          End;
      End;
    Finally
      AnimateEnd;
    End;
  End;
  DoDateChange;
End;

Function TCustomIntlMonthCalendar.GetDateAt(X, Y: Integer;
  Out ADate: TDate): Boolean;
Var
  Cell: Integer;
Begin
  Result := (GetHitTestInfoAt(X, Y, Cell) = ChtDay) And
    GetDateAtCell(Cell, ADate);
End;

Function TCustomIntlMonthCalendar.GetDateAtCell(Index: Integer;
  Out ADate: TDate): Boolean;
Begin
  Result := False;
  If (CurrentView = VmDay) And (Index >= 0) And (Index < CellCount) Then
  Begin
    ADate := FirstCellDate + Index;
    Result := IsDateInRange(ADate);
  End;
End;

Function TCustomIntlMonthCalendar.GetDatesAsText: String;
Var
  I: Integer;
  S: String;
  Continuous: Boolean;
Begin
  Continuous := False;
  S := '';
  For I := 0 To SelectionCount - 2 Do
  Begin
    If Selection[I] + 1 = Selection[I + 1] Then
    Begin
      If Not Continuous Then
      Begin
        S := S + Format('yyyy/mm/dd', Selection[I]) + '-';
        Continuous := True;
      End
    End
    Else
    Begin
      S := S + Format('yyyy/mm/dd', Selection[I]) + ',';
      Continuous := False;
    End;
  End;

  If SelectionCount > 0 Then
    S := S + Format('yyyy/mm/dd', Selection[SelectionCount - 1]);
  Result := S;
End;

Function TCustomIntlMonthCalendar.GetMonthAt(X, Y: Integer;
  Out AMonth: Integer): Boolean;
Var
  Cell: Integer;
Begin
  Result := (GetHitTestInfoAt(X, Y, Cell) = ChtMonth) And
    GetMonthAtCell(Cell, AMonth);
End;

Function TCustomIntlMonthCalendar.GetMonthAtCell(Index: Integer;
  Out AMonth: Integer): Boolean;
Begin
  Result := False;
  If (CurrentView = VmMonth) And (Index >= 0) And (Index < CellCount) Then
  Begin
    AMonth := Index + 1;
    Result := IsMonthInRange(Era, Year, AMonth);
  End;
End;

Function TCustomIntlMonthCalendar.GetYearAt(X, Y: Integer;
  Out AYear: Integer): Boolean;
Var
  Cell: Integer;
Begin
  Result := (GetHitTestInfoAt(X, Y, Cell) = ChtYear) And
    GetYearAtCell(Cell, AYear);
End;

Function TCustomIntlMonthCalendar.GetYearAtCell(Index: Integer;
  Out AYear: Integer): Boolean;
Begin
  Result := False;
  If (CurrentView = VmYear) And (Index >= 0) And (Index < CellCount) Then
  Begin
    AYear := Calendar.OffsetYear(FirstCellYear, Index);
    Result := IsYearInRange(Era, AYear);
  End;
End;

Procedure TCustomIntlMonthCalendar.SelectDate(Const ADate: TDate;
  ShiftState: TShiftState);
Var
  I, Dir: Integer;
  AnchorDate, TheDate: TDate;
  Changed: Boolean;
Begin
  Changed := False;
  AnchorDate := EnsureDateInRange(ADate);
  Case SelectionStyle Of
    SsSingle:
      Changed := (FDate <> AnchorDate);
    SsRange:
      If (SsLeft In ShiftState) And (FSelectionAnchor <> AnchorDate) Then
      Begin
        If Not(SsShift In ShiftState) Then
          FSelectionAnchor := AnchorDate;
        InvalidateDateRange(FSelectionAnchor, AnchorDate);
        Changed := True;
      End;
  Else // ssFree
    If Not(SsCtrl In ShiftState) And (FSelection.Count <> 0) Then
    Begin
      InvalidateDates(FSelection);
      FSelection.Clear;
      Changed := True;
    End;
    If SsLeft In ShiftState Then
    Begin
      If SsShift In ShiftState Then
      Begin
        If AnchorDate < FSelectionAnchor Then
          Dir := +1
        Else
          Dir := -1;
        TheDate := AnchorDate;
        For I := 0 To Calendar.DaysBetween(FSelectionAnchor, AnchorDate) Do
        Begin
          If Not FSelection.Exists(TheDate) And DoDateSelectable(TheDate) Then
          Begin
            FSelection.Add(TheDate);
            InvalidateDate(TheDate);
            Changed := True;
          End;
          Calendar.IncDay(TheDate, Dir);
        End;
      End
      Else If SsCtrl In ShiftState Then
      Begin
        If FSelection.Remove(AnchorDate) >= 0 Then
        Begin
          InvalidateDate(AnchorDate);
          Changed := True;
        End
        Else If DoDateSelectable(AnchorDate) Then
        Begin
          FSelection.Add(AnchorDate);
          InvalidateDate(AnchorDate);
          Changed := True;
        End;
        FSelectionAnchor := AnchorDate;
      End
      Else
      Begin
        If Not FSelection.Exists(AnchorDate) And
          DoDateSelectable(AnchorDate) Then
        Begin
          InvalidateDate(AnchorDate);
          FSelection.Add(AnchorDate);
          Changed := True;
        End;
        FSelectionAnchor := AnchorDate;
      End;
    End;
  End; // case
  FocusDate(AnchorDate);
  If Changed Then
    DoSelectionChange;
End;

Procedure TCustomIntlMonthCalendar.Select(Const ADate: TDate;
  ShiftState: TShiftState);
Begin
  SelectDate(ADate, ShiftState + [SsLeft]);
End;

Procedure TCustomIntlMonthCalendar.ClearSelection;
Begin
  Case SelectionStyle Of
    SsRange:
      Begin
        If FSelectionAnchor <> Date Then
        Begin
          InvalidateDateRange(FSelectionAnchor, Date);
          FSelectionAnchor := Date;
          DoSelectionChange;
        End;
      End;
    SsFree:
      Begin
        If FSelection.Count <> 0 Then
        Begin
          InvalidateDates(FSelection);
          FSelection.Clear;
          DoSelectionChange;
        End;
      End;
  End;
End;

Function TCustomIntlMonthCalendar.IsSelected(Const ADate: TDate): Boolean;
Begin
  Case SelectionStyle Of
    SsSingle:
      Result := Calendar.IsSameDay(ADate, Date);
    SsRange:
      Begin
        Case Calendar.CompareDate(ADate, Date) Of
          - 1:
            Result := (Calendar.CompareDate(ADate, FSelectionAnchor) >= 0);
          +1:
            Result := (Calendar.CompareDate(ADate, FSelectionAnchor) <= 0);
        Else
          Result := True;
        End;
      End
  Else
    Result := FSelection.Exists(Trunc(ADate));
  End;
End;

Function TCustomIntlMonthCalendar.IsYearInRange(AEra, AYear: Integer): Boolean;
Begin
  Result := Calendar.IsValidYear(AEra, AYear) And
    (IsDateInRange(Calendar.StartOfYear(AEra, AYear)) Or
    IsDateInRange(Calendar.EndOfYear(AEra, AYear)));
End;

Function TCustomIntlMonthCalendar.IsYearInRange(AYear: Integer): Boolean;
Begin
  Result := IsYearInRange(Era, AYear);
End;

Function TCustomIntlMonthCalendar.IsMonthInRange(AEra, AYear,
  AMonth: Integer): Boolean;
Begin
  Result := Calendar.IsValidMonth(AEra, AYear, AMonth) And
    (IsDateInRange(Calendar.StartOfMonth(AEra, AYear, AMonth)) Or
    IsDateInRange(Calendar.EndOfMonth(AEra, AYear, AMonth)));
End;

Function TCustomIntlMonthCalendar.IsMonthInRange(AYear,
  AMonth: Integer): Boolean;
Begin
  Result := IsMonthInRange(Era, AYear, AMonth);
End;

Function TCustomIntlMonthCalendar.IsDateInRange(Const ADate: TDate): Boolean;
Begin
  Result := Calendar.IsInRange(Trunc(ADate), ActualDateMin, ActualDateMax);
End;

Function TCustomIntlMonthCalendar.EnsureDateInRange(Const ADate: TDate): TDate;
Begin
  If IsDateInRange(ADate) Then
    Result := Trunc(ADate)
  Else If Calendar.CompareDate(ADate, ActualDateMax) > 0 Then
    Result := ActualDateMax
  Else
    Result := ActualDateMin;
End;

Function TCustomIntlMonthCalendar.Format(Const FmtStr: String;
  Const ADate: TDateTime): String;
Begin
  Result := Calendar.Format(FmtStr, ADate);
  NormalizeDigits(Result);
End;

{ TIntlMonthCalendarPopup }

Constructor TIntlMonthCalendarPopup.Create(AOwner: TWinControl);
Begin
  Inherited Create(AOwner);
  ControlStyle := ControlStyle + [CsNoDesignVisible];
  TabStop := False;
  ShowFooter := True;
  ParentColor := True;
  BevelKind := BkNone;
  Visible := False;
  Parent := AOwner;
End;

Procedure TIntlMonthCalendarPopup.WMMouseActivate(Var Message: TMessage);
Begin
  Message.Result := MA_NOACTIVATE;
End;

Procedure TIntlMonthCalendarPopup.CMVisibleChanged(Var Message: TMessage);
Begin
  Inherited;
  CurrentView := VmDay;
  If Not Visible And Assigned(Parent) Then
    Parent.Invalidate;
End;

Procedure TIntlMonthCalendarPopup.MouseUp(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
// var
// Offset: Integer;
Begin
  // if (Button = mbLeft) and (GetHitTestInfoAt(X, Y, Offset) in [chtDay, chtFooter]) then
  // DoCloseUp(True);
  Inherited MouseUp(Button, Shift, X, Y);
End;

Procedure TIntlMonthCalendarPopup.KeyDown(Var Key: Word; Shift: TShiftState);
Begin
  If Key In [VK_RETURN, VK_ESCAPE] Then
  Begin
    DoCloseUp(Key = VK_RETURN);
    Key := 0;
  End;
  Inherited KeyDown(Key, Shift);
End;

Procedure TIntlMonthCalendarPopup.CreateParams(Var Params: TCreateParams);
Begin
  Inherited CreateParams(Params);
  With Params Do
  Begin
    Style := WS_POPUP Or WS_BORDER;
    WindowClass.Style := WindowClass.Style Or CS_SAVEBITS Or CS_DROPSHADOW;
  End;
End;

Function TIntlMonthCalendarPopup.Focused: Boolean;
Begin
  Result := Parent.Focused;
End;

Procedure TIntlMonthCalendarPopup.Show(X, Y: Integer);
Begin
  SetWindowPos(Handle, HWND_TOP, X, Y, 0, 0, SWP_NOSIZE Or SWP_NOACTIVATE Or
    SWP_SHOWWINDOW);
  Visible := True;
End;

Procedure TIntlMonthCalendarPopup.DblClick;
Var
  Offset: Integer;
Begin
  // if (GetHitTestInfoAt(X, Y, Offset) in [chtDay, chtFooter]) then
  DoCloseUp(True);
  Inherited;
End;

Procedure TIntlMonthCalendarPopup.DoCloseUp(Accept: Boolean);
Begin
  If Assigned(OnCloseUp) Then
    OnCloseUp(Self, Accept);
End;

{ TCustomIntlDatePicker }

Constructor TCustomIntlDatePicker.Create(AOwner: TComponent);
Begin
  Inherited Create(AOwner);
  ControlStyle := ControlStyle + [CsOpaque] - [CsSetCaption];
  Width := 180;
  FDateFormat := 'yyyy/mm/dd dddd';
  FCalendarPopup := TIntlMonthCalendarPopup.Create(Self);
  FCalendarPopup.SetSubComponent(True);
  FCalendarPopup.Name := 'CalendarPopup';
  FCalendarPopup.OnCloseUp := PopupCloseUp;
  FCalendarPopup.OnCalendarChange := CalendarChanged;
  FCalendarPopup.CalendarType := TJalaliCalendar;
  FDate := Calendar.Today;
  FBorderStyle := BsSingle;
  FShowCheckBox := False;
  FChecked := True;
  ParentColor := False;
  Color := ClWindow;
  TabStop := True;
  AutoSize := True;
  UpdateText;
  AutoToday := True;
End;

Destructor TCustomIntlDatePicker.Destroy;
Begin
  If FTheme <> 0 Then
    CloseThemeData(FTheme);
  Inherited Destroy;
End;

Procedure TCustomIntlDatePicker.CalendarChanged(Sender: TObject);
Begin
  UpdateText;
End;

Procedure TCustomIntlDatePicker.PopupCloseUp(Sender: TObject; Accept: Boolean);
Begin
  If Accept Then
    Date := CalendarPopup.Date;
  CalendarPopup.Visible := False;
End;

Procedure TCustomIntlDatePicker.SetDate(Const Value: TDate);
Begin
  If (CsDesigning In ComponentState) And Calendar.IsNoDate(Value) Then
  Begin
    Date := Calendar.Today;
    Exit;
  End;
  If Not IsDateInRange(Value) Then
  Begin
    Raise EDateTimeError.CreateResFmt(@SDateOutOfRangeError,
      [Calendar.Format('ddddd', ActualDateMin), Calendar.Format('ddddd',
      ActualDateMax)]);
  End;
  If Not Calendar.IsSameDay(Date, Value) And CanModify Then
  Begin
    FDate := Calendar.DateOf(Value);
    UpdateText;
    Change;
  End;
End;

Procedure TCustomIntlDatePicker.SetDate(DateStr, FmtStr: String);
Var
  DT: TDateTime;
Begin
  DT := Now;
  If DateStr <> '' Then
    Calendar.TryParse(FmtStr, DateStr, DT);
  Date := DT;
End;

Procedure TCustomIntlDatePicker.SetDateFormat(Const Value: String);
Begin
  If DateFormat <> Value Then
  Begin
    FDateFormat := Value;
    UpdateText;
  End;
End;

Procedure TCustomIntlDatePicker.SetChecked(Value: Boolean);
Begin
  If (Checked <> Value) And CanModify Then
  Begin
    FChecked := Value;
    If Not Checked And CalendarPopup.Visible Then
      PopupCloseUp(CalendarPopup, False);
    InvalidateCheckBox;
    InvalidateDate;
    Change;
  End;
End;

Procedure TCustomIntlDatePicker.SetShowCheckBox(Value: Boolean);
Begin
  If ShowCheckBox <> Value Then
  Begin
    FShowCheckBox := Value;
    UpdateRects;
  End;
End;

Procedure TCustomIntlDatePicker.SetAutoToday(Const Value: Boolean);
Begin
  FAutoToday := Value;
  If FAutoToday Then
    SetToday;
End;

Procedure TCustomIntlDatePicker.SetBorderStyle(Value: TBorderStyle);
Begin
  If BorderStyle <> Value Then
  Begin
    FBorderStyle := Value;
    RecreateWnd;
  End;
End;

Function TCustomIntlDatePicker.GetDateMin: TDate;
Begin
  Result := CalendarPopup.DateMin;
End;

Function TCustomIntlDatePicker.GetDateAsText: String;
Begin
  Result := FCalendarPopup.Format('yyyy/mm/dd', Date);
End;

Procedure TCustomIntlDatePicker.SetDateMin(Const Value: TDate);
Begin
  CalendarPopup.DateMin := Value;
End;

Function TCustomIntlDatePicker.GetDateMax: TDate;
Begin
  Result := CalendarPopup.DateMax;
End;

Procedure TCustomIntlDatePicker.SetDateMax(Const Value: TDate);
Begin
  CalendarPopup.DateMax := Value;
End;

Function TCustomIntlDatePicker.GetCalendarType: TCalendarClass;
Begin
  Result := CalendarPopup.CalendarType;
End;

Procedure TCustomIntlDatePicker.SetCalendarType(Value: TCalendarClass);
Begin
  CalendarPopup.CalendarType := Value;
End;

Function TCustomIntlDatePicker.GetCalendar: TCalendar;
Begin
  Result := CalendarPopup.Calendar;
End;

Function TCustomIntlDatePicker.GetActualDateMin: TDate;
Begin
  Result := CalendarPopup.ActualDateMin;
End;

Function TCustomIntlDatePicker.GetActualDateMax: TDate;
Begin
  Result := CalendarPopup.ActualDateMax;
End;

Procedure TCustomIntlDatePicker.SetToday;
Begin
  SetDate(Now);
End;

Function TCustomIntlDatePicker.GetText: String;
Begin
  Result := Caption;
End;

Function TCustomIntlDatePicker.GetTheme: HTHEME;
Begin
  If FTheme = 0 Then
    FTheme := OpenThemeData(WindowHandle, 'datepicker');
  Result := FTheme;
End;

Function TCustomIntlDatePicker.IsDateStored: Boolean;
Begin
  Result := Not Calendar.IsToday(Date);
End;

Function TCustomIntlDatePicker.IsDateMinStored: Boolean;
Begin
  Result := Not Calendar.IsNoDate(DateMin);
End;

Function TCustomIntlDatePicker.IsDateMaxStored: Boolean;
Begin
  Result := Not Calendar.IsNoDate(DateMax);
End;

Procedure TCustomIntlDatePicker.ReadCalendarType(Reader: TReader);
Begin
  CalendarType := CalendarTypes.ByName(Reader.ReadString);
End;

Procedure TCustomIntlDatePicker.WriteCalendarType(Writer: TWriter);
Begin
  Writer.WriteString(CalendarType.CalendarName);
End;

Procedure TCustomIntlDatePicker.DefineProperties(Filer: TFiler);
Begin
  Inherited DefineProperties(Filer);
  Filer.DefineProperty('CalendarClass', ReadCalendarType, WriteCalendarType,
    Assigned(CalendarType));
End;

Procedure TCustomIntlDatePicker.CreateParams(Var Params: TCreateParams);
Begin
  Inherited CreateParams(Params);
  If BorderStyle = BsSingle Then
  Begin
    If Ctl3D Then
      Params.ExStyle := Params.ExStyle Or WS_EX_CLIENTEDGE
    Else
      Params.Style := Params.Style Or WS_BORDER;
  End;
End;

Procedure TCustomIntlDatePicker.Resize;
Begin
  Inherited Resize;
  UpdateRects;
End;

Function TCustomIntlDatePicker.CanAutoSize(Var NewWidth,
  NewHeight: Integer): Boolean;
Begin
  If AutoSize Then
    NewHeight := GetMinHeight;
  Result := True;
End;

Procedure TCustomIntlDatePicker.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
Begin
  If Button = MbLeft Then
  Begin
    Case HotInfo Of
      PhtCheckBox:
        Begin
          FCheckPressed := True;
          InvalidateCheckBox;
        End;
      PhtDropDown, PhtDate:
        Begin
          If CalendarPopup.Visible Then
            PopupCloseUp(CalendarPopup, False)
          Else
            PopupDropDown;
        End;
    End;
  End;
  Inherited MouseDown(Button, Shift, X, Y);
End;

Procedure TCustomIntlDatePicker.MouseMove(Shift: TShiftState; X, Y: Integer);
Var
  TheHitTest: TIntlPickerHitTest;
Begin
  TheHitTest := GetHitTestInfoAt(X, Y);
  If TheHitTest <> HotInfo Then
  Begin
    Case HotInfo Of
      PhtCheckBox:
        InvalidateCheckBox;
      PhtDropDown:
        InvalidateDropDown;
    End;
    FHotInfo := TheHitTest;
    Case HotInfo Of
      PhtCheckBox:
        InvalidateCheckBox;
      PhtDropDown:
        InvalidateDropDown;
    End;
  End;
  Inherited MouseMove(Shift, X, Y);
End;

Procedure TCustomIntlDatePicker.MouseUp(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
Begin
  If Button = MbLeft Then
  Begin
    If CheckPressed Then
    Begin
      If HotInfo = PhtCheckBox Then
        Checked := Not Checked
      Else
        InvalidateCheckBox;
    End;
    FCheckPressed := False;
  End;
  Inherited MouseUp(Button, Shift, X, Y);
End;

Function TCustomIntlDatePicker.DoMouseWheelDown(Shift: TShiftState;
  MousePos: TPoint): Boolean;
Begin
  Result := Inherited DoMouseWheelDown(Shift, MousePos);
  If Not Result Then
  Begin
    Date := EnsureDateInRange(Calendar.NextDay(Date));
    Result := True;
  End;
End;

Function TCustomIntlDatePicker.DoMouseWheelUp(Shift: TShiftState;
  MousePos: TPoint): Boolean;
Begin
  Result := Inherited DoMouseWheelUp(Shift, MousePos);
  If Not Result Then
  Begin
    Date := EnsureDateInRange(Calendar.PrevDay(Date));
    Result := True;
  End;
End;

Procedure TCustomIntlDatePicker.KeyDown(Var Key: Word; Shift: TShiftState);
Begin
  Inherited KeyDown(Key, Shift);
  If CalendarPopup.Visible Then
  Begin
    CalendarPopup.KeyDown(Key, Shift);
    Exit;
  End;
  Case Key Of
    VK_DOWN:
      Begin
        If SsAlt In Shift Then
          PopupDropDown
        Else
          Date := EnsureDateInRange(Calendar.NextDay(Date));
        Key := 0;
      End;
    VK_UP:
      Begin
        Date := EnsureDateInRange(Calendar.PrevDay(Date));
        Key := 0;
      End;
    VK_NEXT:
      Begin
        Date := EnsureDateInRange(Calendar.NextMonth(Date));
        Key := 0;
      End;
    VK_PRIOR:
      Begin
        Date := EnsureDateInRange(Calendar.PrevMonth(Date));
        Key := 0;
      End;
    VK_SPACE:
      Begin
        If ShowCheckBox Then
          Checked := Not Checked;
        Key := 0;
      End;
  End;
End;

Function TCustomIntlDatePicker.CanModify: Boolean;
Begin
  Result := True;
End;

Procedure TCustomIntlDatePicker.Paint;
Const
  CheckStates: Array [Boolean] Of TCheckBoxState = (CbUnchecked, CbChecked);
Var
  DrawState: TPartDrawState;
  DrawFlags: Cardinal;
  TextRect, FocusRect: TRect;
  StateId: Integer;
Begin
  // check box
  If ShowCheckBox And RectVisible(Canvas.Handle, FCheckBoxRect) Then
  Begin
    If Enabled Then
      If HotInfo = PhtCheckBox Then
        If CheckPressed Then
          DrawState := PdsPressed
        Else
          DrawState := PdsHot
      Else
        DrawState := PdsNormal
    Else
      DrawState := PdsDisabled;
    FCheckBoxRect := DrawCheckBox(Canvas.Handle, FCheckBoxRect,
      CheckStates[Checked], DrawState, False, ThemeControl(Self));
  End;
  // drop down
  If RectVisible(Canvas.Handle, FDropDownRect) Then
  Begin
    If Enabled Then
      If CalendarPopup.Visible Then
        DrawState := PdsPressed
      Else If HotInfo = PhtDropDown Then
        DrawState := PdsHot
      Else
        DrawState := PdsNormal
    Else
      DrawState := PdsDisabled;
    DrawDropDown(Canvas.Handle, FDropDownRect, DrawState, ThemeControl(Self));
  End;
  // content
  If RectVisible(Canvas.Handle, FDateRect) Then
  Begin
    Canvas.Font := Font;
    Canvas.Brush.Color := Color;
    Canvas.FillRect(FDateRect);
    TextRect := FDateRect;
    InflateRect(TextRect, -2, 0);
    DrawFlags := DrawTextBiDiModeFlags(DT_LEFT Or DT_VCENTER Or DT_NOPREFIX Or
      DT_SINGLELINE);
    If CalendarPopup.IsThemed Then
    Begin
      If Checked And Enabled Then
        StateId := DPDT_NORMAL
      Else
        StateId := DPDT_DISABLED;
      DrawThemeText(Theme, Canvas.Handle, DP_DATETEXT, StateId, PChar(Caption),
        Length(Caption), DrawFlags, 0, TextRect);
    End
    Else
    Begin
      If Not(Checked And Enabled) Then
        Canvas.Font.Color := ClGrayText;
      DrawText(Canvas.Handle, PChar(Caption), Length(Caption), TextRect,
        DrawFlags);
    End;
    If Focused And Not CalendarPopup.Visible Then
    Begin
      FillChar(FocusRect, SizeOf(FocusRect), 0);
      DrawText(Canvas.Handle, PChar(Caption), Length(Caption), FocusRect,
        DrawFlags Or DT_CALCRECT);
      OffsetRect(FocusRect, 0,
        (TextRect.Top + TextRect.Bottom - FocusRect.Bottom) Div 2);
      If UseRightToLeftAlignment Then
        OffsetRect(FocusRect, TextRect.Right - FocusRect.Right, 0)
      Else
        OffsetRect(FocusRect, TextRect.Left, 0);
      InflateRect(FocusRect, 2, 4);
      SetTextColor(Canvas.Handle, ColorToRGB(Font.Color));
      Canvas.DrawFocusRect(FocusRect);
    End;
  End;
End;

Procedure TCustomIntlDatePicker.UpdateText;
Begin
  Caption := CalendarPopup.Format(DateFormat, Date);
  InvalidateDate;
End;

Procedure TCustomIntlDatePicker.UpdateRects;
Var
  Rect: TRect;
Begin
  If Not HandleAllocated Then
    Exit;
  Rect := ClientRect;
  FDateRect := Rect;
  FDropDownRect := Rect;
  FCheckBoxRect := Rect;
  If UseRightToLeftAlignment Then
  Begin
    FDropDownRect.Right := FDropDownRect.Left + GetDropDownSize.Cx;
    FDateRect.Left := FDropDownRect.Right + 2;
    If FShowCheckBox Then
    Begin
      FCheckBoxRect.Left := FCheckBoxRect.Right - GetCheckBoxSize.Cx - 4;
      FDateRect.Right := FCheckBoxRect.Left - 2;
    End;
  End
  Else
  Begin
    FDropDownRect.Left := FDropDownRect.Right - GetDropDownSize.Cx;
    FDateRect.Right := FDropDownRect.Left - 2;
    If FShowCheckBox Then
    Begin
      FCheckBoxRect.Right := FCheckBoxRect.Left + GetCheckBoxSize.Cx + 4;
      FDateRect.Left := FCheckBoxRect.Right + 2;
    End;
  End;
  Invalidate;
End;

Function TCustomIntlDatePicker.GetMinHeight: Integer;
Var
  DC: HDC;
  OldFont: HFONT;
  Metrics: TTextMetric;
Begin
  DC := GetDC(WindowHandle);
  Try
    OldFont := SelectObject(DC, Font.Handle);
    GetTextMetrics(DC, Metrics);
    SelectObject(DC, OldFont);
  Finally
    ReleaseDC(WindowHandle, DC);
  End;
  Result := Metrics.TmHeight + (Height - ClientHeight) + 4;
  If Result < GetDropDownSize.Cy Then
    Result := GetDropDownSize.Cy;
  If ShowCheckBox And (Result < GetCheckBoxSize.Cy) Then
    Result := GetCheckBoxSize.Cy;
End;

Function TCustomIntlDatePicker.GetHitTestInfoAt(X, Y: Integer)
  : TIntlPickerHitTest;
Var
  Pt: TPoint;
Begin
  Pt.X := X;
  Pt.Y := Y;
  If ShowCheckBox And PtInRect(FCheckBoxRect, Pt) Then
    Result := PhtCheckBox
  Else If PtInRect(FDropDownRect, Pt) Then
    Result := PhtDropDown
  Else If PtInRect(FDateRect, Pt) Then
    Result := PhtDate
  Else If PtInRect(ClientRect, Pt) Then
    Result := PhtControl
  Else
    Result := PhtNowhere;
End;

Procedure TCustomIntlDatePicker.InvalidateCheckBox;
Begin
  If ShowCheckBox And HandleAllocated Then
    InvalidateRect(WindowHandle, FCheckBoxRect, False);
End;

Procedure TCustomIntlDatePicker.InvalidateDropDown;
Begin
  If HandleAllocated Then
    InvalidateRect(WindowHandle, FDropDownRect, False);
End;

Procedure TCustomIntlDatePicker.InvalidateDate;
Begin
  If HandleAllocated Then
    InvalidateRect(WindowHandle, FDateRect, False);
End;

Procedure TCustomIntlDatePicker.InvalidateFrame;
Begin
  If ThemeControl(Self) And HandleAllocated Then
    RedrawWindow(WindowHandle, Nil, 0, RDW_FRAME Or RDW_INVALIDATE Or
      RDW_UPDATENOW);
End;

Procedure TCustomIntlDatePicker.PopupDropDown;
Var
  TopLeft: TPoint;
  X, Y, W, H: Integer;
  Monitor: TMonitor;
Begin
  If Enabled And CanModify And Not CalendarPopup.Visible Then
  Begin
    Monitor := Screen.MonitorFromWindow(Handle);
    CalendarPopup.HandleNeeded;
    TopLeft := Parent.ClientToScreen(Point(Left, Top));
    // CalendarPopup.GetPreferredControlSize(W, H);
    W := CalendarPopup.Width;
    H := CalendarPopup.Height;
    If W < Width Then
      W := Width;
    // ShowMessage(IntToStr(CalendarPopup.Width));
    If UseRightToLeftAlignment Then
      X := TopLeft.X + Width - W
    Else
      X := TopLeft.X;
    If X < Monitor.Left Then
      X := Monitor.Left
    Else If X + W > Monitor.Left + Monitor.Width Then
      X := Monitor.Left + Monitor.Width - W;
    Y := TopLeft.Y + Height;
    If Y + H > Monitor.Top + Monitor.Height Then
      Y := TopLeft.Y - H;
    CalendarPopup.Date := Date;
    Checked := True;
    CalendarPopup.SetBounds(X, Y, W, H);
    CalendarPopup.Show(X, Y);
    InvalidateDropDown;
    InvalidateDate;
  End;
End;

Procedure TCustomIntlDatePicker.Change;
Begin
  If Assigned(OnChange) Then
    OnChange(Self);
End;

Procedure TCustomIntlDatePicker.WMNCPaint(Var Message: TWMNCPaint);
Var
  DC: HDC;
  StateId: Integer;
  Rect: TRect;
Begin
  Inherited;
  If (BorderStyle = BsSingle) And ThemeControl(Self) Then
  Begin
    Rect := BoundsRect;
    OffsetRect(Rect, -Rect.Left, -Rect.Top);
    If Enabled Then
      If Focused Then
        StateId := DPDB_FOCUSED
      Else If HotInfo <> PhtNowhere Then
        StateId := DPDB_HOT
      Else
        StateId := DPDB_NORMAL
    Else
      StateId := DPDB_DISABLED;
    DC := GetWindowDC(WindowHandle);
    Try
      ExcludeClipRect(DC, Rect.Left + 2, Rect.Top + 2, Rect.Right - 2,
        Rect.Bottom - 2);
      DrawThemeBackground(Theme, DC, DP_DATEBORDER, StateId, Rect, Nil);
    Finally
      ReleaseDC(WindowHandle, DC);
    End;
  End;
End;

Procedure TCustomIntlDatePicker.WMMouseActivate(Var Message: TWMMouseActivate);
Begin
  Inherited;
  If CanFocus Then
    SetFocus;
End;

Procedure TCustomIntlDatePicker.WMSetFocus(Var Message: TWMSetFocus);
Begin
  Inherited;
  InvalidateDate;
  InvalidateFrame;
End;

Procedure TCustomIntlDatePicker.WMKillFocus(Var Message: TWMKillFocus);
Begin
  Inherited;
  If CalendarPopup.Visible Then
    PopupCloseUp(CalendarPopup, False);
  InvalidateDate;
  InvalidateFrame;
End;

Procedure TCustomIntlDatePicker.WMGetDlgCode(Var Message: TWMGetDlgCode);
Begin
  Message.Result := DLGC_WANTARROWS;
End;

Procedure TCustomIntlDatePicker.CMEnableChanged(Var Message: TMessage);
Begin
  Inherited;
  Invalidate;
End;

Procedure TCustomIntlDatePicker.CMBiDiModeChanged(Var Message: TMessage);
Begin
  Inherited;
  UpdateRects;
End;

Procedure TCustomIntlDatePicker.CMBorderChanged(Var Message: TMessage);
Begin
  Inherited;
  AdjustSize;
  InvalidateFrame;
End;

Procedure TCustomIntlDatePicker.CMCtl3DChanged(Var Message: TMessage);
Begin
  Inherited;
  AdjustSize;
  InvalidateFrame;
End;

Procedure TCustomIntlDatePicker.CMCancelMode(Var Message: TCMCancelMode);
Begin
  If (Message.Sender <> Self) And (Message.Sender <> CalendarPopup) And
    CalendarPopup.Visible And Not CalendarPopup.ContainsControl
    (Message.Sender) Then
    PopupCloseUp(CalendarPopup, False);
End;

Procedure TCustomIntlDatePicker.CMMouseEnter(Var Message: TMessage);
Begin
  Inherited;
  FHotInfo := PhtControl;
  InvalidateFrame;
End;

Procedure TCustomIntlDatePicker.CMMouseLeave(Var Message: TMessage);
Begin
  Case HotInfo Of
    PhtCheckBox:
      InvalidateCheckBox;
    PhtDropDown:
      InvalidateDropDown;
  End;
  FHotInfo := PhtNowhere;
  InvalidateFrame;
  Inherited;
End;

Function TCustomIntlDatePicker.IsYearInRange(AEra, AYear: Integer): Boolean;
Begin
  Result := CalendarPopup.IsYearInRange(AEra, AYear);
End;

Function TCustomIntlDatePicker.IsYearInRange(AYear: Integer): Boolean;
Begin
  Result := CalendarPopup.IsYearInRange(AYear);
End;

Function TCustomIntlDatePicker.IsMonthInRange(AEra, AYear,
  AMonth: Integer): Boolean;
Begin
  Result := CalendarPopup.IsMonthInRange(AEra, AYear, AMonth);
End;

Function TCustomIntlDatePicker.IsMonthInRange(AYear, AMonth: Integer): Boolean;
Begin
  Result := CalendarPopup.IsMonthInRange(AYear, AMonth);
End;

Function TCustomIntlDatePicker.IsDateInRange(Const ADate: TDate): Boolean;
Begin
  Result := CalendarPopup.IsDateInRange(ADate);
End;

Function TCustomIntlDatePicker.EnsureDateInRange(Const ADate: TDate): TDate;
Begin
  Result := CalendarPopup.EnsureDateInRange(ADate);
End;

{ TEventItems }

Procedure TEventItems.AddDate(Date: TDate);
Begin
  TEventItem(Add).Date := Date;
  Changed;
End;

Procedure TEventItems.AddDate(Date: String);
Begin
  TEventItem(Add).DateStr := Date;
  // TNasimCalendar(GetOwner).Invalidate;
End;

Constructor TEventItems.Create(AOwner: TPersistent);
Begin
  Inherited Create(AOwner, TEventItem);
End;

Function TEventItems.FindItem(Date: TDate): Integer;
Var
  I: Integer;
Begin
  Result := -1;
  For I := 0 To Count - 1 Do
  Begin
    If Items[I].FDate = Date Then
    Begin
      Result := I;
      Break;
    End;
  End;
End;

Function TEventItems.GetItem(Index: Integer): TEventItem;
Begin
  Result := TEventItem( Inherited GetItem(Index));
End;

Function TEventItems.GetItemComment(Date: TDate): String;
Var
  I: Integer;
Begin
  Result := '';
  For I := 0 To Count - 1 Do
  Begin
    If Items[I].FDate = Date Then
    Begin
      Result := Items[I].Comment;
      Break;
    End;
  End;
End;

Procedure TEventItems.SetItem(Index: Integer; Value: TEventItem);
Begin
  Inherited SetItem(Index, Value);
End;

Procedure TEventItems.Update(Item: TCollectionItem);
Begin
  Inherited;
  // TEventItem(GetOwner).Changed(True);
End;

{ TEventItem }

Procedure TEventItem.Assign(Source: TPersistent);
Begin
  FDate := TEventItem(Source).FDate;
  FDateStr := TEventItem(Source).FDateStr;
  FComment := TEventItem(Source).FComment;
End;

Procedure TEventItem.Changed(AllItems: Boolean);
Begin
  Inherited;
  // TEventItems(GetOwner).Changed;//(True);
End;

Constructor TEventItem.Create(Collection: TCollection);
Begin
  Inherited Create(Collection);
  FDate := Now;
  FDateStr := TJalaliCalendar.Create.Format('yyyy/mm/dd', FDate);
End;

Destructor TEventItem.Destroy;
Begin
  Inherited Destroy;
End;

Function TEventItem.GetDisplayName: String;
Begin
  If FDateStr <> '' Then
    Result := FDateStr
  Else
    Result := Inherited;
End;

Procedure TEventItem.SetComment(Const Value: String);
Begin
  If (FComment <> Value) Then
  Begin
    FComment := Value;
    Changed(True);
  End;
End;

Procedure TEventItem.SetDate(Const Value: TDate);
Begin
  If (FDate <> Value) Then
  Begin
    FDate := Value;
    FDateStr := TJalaliCalendar.Create.Format('yyyy/mm/dd', FDate);
    Changed(True);
  End;
End;

Procedure TEventItem.SetDateStr(Const Value: String);
Begin
  If (FDateStr <> Value) Then
  Begin
    FDateStr := Value;
    FDate := TJalaliCalendar.Create.Parse('yyyy/mm/dd', FDateStr);
    Changed(True);
  End;
End;

Procedure TEventItem.SetIndex(Value: Integer);
Begin
  Inherited SetIndex(Value);
End;

End.
