Unit NReg;

{$I DELPHIAREA.INC}

Interface

Uses
  Windows, Classes, SysUtils, DesignIntf, DesignEditors, Forms, ColnEdit,
  NCtrls, NCalGregorian,
  NCalJalali, NCalHijri, NCalendars;

Type

  TCalendarSystemPropertyEditor = Class(TPropertyEditor)
  Public
    Function GetAttributes: TPropertyAttributes; Override;
    Procedure GetValues(Proc: TGetStrProc); Override;
    Function GetValue: String; Override;
    Procedure SetValue(Const Value: String); Override;
  End;

  TIntlDateTimePropertyEditor = Class(TDateProperty)
  Public
    Function GetAttributes: TPropertyAttributes; Override;
    Function GetValue: String; Override;
    Procedure SetValue(Const Value: String); Override;
    Procedure Edit; Override;
  End;

  TIntlDatePropertyEditor = Class(TDateProperty)
  Public
    Function GetAttributes: TPropertyAttributes; Override;
    Function GetValue: String; Override;
    Procedure SetValue(Const Value: String); Override;
    Procedure Edit; Override;
  End;

  TCalendarEventEditor = Class(TComponentEditor)
  Protected
    Function GetVerbCount: Integer; Override;
    Function GetVerb(Index: Integer): String; Override;
    Procedure ExecuteVerb(Index: Integer); Override;
  End;

  TCalendarSelectionEditor = Class(TSelectionEditor)
  Public
    Procedure RequiresUnits(Proc: TGetStrProc); Override;
  End;

  TStringDatePropEditor = Class(TStringProperty)
  Public
    Function GetAttributes: TPropertyAttributes; Override;
    Procedure Edit; Override;
  End;

Procedure Register;

Implementation

{$R NReg.dcr}

Uses
  NCalendar, NDateCtrls, Types, TypInfo, Graphics, Dialogs, Controls,
  StdCtrls, ExtCtrls, Menus, SysConst;

Procedure Register;
Begin
  // RegisterComponents('MyCal', [TIntlMonthCalendar, TIntlDatePicker]);
  RegisterComponents('NasimCalendar', [TNasimMonthCalendar, TNasimDatePicker]);
  // RegisterComponents('MyCal', [TJalaliCalendar, TGregorianCalendar, THijriCalendar]);
  RegisterComponentEditor(TIntlMonthCalendar, TCalendarEventEditor);

{$IFDEF COMPILER2010_UP}
  RegisterPropertyEditor(TypeInfo(TCalendarClass), Nil, '',
    TCalendarSystemPropertyEditor);
{$ENDIF}
  RegisterPropertyEditor(TypeInfo(TDate), TCustomIntlMonthCalendar, '',
    TIntlDatePropertyEditor);
  RegisterPropertyEditor(TypeInfo(TDate), TCustomIntlDatePicker, '',
    TIntlDatePropertyEditor);

  RegisterSelectionEditor(TCustomIntlMonthCalendar, TCalendarSelectionEditor);
  RegisterSelectionEditor(TCustomIntlDatePicker, TCalendarSelectionEditor);

  RegisterPropertyEditor(TypeInfo(String), TEventItem, 'DateStr',
    TStringDatePropEditor);
End;

Function PromptForDate(Const ACaption: String; Var DateTime: TDateTime;
  ACalendarType: TCalendarClass): Boolean;
Var
  Dialog: TForm;
  IntlCal: TIntlMonthCalendar;
Begin
  Result := False;
  Dialog := TForm.Create(Application);
  Try
    Dialog.BorderStyle := BsDialog;
    Dialog.Position := PoMainFormCenter;
    Dialog.ClientWidth := 360;
    Dialog.ClientHeight := 240;
    Dialog.BorderWidth := 8;
    Dialog.Caption := ACaption;
    IntlCal := TIntlMonthCalendar.Create(Dialog);
    // IntlCal.
    With IntlCal Do
    Begin

      CalendarType := ACalendarType;
      If ACalendarType <> TGregorianCalendar Then
        BiDiMode := BdRightToLeft;
      SetBounds(0, 0, Dialog.ClientWidth, Dialog.ClientHeight - 33);
      FontHeader.Style := [FsBold];
      FontHeader.Size := FontHeader.Size + 2;
      FontFooter.Style := [FsBold];
      FontDayOfWeek.Color := ClNavy;
      If Not Calendar.IsNoDate(DateTime) Then
        Date := DateTime;
      Parent := Dialog;
    End;
    With TButton.Create(Dialog) Do
    Begin
      SetBounds(Dialog.ClientWidth - 158, Dialog.ClientHeight - 25, 75, 25);
      Caption := 'OK';
      ModalResult := MrOk;
      Default := True;
      Parent := Dialog;
    End;
    With TButton.Create(Dialog) Do
    Begin
      SetBounds(Dialog.ClientWidth - 75, Dialog.ClientHeight - 25, 75, 25);
      Caption := 'Cancel';
      ModalResult := MrCancel;
      Cancel := True;
      Parent := Dialog;
    End;
    If Dialog.ShowModal = MrOk Then
    Begin
      DateTime := IntlCal.Date;
      Result := True;
    End;
  Finally
    Dialog.Free;
  End;
End;

{ TCalendarSystemPropertyEditor }

Function TCalendarSystemPropertyEditor.GetAttributes: TPropertyAttributes;
Begin
  Result := [PaValueList, PaSortList, PaRevertable];
End;

Procedure TCalendarSystemPropertyEditor.GetValues(Proc: TGetStrProc);
Var
  I: Integer;
Begin
  For I := 0 To CalendarTypes.Count - 1 Do
    Proc(CalendarTypes.ByIndex(I).CalendarName);
End;

Function TCalendarSystemPropertyEditor.GetValue: String;
Var
  CalendarClass: TCalendarClass;
Begin
  CalendarClass := TCalendarClass(GetOrdValue);
  If Assigned(CalendarClass) Then
    Result := CalendarClass.CalendarName
  Else
    Result := '';
End;

Procedure TCalendarSystemPropertyEditor.SetValue(Const Value: String);
Var
  CalendarClass: TCalendarClass;
Begin
  CalendarClass := CalendarTypes.ByName(Trim(Value));
  If Assigned(CalendarClass) Then
    SetOrdValue(Integer(CalendarClass))
  Else
    SetOrdValue(0);
End;

{ TIntlDateTimePropertyEditor }

Function TIntlDateTimePropertyEditor.GetAttributes: TPropertyAttributes;
Begin
  Result := [PaDialog];
End;

Function TIntlDateTimePropertyEditor.GetValue: String;
Var
  DT: TDateTime;
Begin
  DT := GetFloatValue;
  If TCalendar.IsNoDate(DT) Then
    Result := ''
  Else
    Result := DefaultCalendar.Format('yyyy/mm/dd', DT);
End;

Procedure TIntlDateTimePropertyEditor.SetValue(Const Value: String);
Var
  DT: TDateTime;
Begin
  If Value = '' Then
    SetFloatValue(TCalendar.NoDate)
  Else If DefaultCalendar.TryParse('C', Value, DT) Then
    SetFloatValue(DT)
  Else
    Raise EConvertError.CreateResFmt(@SInvalidDateTime, [Value]);
End;

Procedure TIntlDateTimePropertyEditor.Edit;
Var
  DT: TDateTime;
  Caption: String;
Begin
  DT := GetFloatValue;
  Caption := GetComponent(0).GetNamePath + '.' + GetName;
  If PromptForDate(Caption, DT, TCalendarClass(GetObjectProp(GetComponent(0),
    'CalendarType'))) Then
    SetFloatValue(DT);
End;

{ TIntlDatePropertyEditor }

Function TIntlDatePropertyEditor.GetAttributes: TPropertyAttributes;
Begin
  Result := [PaDialog];
End;

Function TIntlDatePropertyEditor.GetValue: String;
Var
  DT: TDateTime;
Begin
  DT := GetFloatValue;
  If TCalendar.IsNoDate(DT) Then
    Result := ''
  Else
    Result := DefaultCalendar.Format('yyyy/mm/dd', Trunc(DT));
End;

Procedure TIntlDatePropertyEditor.SetValue(Const Value: String);
Var
  DT: TDateTime;
Begin
  If Value = '' Then
    SetFloatValue(TCalendar.NoDate)
  Else If DefaultCalendar.TryParse('C', Value, DT) Then
    SetFloatValue(Trunc(DT))
  Else
    Raise EConvertError.CreateResFmt(@SInvalidDate, [Value]);
End;

Procedure TIntlDatePropertyEditor.Edit;
Var
  DT: TDateTime;
  Caption: String;
Begin
  DT := GetFloatValue;
  Caption := GetComponent(0).GetNamePath + '.' + GetName;
  If PromptForDate(Caption, DT, TCalendarClass(GetObjectProp(GetComponent(0),
    'CalendarType'))) Then
    SetFloatValue(DT);
End;

{ TCalendarSelectionEditor }

Procedure TCalendarSelectionEditor.RequiresUnits(Proc: TGetStrProc);
Begin
  // Proc('NCore');
  Proc('NCalendar');
End;

{ TCalendarEventEditor }

Procedure TCalendarEventEditor.ExecuteVerb(Index: Integer);
Begin
  Case Index Of
    0:
      ShowCollectionEditor(Designer, Component, TIntlMonthCalendar(Component)
        .EventItems, 'EventItems');
  End;
End;

Function TCalendarEventEditor.GetVerb(Index: Integer): String;
Begin
  Case Index Of
    0:
      Result := 'Edit EventItems';
  End;
End;

Function TCalendarEventEditor.GetVerbCount: Integer;
Begin
  Result := 1;
End;

{ TStringDatePropEditor }

Function TStringDatePropEditor.GetAttributes: TPropertyAttributes;
Begin
  Result := [PaDialog];
End;

Procedure TStringDatePropEditor.Edit;
Var
  DT: TDateTime;
  Caption: String;
  Jc: TJalaliCalendar;
Begin
  Jc := TJalaliCalendar.Create;
  If Not Jc.TryParse('yyyy/mm/dd', Value, DT) Then
    DT := Jc.NoDate;
  Caption := GetComponent(0).GetNamePath + '.' + GetName;
  If PromptForDate(Caption, DT, TJalaliCalendar) Then
    SetValue(Jc.Format('yyyy/mm/dd', DT));
End;

End.
