Unit NCalendar;
{$I DELPHIAREA.INC}

Interface

Uses
  SysUtils, Classes;

Type

  TCalendar = Class;
  TCalendarSettings = Class;
  TCalendarClass = Class Of TCalendar;
  TCalendarSettingsClass = Class Of TCalendarSettings;
  TDayOfWeek = (Default, Monday, Tuesday, Wednesday, Thursday, Friday, Saturday, Sunday);
  TCalendarLang = (ClPersian, ClArabic, ClEnglish);
  // TCalendarDigit = (CdPersian, CdEnglish);
  TWeekRule = (WrFirstDayWeek, WrFirstFourDayWeek, WrFullWeek);
  TCalendarKind = (CkUnknown, CkSolar, CkLunar, CkLunisolar);

  TDateYMD = Packed Record
    Year: SmallInt;
    Month: Byte;
    Day: Byte;
  End;

  TCalendarSettings = Class Abstract(TPersistent)
  Private
    FLocale: String;
    FCalendarName: String;
    FEraNames: Array Of String;
    FShortEraNames: Array Of String;
    FMonthNames: Array Of String;
    FShortMonthNames: Array Of String;
    FGenitiveMonthNames: Array Of String;
    FDayNames: Array [TDayOfWeek] Of String;
    FShortDayNames: Array [TDayOfWeek] Of String;
    FShortestDayNames: Array [TDayOfWeek] Of String;
    FLongDateFormat: String;
    FShortDateFormat: String;
    FYearMonthFormat: String;
    FMonthDayFormat: String;
    FYearFormat: String;
    FTwoDigitYearMax: Word;
    FWeekRule: TWeekRule;
    FFirstDayOfWeek: TDayOfWeek;
    FDateSeparator: String;
    FTimeSeparator: String;
    FTimeAMString: String;
    FTimePMString: String;
    FShortTimeFormat: String;
    FLongTimeFormat: String;
    FUpdateCount: Integer;
    FUpdateAffected: Boolean;
    FOnChange: TNotifyEvent;
    FCalendarLang: TCalendarLang;
    // FCalendarDigit: TCalendarDigit;
    Procedure SetCalendarName(Const Value: String);
    Function GetEraCount: Integer; Inline;
    Function GetEraNames(Era: Integer): String; Inline;
    Procedure SetEraNames(Era: Integer; Const Value: String);
    Function GetShortEraNames(Era: Integer): String; Inline;
    Procedure SetShortEraNames(Era: Integer; Const Value: String);
    Function GetMonthCount: Integer; Inline;
    Function GetMonthNames(Month: Integer): String; Inline;
    Procedure SetMonthNames(Month: Integer; Const Value: String);
    Function GetShortMonthNames(Month: Integer): String; Inline;
    Procedure SetShortMonthNames(Month: Integer; Const Value: String);
    Function GetGenitiveMonthNames(Month: Integer): String; Inline;
    Procedure SetGenitiveMonthNames(Month: Integer; Const Value: String);
    Function GetDayNames(DoW: TDayOfWeek): String; Inline;
    Procedure SetDayNames(DoW: TDayOfWeek; Const Value: String);
    Function GetShortDayNames(DoW: TDayOfWeek): String; Inline;
    Procedure SetShortDayNames(DoW: TDayOfWeek; Const Value: String);
    Function GetShortestDayNames(DoW: TDayOfWeek): String; Inline;
    Procedure SetShortestDayNames(DoW: TDayOfWeek; Const Value: String);
    Procedure SetLongDateFormat(Const Value: String);
    Procedure SetShortDateFormat(Const Value: String);
    Procedure SetYearMonthFormat(Const Value: String);
    Procedure SetMonthDayFormat(Const Value: String);
    Procedure SetYearFormat(Const Value: String);
    Procedure SetTwoDigitYearMax(Value: Word);
    Procedure SetWeekRule(Value: TWeekRule);
    Procedure SetFirstDayOfWeek(Value: TDayOfWeek);
    Procedure SetDateSeparator(Const Value: String);
    Procedure SetTimeSeparator(Const Value: String);
    Procedure SetTimeAMString(Const Value: String);
    Procedure SetTimePMString(Const Value: String);
    Procedure SetShortTimeFormat(Const Value: String);
    Procedure SetLongTimeFormat(Const Value: String);
    Procedure SetCalendarLang(Const Value: TCalendarLang);
    // Procedure SetCalendarDigit(Const Value: TCalendarDigit);
  Protected
    Function ScanNames(Var S: PChar; Names: Array Of String): Integer;
    Function ScanEraNames(Var S: PChar): Integer; Inline;
    Function ScanShortEraNames(Var S: PChar): Integer; Inline;
    Function ScanMonthNames(Var S: PChar): Integer; Inline;
    Function ScanShortMonthNames(Var S: PChar): Integer; Inline;
    Function ScanGenitiveMonthNames(Var S: PChar): Integer; Inline;
    Function ScanDayNames(Var S: PChar): Integer; Inline;
    Function ScanShortDayNames(Var S: PChar): Integer; Inline;
    Function ScanShortestDayNames(Var S: PChar): Integer; Inline;
    Function ScanAMPM(Var S: PChar): Integer; Inline;
    Function SafeGetInfoStr(Const Locale: String; CalID: Cardinal; LType, CType: Cardinal;
      Const Default: String): String;
    Function GetMonthInfoIndex(Month: Integer): Integer; Virtual;
  Protected
    Procedure DoChange; Virtual;
    Procedure PrepareEraName(Era: Integer; Const Locale: String; CalendarID: Cardinal); Virtual;
    Procedure PrepareMonthNames(Const Locale: String; CalendarID: Cardinal); Virtual;
    Procedure PrepareDayNames(Const Locale: String; CalendarID: Cardinal); Virtual;
    Procedure PrepareDateFormats(Const Locale: String; CalendarID: Cardinal); Virtual;
    Procedure PrepareTimeFormats(Const Locale: String); Virtual;
    Procedure PrepareWeekParams(Const Locale: String); Virtual;
    Procedure PrepareSettings(Const Locale: String; CalendarID: Cardinal); Virtual;
    Procedure ReadSettings(Reader: TReader); Virtual;
    Procedure WriteSettings(Writer: TWriter); Virtual;
  Public
    Constructor Create; Overload; Virtual;
    Constructor Create(LocaleID: Cardinal); Overload;
    Constructor Create(Const Locale: String); Overload;
    Destructor Destroy; Override;
    Class Function CalendarClass: TCalendarClass; Virtual; Abstract;
    Procedure Prepare(LocaleID: Cardinal); Overload;
    Procedure Prepare(Const Locale: String); Overload;
    Procedure LoadFromStream(Stream: TStream);
    Procedure SaveToStream(Stream: TStream);
    Procedure Assign(Source: TPersistent); Override;
    Procedure BeginUpdate;
    Procedure EndUpdate;
    Property Locale: String Read FLocale;
    Property CalendarName: String Read FCalendarName Write SetCalendarName;
    Property EraCount: Integer Read GetEraCount;
    Property EraNames[Era: Integer]: String Read GetEraNames Write SetEraNames;
    Property ShortEraNames[Era: Integer]: String Read GetShortEraNames Write SetShortEraNames;
    Property MonthCount: Integer Read GetMonthCount;
    Property MonthNames[Month: Integer]: String Read GetMonthNames Write SetMonthNames;
    Property ShortMonthNames[Month: Integer]: String Read GetShortMonthNames Write SetShortMonthNames;
    Property GenitiveMonthNames[Month: Integer]: String Read GetGenitiveMonthNames Write SetGenitiveMonthNames;
    Property DayNames[DoW: TDayOfWeek]: String Read GetDayNames Write SetDayNames;
    Property ShortDayNames[DoW: TDayOfWeek]: String Read GetShortDayNames Write SetShortDayNames;
    Property ShortestDayNames[DoW: TDayOfWeek]: String Read GetShortestDayNames Write SetShortestDayNames;
  Published
    Property LongDateFormat: String Read FLongDateFormat Write SetLongDateFormat;
    Property ShortDateFormat: String Read FShortDateFormat Write SetShortDateFormat;
    Property YearMonthFormat: String Read FYearMonthFormat Write SetYearMonthFormat;
    Property MonthDayFormat: String Read FMonthDayFormat Write SetMonthDayFormat;
    Property YearFormat: String Read FYearFormat Write SetYearFormat;
    Property TwoDigitYearMax: Word Read FTwoDigitYearMax Write SetTwoDigitYearMax;
    Property WeekRule: TWeekRule Read FWeekRule Write SetWeekRule;
    Property FirstDayOfWeek: TDayOfWeek Read FFirstDayOfWeek Write SetFirstDayOfWeek;
    Property CalendarLang: TCalendarLang Read FCalendarLang Write SetCalendarLang;
    // Property CalendarDigit: TCalendarDigit Read FCalendarDigit Write SetCalendarDigit;
    Property DateSeparator: String Read FDateSeparator Write SetDateSeparator;
    Property TimeSeparator: String Read FTimeSeparator Write SetTimeSeparator;
    Property TimeAMString: String Read FTimeAMString Write SetTimeAMString;
    Property TimePMString: String Read FTimePMString Write SetTimePMString;
    Property ShortTimeFormat: String Read FShortTimeFormat Write SetShortTimeFormat;
    Property LongTimeFormat: String Read FLongTimeFormat Write SetLongTimeFormat;
    Property OnChange: TNotifyEvent Read FOnChange Write FOnChange;
  End;

  TCalendar = Class Abstract(TPersistent)
  Private
    FSettings: TCalendarSettings;
    FDefaultEra: Integer;
    FOnChange: TNotifyEvent;
    Procedure SetSettings(Value: TCalendarSettings); Inline;
    Procedure SetDefaultEra(Value: Integer);
    Procedure SettingsChanged(Sender: TObject);
  Protected
    Class Procedure EraError(Era: Integer);
    Class Procedure YearError(Era, Year: Integer);
    Class Procedure MonthError(Era, Year, Month: Integer);
    Class Procedure DayError(Era, Year, Month, Day: Integer);
    Class Procedure DayNumError(DayNum: Integer);
    Class Procedure DateError(Const DateTime: TDateTime);
    Class Procedure DateDayError(Era, Year, DayOfYear: Integer);
    Class Procedure DateWeekError(Era, Year, Week: Integer; DoW: TDayOfWeek);
    Class Procedure ConvertEraError(Year, FromEra, ToEra: Integer);
  Protected
    Class Procedure Divide(Numerator, Denominator: Integer; Out Quotient, Reminder: Integer); Inline;
    Class Function Modulus(Numerator, Denominator: Integer): Integer; Inline;
  Protected Const
    DATETIME_EPOCH = 2415018.5;
    Procedure DoChange; Virtual;
    Function ToJulianDay(Year, Month, Day: Integer): Extended; Virtual; Abstract;
    Function FromJulianDay(JD: Extended; Out Year, Month, Day: Integer): Boolean; Virtual; Abstract;
    Function AdjustJulianDay(JD: Extended; Backward: Boolean): Extended; Virtual;
    Function BaseEra: Integer; Virtual;
    Function BestEraOfBaseEraYear(Year: Integer): Integer; Virtual;
    Function ToZeroBase(Era, Year: Integer): Integer; Inline;
    Function FromZeroBase(Era, Year: Integer): Integer; Inline;
  Public Const
    DaysPerWeek = 7;

  Const
    OneDay = 1.0;

  Const
    OneHour = OneDay / HoursPerDay;

  Const
    OneMinute = OneDay / MinsPerDay;

  Const
    OneSecond = OneDay / SecsPerDay;

  Const
    OneMillisecond = OneDay / MSecsPerDay;

  Const
    OneWeek = OneDay * DaysPerWeek;

  Const
    NoDate: TDateTime = -0.5;
  Public
    Constructor Create; Overload; Virtual;
    Constructor Create(LocaleID: Integer); Overload;
    Constructor Create(Const Locale: String); Overload;
    Destructor Destroy; Override;
    Procedure Assign(Source: TPersistent); Override;
    Procedure AssignTo(Dest: TPersistent); Override;
    Function Clone: TCalendar; Virtual;
    Property Settings: TCalendarSettings Read FSettings Write SetSettings;
    Property DefaultEra: Integer Read FDefaultEra Write SetDefaultEra;
    Property OnChange: TNotifyEvent Read FOnChange Write FOnChange;
  Public
    Class Function CalendarName: String;
    Class Function CalendarID: Cardinal; Virtual;
    Class Function CalendarKind: TCalendarKind; Virtual;
    Class Function SettingsClass: TCalendarSettingsClass; Virtual;
    { Min/Max Functions }
    Class Function MaxEra: Integer; Virtual;
    Class Function MaxMonthsPerYear: Integer; Virtual;
    Class Function MinSupportedDateTime: TDateTime; Virtual;
    Class Function MaxSupportedDateTime: TDateTime; Virtual;
    { Miscellaneous Functions }
    Class Function Today: TDate; Static; Inline;
    Class Function Yesterday: TDate; Static; Inline;
    Class Function Tomorrow: TDate; Static; Inline;
    Class Function FirstMonthOfYear: Integer; Virtual;
    Function DaysToMonth(Era, Year, Month: Integer): Integer; Overload; Virtual;
    Function DaysToMonth(Year, Month: Integer): Integer; Overload; Inline;
    Function DaysToMonth(Const DateTime: TDateTime): Integer; Overload;
    Function DaysInMonth(Era, Year, Month: Integer): Integer; Overload; Virtual;
    Function DaysInMonth(Year, Month: Integer): Integer; Overload; Inline;
    Function DaysInMonth(Const DateTime: TDateTime): Integer; Overload;
    Function DaysInYear(Era, Year: Integer): Integer; Overload; Virtual;
    Function DaysInYear(Year: Integer): Integer; Overload; Inline;
    Function DaysInYear(Const DateTime: TDateTime): Integer; Overload;
    Function MonthsInYear(Era, Year: Integer): Integer; Overload; Virtual;
    Function MonthsInYear(Year: Integer): Integer; Overload; Inline;
    Function MonthsInYear(Const DateTime: TDateTime): Integer; Overload;
    Function WeeksInYear(Era, Year: Integer): Integer; Overload; Virtual;
    Function WeeksInYear(Year: Integer): Integer; Overload; Inline;
    Function WeeksInYear(Const DateTime: TDateTime): Integer; Overload;
    Function WeeksInMonth(Era, Year, Month: Integer): Integer; Overload; Inline;
    Function WeeksInMonth(Year, Month: Integer): Integer; Overload; Inline;
    Function WeeksInMonth(Const DateTime: TDateTime): Integer; Overload;
    Function DayOfYearToDayOfMonth(Era, Year, DayOfYear: Integer; Var Month, Day: Integer): Boolean;
      Overload; Virtual;
    Function DayOfYearToDayOfMonth(Year, DayOfYear: Integer; Var Month, Day: Integer): Boolean;
      Overload; Inline;
    Function DayOfMonthToDayOfYear(Era, Year, Month, Day: Integer; Out DayOfYear: Integer): Boolean;
      Overload; Virtual;
    Function DayOfMonthToDayOfYear(Year, Month, Day: Integer; Out DayOfYear: Integer): Boolean;
      Overload; Inline;
    Function ConvertYear(Year, FromEra, ToEra: Integer): Integer; Virtual;
    Function ToFourDigitYear(Year: Integer): Integer; Virtual;
    Function CanHaveZeroYear(Era: Integer): Boolean; Virtual;
    Function OffsetYear(Year, NumberOfYears: Integer; FromEra, ToEra: Integer): Integer; Overload; Inline;
    Function OffsetYear(Era, Year, NumberOfYears: Integer): Integer; Overload; Inline;
    Function OffsetYear(Year, NumberOfYears: Integer): Integer; Overload; Inline;
    Class Function OffsetDateTime(Const DateTime: TDateTime; Const Delta: Double): TDateTime; Static; Inline;
    Class Function DaysBetween(DateTime1, DateTime2: TDateTime): Integer; Static; Inline;
    Class Function DaysOfWeekBetween(DoW1, DoW2: TDayOfWeek): Integer; Static; Inline;
    Function DayOfWeekToDayNum(DoW: TDayOfWeek): Integer; Inline;
    Function DayNumToDayOfWeek(DayNum: Integer): TDayOfWeek;
    { Is Functions }
    Class Function IsNoDate(Const DateTime: TDateTime): Boolean; Static; Inline;
    Function IsValidDate(Const DateTime: TDateTime): Boolean; Virtual;
    Function IsValidDateDay(Era, Year, DayOfYear: Integer): Boolean; Overload; Virtual;
    Function IsValidDateDay(Year, DayOfYear: Integer): Boolean; Overload; Inline;
    Function IsValidDateWeek(Era, Year, Week: Integer; DoW: TDayOfWeek): Boolean; Overload; Virtual;
    Function IsValidDateWeek(Year, Week: Integer; DoW: TDayOfWeek): Boolean; Overload; Inline;
    Function IsValidEra(Era: Integer): Boolean; Virtual;
    Function IsValidYear(Era, Year: Integer): Boolean; Overload; Virtual;
    Function IsValidYear(Year: Integer): Boolean; Overload; Inline;
    Function IsValidMonth(Era, Year, Month: Integer): Boolean; Overload; Virtual;
    Function IsValidMonth(Year, Month: Integer): Boolean; Overload; Inline;
    Function IsValidDay(Era, Year, Month, Day: Integer): Boolean; Overload; Virtual;
    Function IsValidDay(Year, Month, Day: Integer): Boolean; Overload; Inline;
    Function IsLeapYear(Era, Year: Integer): Boolean; Overload; Virtual;
    Function IsLeapYear(Year: Integer): Boolean; Overload; Inline;
    Function IsInLeapYear(Const DateTime: TDateTime): Boolean;
    Class Function IsInRange(Const DateTime, RangeStart, RangeEnd: TDateTime): Boolean; Static; Inline;
    Function IsThisEra(Const DateTime: TDateTime): Boolean; Overload; Inline;
    Function IsThisEra(Era: Integer): Boolean; Overload; Inline;
    Function IsThisYear(Const DateTime: TDateTime): Boolean; Overload; Inline;
    Function IsThisYear(Era, Year: Integer): Boolean; Overload; Inline;
    Function IsThisYear(Year: Integer): Boolean; Overload; Inline;
    Function IsThisMonth(Const DateTime: TDateTime): Boolean; Overload; Inline;
    Function IsThisMonth(Era, Year, Month: Integer): Boolean; Overload; Inline;
    Function IsThisMonth(Year, Month: Integer): Boolean; Overload; Inline;
    Function IsThisMonth(Month: Integer): Boolean; Overload; Inline;
    Function IsThisWeek(Const DateTime: TDateTime): Boolean; Inline;
    Class Function IsToday(Const DateTime: TDateTime): Boolean; Overload; Static; Inline;
    Class Function IsToday(DoW: TDayOfWeek): Boolean; Overload; Static; Inline;
    Class Function IsPM(Const DateTime: TDateTime): Boolean; Static; Inline;
    Class Function IsSame(Const DateTime1, DateTime2: TDateTime): Boolean; Static; Inline;
    Function IsSameEra(Const DateTime1, DateTime2: TDateTime): Boolean; Inline;
    Function IsSameYear(Const DateTime1, DateTime2: TDateTime): Boolean;
    Function IsSameMonth(Const DateTime1, DateTime2: TDateTime): Boolean;
    Function IsSameWeek(Const DateTime1, DateTime2: TDateTime): Boolean;
    Class Function IsSameDay(Const DateTime1, DateTime2: TDateTime): Boolean; Static; Inline;
    Class Function Compare(Const DateTime1, DateTime2: TDateTime): Integer; Static;
    Class Function CompareDate(Const DateTime1, DateTime2: TDateTime): Integer; Overload; Static; Inline;
    Class Function CompareTime(Const DateTime1, DateTime2: TDateTime): Integer; Overload; Static; Inline;
    { Encode/Decode Functions }
    Function TryEncodeDateWeek(Era, Year, Week: Integer; DoW: TDayOfWeek; Out DateTime: TDateTime)
      : Boolean; Overload;
    Function TryEncodeDateWeek(Year, Week: Integer; DoW: TDayOfWeek; Out DateTime: TDateTime): Boolean;
      Overload; Inline;
    Function EncodeDateWeek(Era, Year, Week: Integer; DoW: TDayOfWeek): TDateTime; Overload;
    Function EncodeDateWeek(Year, Week: Integer; DoW: TDayOfWeek): TDateTime; Overload; Inline;
    Function TryDecodeDateWeek(Const DateTime: TDateTime; Out Era, Year, Week: Integer; Out DoW: TDayOfWeek)
      : Boolean; Overload;
    Function TryDecodeDateWeek(Const DateTime: TDateTime; Out Year, Week: Integer; Out DoW: TDayOfWeek)
      : Boolean; Overload; Inline;
    Procedure DecodeDateWeek(Const DateTime: TDateTime; Out Era, Year, Week: Integer;
      Out DoW: TDayOfWeek); Overload;
    Procedure DecodeDateWeek(Const DateTime: TDateTime; Out Year, Week: Integer; Out DoW: TDayOfWeek);
      Overload; Inline;
    Function TryEncodeDateDay(Era, Year, DayOfYear: Integer; Out DateTime: TDateTime): Boolean; Overload;
    Function TryEncodeDateDay(Year, DayOfYear: Integer; Out DateTime: TDateTime): Boolean; Overload; Inline;
    Function EncodeDateDay(Era, Year, DayOfYear: Integer): TDateTime; Overload;
    Function EncodeDateDay(Year, DayOfYear: Integer): TDateTime; Overload; Inline;
    Function TryDecodeDateDay(Const DateTime: TDateTime; Out Era, Year, DayOfYear: Integer): Boolean; Overload;
    Function TryDecodeDateDay(Const DateTime: TDateTime; Out Year, DayOfYear: Integer): Boolean;
      Overload; Inline;
    Procedure DecodeDateDay(Const DateTime: TDateTime; Out Era, Year, DayOfYear: Integer); Overload;
    Procedure DecodeDateDay(Const DateTime: TDateTime; Out Year, DayOfYear: Integer); Overload; Inline;
    Function TryEncodeDate(Era, Year, Month, Day: Integer; Out DateTime: TDateTime): Boolean; Overload;
    Function TryEncodeDate(Year, Month, Day: Integer; Out DateTime: TDateTime): Boolean; Overload; Inline;
    Function EncodeDate(Era, Year, Month, Day: Integer): TDateTime; Overload;
    Function EncodeDate(Year, Month, Day: Integer): TDateTime; Overload; Inline;
    Function TryDecodeDate(Const DateTime: TDateTime; Out Era, Year, Month, Day: Integer): Boolean;
      Overload; Inline;
    Function TryDecodeDate(Const DateTime: TDateTime; Out Year, Month, Day: Integer): Boolean; Overload; Inline;
    Procedure DecodeDate(Const DateTime: TDateTime; Out Era, Year, Month, Day: Integer); Overload;
    Procedure DecodeDate(Const DateTime: TDateTime; Out Year, Month, Day: Integer); Overload; Inline;
    Class Function DateTimeToJulianDay(Const DateTime: TDateTime): Extended; Static; Inline;
    Class Function JulianDayToDateTime(Const JulianDate: Extended): TDateTime; Static; Inline;
    Class Function CombineDateTime(Const Date: TDate; Const Time: TTime): TDateTime; Static; Inline;
    Class Function LocalToUniversalDateTime(Const DateTime: TDateTime): TDateTime; Static;
    Class Function UniversalToLocalDateTime(Const DateTime: TDateTime): TDateTime; Static;
    { Pick-a-Field Functions }
    Class Function DateOf(Const DateTime: TDateTime): TDate; Static; Inline;
    Function EraOf(Const DateTime: TDateTime): Integer; Inline;
    Function YearOf(Const DateTime: TDateTime): Integer; Inline;
    Function MonthOfYear(Const DateTime: TDateTime): Integer; Inline;
    Function WeekOfYear(Const DateTime: TDateTime): Integer; Inline;
    Function WeekOfMonth(Const DateTime: TDateTime): Integer; Inline;
    Function DayOfYear(Const DateTime: TDateTime): Integer; Inline;
    Function DayOfMonth(Const DateTime: TDateTime): Integer; Inline;
    Class Function DayOfWeek(Const DateTime: TDateTime): TDayOfWeek; Static; Inline;
    Class Function TimeOf(Const DateTime: TDateTime): TTime; Static; Inline;
    Class Function HourOfDay(Const DateTime: TDateTime): Integer; Static; Inline;
    Class Function MinuteOfDay(Const DateTime: TDateTime): Integer; Static; Inline;
    Class Function MinuteOfHour(Const DateTime: TDateTime): Integer; Static; Inline;
    Class Function SecondOfDay(Const DateTime: TDateTime): Integer; Static; Inline;
    Class Function SecondOfMinute(Const DateTime: TDateTime): Integer; Static; Inline;
    Class Function MillisecondOfDay(Const DateTime: TDateTime): Integer; Static; Inline;
    Class Function MillisecondOfSecond(Const DateTime: TDateTime): Integer; Static; Inline;
    { Start of a Field Functions }
    Function StartOfEra(Era: Integer): TDateTime; Virtual;
    Function StartOfYear(Era, Year: Integer): TDateTime; Overload; Inline;
    Function StartOfYear(Year: Integer): TDateTime; Overload; Inline;
    Function StartOfYear(Const DateTime: TDateTime): TDateTime; Overload; Inline;
    Function StartOfMonth(Era, Year, Month: Integer): TDateTime; Overload; Inline;
    Function StartOfMonth(Year, Month: Integer): TDateTime; Overload; Inline;
    Function StartOfMonth(Const DateTime: TDateTime): TDateTime; Overload; Inline;
    Function StartOfWeek(Const DateTime: TDateTime): TDateTime; Inline;
    Class Function StartOfDay(Const DateTime: TDateTime): TDateTime; Static; Inline;
    Class Function StartOfHour(Const DateTime: TDateTime): TDateTime; Static; Inline;
    Class Function StartOfMinute(Const DateTime: TDateTime): TDateTime; Static; Inline;
    Class Function StartOfSecond(Const DateTime: TDateTime): TDateTime; Static; Inline;
    { End of a Field Functions }
    Function EndOfEra(Era: Integer): TDateTime; Virtual;
    Function EndOfYear(Era, Year: Integer): TDateTime; Overload; Inline;
    Function EndOfYear(Year: Integer): TDateTime; Overload; Inline;
    Function EndOfYear(Const DateTime: TDateTime): TDateTime; Overload; Inline;
    Function EndOfMonth(Era, Year, Month: Integer): TDateTime; Overload; Inline;
    Function EndOfMonth(Year, Month: Integer): TDateTime; Overload; Inline;
    Function EndOfMonth(Const DateTime: TDateTime): TDateTime; Overload; Inline;
    Function EndOfWeek(Const DateTime: TDateTime): TDateTime; Inline;
    Class Function EndOfDay(Const DateTime: TDateTime): TDateTime; Static; Inline;
    Class Function EndOfHour(Const DateTime: TDateTime): TDateTime; Static; Inline;
    Class Function EndOfMinute(Const DateTime: TDateTime): TDateTime; Static; Inline;
    Class Function EndOfSecond(Const DateTime: TDateTime): TDateTime; Static; Inline;
    { Previous of a Field Functions }
    Function PrevYear(Const DateTime: TDateTime; NumberOfYears: Integer = 1): TDateTime; Inline;
    Function PrevMonth(Const DateTime: TDateTime; NumberOfMonths: Integer = 1): TDateTime; Inline;
    Class Function PrevWeek(Const DateTime: TDateTime; NumberOfWeeks: Integer = 1): TDateTime; Static; Inline;
    Class Function PrevDay(Const DateTime: TDateTime; NumberOfDays: Integer = 1): TDateTime; Static; Inline;
    Function PrevDayOfWeek(Const DateTime: TDateTime; DoW: TDayOfWeek): TDateTime;
    Class Function PrevHour(Const DateTime: TDateTime; NumberOfHours: Integer = 1): TDateTime; Static; Inline;
    Class Function PrevMinute(Const DateTime: TDateTime; NumberOfMinutes: Integer = 1): TDateTime;
      Static; Inline;
    Class Function PrevSecond(Const DateTime: TDateTime; NumberOfSeconds: Integer = 1): TDateTime;
      Static; Inline;
    Class Function PrevMillisecond(Const DateTime: TDateTime; NumberOfMilliseconds: Integer = 1): TDateTime;
      Static; Inline;
    { Next of a Field Functions }
    Function NextYear(Const DateTime: TDateTime; NumberOfYears: Integer = 1): TDateTime;
    Function NextMonth(Const DateTime: TDateTime; NumberOfMonths: Integer = 1): TDateTime;
    Class Function NextWeek(Const DateTime: TDateTime; NumberOfWeeks: Integer = 1): TDateTime; Static; Inline;
    Class Function NextDay(Const DateTime: TDateTime; NumberOfDays: Integer = 1): TDateTime; Static; Inline;
    Function NextDayOfWeek(Const DateTime: TDateTime; DoW: TDayOfWeek): TDateTime;
    Class Function NextHour(Const DateTime: TDateTime; NumberOfHours: Integer = 1): TDateTime; Static; Inline;
    Class Function NextMinute(Const DateTime: TDateTime; NumberOfMinutes: Integer = 1): TDateTime;
      Static; Inline;
    Class Function NextSecond(Const DateTime: TDateTime; NumberOfSeconds: Integer = 1): TDateTime;
      Static; Inline;
    Class Function NextMillisecond(Const DateTime: TDateTime; NumberOfMilliseconds: Integer = 1): TDateTime;
      Static; Inline;
    { Increment Field Functions }
    Procedure IncYear(Var DateTime: TDateTime; NumberOfYears: Integer = 1); Overload; Inline;
    Procedure IncYear(Var Date: TDate; NumberOfYears: Integer = 1); Overload; Inline;
    Procedure IncMonth(Var DateTime: TDateTime; NumberOfMonths: Integer = 1); Overload; Inline;
    Procedure IncMonth(Var Date: TDate; NumberOfMonths: Integer = 1); Overload; Inline;
    Class Procedure IncWeek(Var DateTime: TDateTime; NumberOfWeeks: Integer = 1); Overload; Static; Inline;
    Class Procedure IncWeek(Var Date: TDate; NumberOfWeeks: Integer = 1); Overload; Static; Inline;
    Class Procedure IncDay(Var DateTime: TDateTime; NumberOfDays: Integer = 1); Overload; Static; Inline;
    Class Procedure IncDay(Var Date: TDate; NumberOfDays: Integer = 1); Overload; Static; Inline;
    Class Procedure IncHour(Var DateTime: TDateTime; NumberOfHours: Integer = 1); Overload; Static; Inline;
    Class Procedure IncHour(Var Time: TTime; NumberOfHours: Integer = 1); Overload; Static; Inline;
    Class Procedure IncMinute(Var DateTime: TDateTime; NumberOfMinutes: Integer = 1); Overload; Static; Inline;
    Class Procedure IncMinute(Var Time: TTime; NumberOfMinutes: Integer = 1); Overload; Static; Inline;
    Class Procedure IncSecond(Var DateTime: TDateTime; NumberOfSeconds: Integer = 1); Overload; Static; Inline;
    Class Procedure IncSecond(Var Time: TTime; NumberOfSeconds: Integer = 1); Overload; Static; Inline;
    Class Procedure IncMillisecond(Var DateTime: TDateTime; NumberOfMilliseconds: Integer = 1); Overload;
      Static; Inline;
    Class Procedure IncMillisecond(Var Time: TTime; NumberOfMilliseconds: Integer = 1); Overload;
      Static; Inline;
    Class Procedure IncDateTime(Var DateTime: TDateTime; Const Delta: Double); Static; Inline;
    { Decrement Field procedures }
    Procedure DecYear(Var DateTime: TDateTime; NumberOfYears: Integer = 1); Overload; Inline;
    Procedure DecYear(Var Date: TDate; NumberOfYears: Integer = 1); Overload; Inline;
    Procedure DecMonth(Var DateTime: TDateTime; NumberOfMonths: Integer = 1); Overload; Inline;
    Procedure DecMonth(Var Date: TDate; NumberOfMonths: Integer = 1); Overload; Inline;
    Class Procedure DecWeek(Var DateTime: TDateTime; NumberOfWeeks: Integer = 1); Overload; Static; Inline;
    Class Procedure DecWeek(Var Date: TDate; NumberOfWeeks: Integer = 1); Overload; Static; Inline;
    Class Procedure DecDay(Var DateTime: TDateTime; NumberOfDays: Integer = 1); Overload; Static; Inline;
    Class Procedure DecDay(Var Date: TDate; NumberOfDays: Integer = 1); Overload; Static; Inline;
    Class Procedure DecHour(Var DateTime: TDateTime; NumberOfHours: Integer = 1); Overload; Static; Inline;
    Class Procedure DecHour(Var Time: TTime; NumberOfHours: Integer = 1); Overload; Static; Inline;
    Class Procedure DecMinute(Var DateTime: TDateTime; NumberOfMinutes: Integer = 1); Overload; Static; Inline;
    Class Procedure DecMinute(Var Time: TTime; NumberOfMinutes: Integer = 1); Overload; Static; Inline;
    Class Procedure DecSecond(Var DateTime: TDateTime; NumberOfSeconds: Integer = 1); Overload; Static; Inline;
    Class Procedure DecSecond(Var Time: TTime; NumberOfSeconds: Integer = 1); Overload; Static; Inline;
    Class Procedure DecMillisecond(Var DateTime: TDateTime; NumberOfMilliseconds: Integer = 1); Overload;
      Static; Inline;
    Class Procedure DecMillisecond(Var Time: TTime; NumberOfMilliseconds: Integer = 1); Overload;
      Static; Inline;
    Class Procedure DecDateTime(Var DateTime: TDateTime; Const Delta: Double); Static; Inline;
    { String Conversion Functions }
    Function Format(Const FmtStr: String; Const DateTime: TDateTime): String; Virtual;
    Function Parse(Const FmtStr: String; Const DateTimeStr: String): TDateTime; Virtual;
    Function TryParse(Const FmtStr: String; Const DateTimeStr: String; Var DateTime: TDateTime): Boolean;
  End;

  TCalendarClassList = Class(TList)
  Private
    Function GetItems(Index: Integer): TCalendarClass; Inline;
    Procedure SetItems(Index: Integer; Value: TCalendarClass);
  Public
    Function First: TCalendarClass; Inline;
    Function Last: TCalendarClass; Inline;
    Function Add(Item: TCalendarClass): Integer; Inline;
    Procedure Insert(Index: Integer; Item: TCalendarClass); Inline;
    Function Remove(Item: TCalendarClass): Integer; Inline;
{$IFDEF COMPILER2010_UP}
    Function RemoveItem(Item: TCalendarClass; Direction: TList.TDirection): Integer;
{$ENDIF}
    Function Extract(Item: TCalendarClass): TCalendarClass; Inline;
{$IFDEF COMPILER2010_UP}
    Function ExtractItem(Item: TCalendarClass; Direction: TList.TDirection): TCalendarClass; Inline;
{$ENDIF}
    Function IndexOf(Item: TCalendarClass): Integer; Inline;
{$IFDEF COMPILER2010_UP}
    Function IndexOfItem(Item: TCalendarClass; Direction: TList.TDirection): Integer; Inline;
{$ENDIF}
    Property Items[Index: Integer]: TCalendarClass Read GetItems Write SetItems; Default;
  End;

  CalendarTypes = Class
  Private
    Class Var CalendarList: TStringList;
    Class Procedure Cleanup;
  Protected Type
    TCalendarRegInfo = Class(TObject)
    Private
      FCalendarClass: TCalendarClass;
      FLocales: TStringList;
    Public
      Constructor Create(ACalendarClass: TCalendarClass);
      Destructor Destroy; Override;
      Procedure Associate(Const Locales: Array Of String);
      Procedure Dissociate(Const Locales: Array Of String);
      Property AssociatedLocales: TStringList Read FLocales;
      Property CalendarClass: TCalendarClass Read FCalendarClass;
    End;
  Public
    Class Procedure Register(CalendarClass: TCalendarClass); Overload; Static;
    Class Procedure Register(CalendarClass: TCalendarClass; Const Locales: Array Of String); Overload; Static;
    Class Procedure Unregister(CalendarClass: TCalendarClass); Overload; Static;
    Class Procedure Unregister(CalendarClass: TCalendarClass; Const Locales: Array Of String); Overload; Static;
    Class Function MinSupportedDateTime: TDateTime; Static;
    Class Function MaxSupportedDateTime: TDateTime; Static;
    Class Function Count: Integer; Static;
    Class Function ByIndex(Index: Integer): TCalendarClass; Inline; Static;
    Class Function ByID(ID: Cardinal): TCalendarClass; Static;
    Class Function ByName(Const Name: String): TCalendarClass; Static;
    Class Function ByLocale(Const Locale: String): TCalendarClass; Static;
    Class Procedure GetCalendersOfLocale(Const Locale: String; CalendarClasses: TCalendarClassList); Static;
  End;

Implementation

Uses
  Math, Types, SysConst, RTLConsts, Windows,NWinNLS, NUnicode;

Resourcestring
  SStreamReadError = 'Incompatible calendar settings';
  SInvalidEraError = '%d is not a valid era in %s';
  SInvalidYearError = '%d is not a valid year in %s (Era = %d)';
  SInvalidMonthError = '%d/%d is not a valid year/month in %s (Era = %d)';
  SInvalidDayError = '%d/%d/%d is not a valid year/month/day in %s (Era = %d)';
  SInvalidDayNumError = 'The numeric value of a day of the week (%d) must be in range 1 to 7.';
  SInvalidDateError = 'The Gregorian date (%s) is not supported by %s';
  SInvalidDateDayError = '(%d, %d) is not a valid DateDay pair in %s (Era = %d)';
  SInvalidDateWeekError = '(%d, %d, %d) is not a valid DateWeek triplet in %s (Era = %d)';
  SInvalidConvertEraError = 'Year %d cannot be converted from era %d to era %d in %s';

  { TCalendarSettings }

Constructor TCalendarSettings.Create;
Begin
  SetLength(FEraNames, CalendarClass.MaxEra);
  SetLength(FShortEraNames, CalendarClass.MaxEra);
  SetLength(FMonthNames, CalendarClass.MaxMonthsPerYear);
  SetLength(FShortMonthNames, CalendarClass.MaxMonthsPerYear);
  SetLength(FGenitiveMonthNames, CalendarClass.MaxMonthsPerYear);
End;

Constructor TCalendarSettings.Create(LocaleID: Cardinal);
Begin
  Create;
  Prepare(LocaleID);
End;

Constructor TCalendarSettings.Create(Const Locale: String);
Begin
  Create;
  Prepare(Locale);
End;

Destructor TCalendarSettings.Destroy;
Begin
  FEraNames := Nil;
  FShortEraNames := Nil;
  FMonthNames := Nil;
  FShortMonthNames := Nil;
  FGenitiveMonthNames := Nil;
  Inherited Destroy;
End;

Procedure TCalendarSettings.SetCalendarName(Const Value: String);
Begin
  If CalendarName <> Value Then
  Begin
    FCalendarName := Value;
    DoChange;
  End;
End;

Function TCalendarSettings.GetEraCount: Integer;
Begin
  Result := Length(FEraNames);
End;

Function TCalendarSettings.GetEraNames(Era: Integer): String;
Begin
  Result := FEraNames[Era - 1];
End;

Procedure TCalendarSettings.SetEraNames(Era: Integer; Const Value: String);
Begin
  If EraNames[Era] <> Value Then
  Begin
    FEraNames[Era - 1] := Value;
    DoChange;
  End;
End;

Function TCalendarSettings.GetShortEraNames(Era: Integer): String;
Begin
  Result := FShortEraNames[Era - 1];
End;

Procedure TCalendarSettings.SetShortEraNames(Era: Integer; Const Value: String);
Begin
  If ShortEraNames[Era] <> Value Then
  Begin
    FShortEraNames[Era - 1] := Value;
    DoChange;
  End;
End;

Function TCalendarSettings.GetMonthCount: Integer;
Begin
  Result := Length(FMonthNames);
End;

Function TCalendarSettings.GetMonthNames(Month: Integer): String;
Begin
  Result := FMonthNames[Month - 1];
End;

Procedure TCalendarSettings.SetMonthNames(Month: Integer; Const Value: String);
Begin
  If MonthNames[Month] <> Value Then
  Begin
    FMonthNames[Month - 1] := Value;
    DoChange;
  End;
End;

// Procedure TCalendarSettings.SetCalendarDigit(Const Value: TCalendarDigit);
// Begin
// If FCalendarDigit <> Value Then
// Begin
// FCalendarDigit := Value;
// DoChange;
// End;
// End;

Procedure TCalendarSettings.SetCalendarLang(Const Value: TCalendarLang);
Begin
  If FCalendarLang <> Value Then
  Begin
    FCalendarLang := Value;
    PrepareSettings(FLocale, 0);
    // PrepareMonthNames(FLocale, 0);
    // PrepareDayNames(FLocale, 0);

    DoChange;
  End;
End;

Function TCalendarSettings.GetShortMonthNames(Month: Integer): String;
Begin
  Result := FShortMonthNames[Month - 1];
End;

Procedure TCalendarSettings.SetShortMonthNames(Month: Integer; Const Value: String);
Begin
  If ShortMonthNames[Month] <> Value Then
  Begin
    FShortMonthNames[Month - 1] := Value;
    DoChange;
  End;
End;

Function TCalendarSettings.GetGenitiveMonthNames(Month: Integer): String;
Begin
  Result := FGenitiveMonthNames[Month - 1];
End;

Procedure TCalendarSettings.SetGenitiveMonthNames(Month: Integer; Const Value: String);
Begin
  If GenitiveMonthNames[Month] <> Value Then
  Begin
    FGenitiveMonthNames[Month - 1] := Value;
    DoChange;
  End;
End;

Function TCalendarSettings.GetDayNames(DoW: TDayOfWeek): String;
Begin
  Result := FDayNames[DoW];
End;

Procedure TCalendarSettings.SetDayNames(DoW: TDayOfWeek; Const Value: String);
Begin
  If DayNames[DoW] <> Value Then
  Begin
    FDayNames[DoW] := Value;
    DoChange;
  End;
End;

Function TCalendarSettings.GetShortDayNames(DoW: TDayOfWeek): String;
Begin
  Result := FShortDayNames[DoW];
End;

Procedure TCalendarSettings.SetShortDayNames(DoW: TDayOfWeek; Const Value: String);
Begin
  If ShortDayNames[DoW] <> Value Then
  Begin
    FShortDayNames[DoW] := Value;
    DoChange;
  End;
End;

Function TCalendarSettings.GetShortestDayNames(DoW: TDayOfWeek): String;
Begin
  Result := FShortestDayNames[DoW];
End;

Procedure TCalendarSettings.SetShortestDayNames(DoW: TDayOfWeek; Const Value: String);
Begin
  If ShortestDayNames[DoW] <> Value Then
  Begin
    FShortestDayNames[DoW] := Value;
    DoChange;
  End;
End;

Procedure TCalendarSettings.SetLongDateFormat(Const Value: String);
Begin
  If LongDateFormat <> Value Then
  Begin
    FLongDateFormat := Value;
    DoChange;
  End;
End;

Procedure TCalendarSettings.SetShortDateFormat(Const Value: String);
Begin
  If ShortDateFormat <> Value Then
  Begin
    FShortDateFormat := Value;
    DoChange;
  End;
End;

Procedure TCalendarSettings.SetYearMonthFormat(Const Value: String);
Begin
  If YearMonthFormat <> Value Then
  Begin
    FYearMonthFormat := Value;
    DoChange;
  End;
End;

Procedure TCalendarSettings.SetMonthDayFormat(Const Value: String);
Begin
  If MonthDayFormat <> Value Then
  Begin
    FMonthDayFormat := Value;
    DoChange;
  End;
End;

Procedure TCalendarSettings.SetYearFormat(Const Value: String);
Begin
  If YearFormat <> Value Then
  Begin
    FYearFormat := Value;
    DoChange;
  End;
End;

Procedure TCalendarSettings.SetTwoDigitYearMax(Value: Word);
Begin
  If TwoDigitYearMax <> Value Then
  Begin
    FTwoDigitYearMax := Value;
    DoChange;
  End;
End;

Procedure TCalendarSettings.SetWeekRule(Value: TWeekRule);
Begin
  If WeekRule <> Value Then
  Begin
    FWeekRule := Value;
    DoChange;
  End;
End;

Procedure TCalendarSettings.SetFirstDayOfWeek(Value: TDayOfWeek);
Begin
  If FirstDayOfWeek <> Value Then
  Begin
    If Value = Default Then
      Value := Saturday;
    FFirstDayOfWeek := Value;
    DoChange;
  End;
End;

Procedure TCalendarSettings.SetDateSeparator(Const Value: String);
Begin
  If DateSeparator <> Value Then
  Begin
    FDateSeparator := Value;
    DoChange;
  End;
End;

Procedure TCalendarSettings.SetTimeSeparator(Const Value: String);
Begin
  If TimeSeparator <> Value Then
  Begin
    FTimeSeparator := Value;
    DoChange;
  End;
End;

Procedure TCalendarSettings.SetTimeAMString(Const Value: String);
Begin
  If TimeAMString <> Value Then
  Begin
    FTimeAMString := Value;
    DoChange;
  End;
End;

Procedure TCalendarSettings.SetTimePMString(Const Value: String);
Begin
  If FTimePMString <> Value Then
  Begin
    FTimePMString := Value;
    DoChange;
  End;
End;

Procedure TCalendarSettings.SetShortTimeFormat(Const Value: String);
Begin
  If ShortTimeFormat <> Value Then
  Begin
    FShortTimeFormat := Value;
    DoChange;
  End;
End;

Procedure TCalendarSettings.SetLongTimeFormat(Const Value: String);
Begin
  If LongTimeFormat <> Value Then
  Begin
    FLongTimeFormat := Value;
    DoChange;
  End;
End;

Function TCalendarSettings.ScanNames(Var S: PChar; Names: Array Of String): Integer;
Var
  I, Len, MaxMatchedLen: Integer;
Begin
  Result := 0;
  MaxMatchedLen := 0;
  For I := Low(Names) To High(Names) Do
  Begin
    Len := Length(Names[I]);
    If (AnsiStrLIComp(S, PChar(Names[I]), Len) = 0) Then
    Begin
      If Len > MaxMatchedLen Then
      Begin
        MaxMatchedLen := Len;
        Result := I + 1;
      End;
    End;
  End;
  Inc(S, MaxMatchedLen);
End;

Function TCalendarSettings.ScanEraNames(Var S: PChar): Integer;
Begin
  Result := ScanNames(S, FEraNames);
End;

Function TCalendarSettings.ScanShortEraNames(Var S: PChar): Integer;
Begin
  Result := ScanNames(S, FShortEraNames);
End;

Function TCalendarSettings.ScanMonthNames(Var S: PChar): Integer;
Begin
  Result := ScanNames(S, FMonthNames);
End;

Function TCalendarSettings.ScanShortMonthNames(Var S: PChar): Integer;
Begin
  Result := ScanNames(S, FShortMonthNames);
End;

Function TCalendarSettings.ScanGenitiveMonthNames(Var S: PChar): Integer;
Begin
  Result := ScanNames(S, FGenitiveMonthNames);
End;

Function TCalendarSettings.ScanDayNames(Var S: PChar): Integer;
Begin
  Result := ScanNames(S, FDayNames);
End;

Function TCalendarSettings.ScanShortDayNames(Var S: PChar): Integer;
Begin
  Result := ScanNames(S, FShortDayNames);
End;

Function TCalendarSettings.ScanShortestDayNames(Var S: PChar): Integer;
Begin
  Result := ScanNames(S, FShortestDayNames);
End;

Function TCalendarSettings.ScanAMPM(Var S: PChar): Integer;
Begin
  Result := ScanNames(S, [TimeAMString, TimePMString]);
End;

Function TCalendarSettings.SafeGetInfoStr(Const Locale: String; CalID: Cardinal; LType, CType: Cardinal;
  Const Default: String): String;
Begin
  If CalID <> 0 Then
  Begin
    Result := GetCalendarStr(Locale, CalID, CType, '');
    If Result <> '' Then
      Exit;
  End;
  Result := GetLocaleStr(Locale, LType, Default);
End;

Function TCalendarSettings.GetMonthInfoIndex(Month: Integer): Integer;
Begin
  Result := Month - 1;
End;

Procedure TCalendarSettings.DoChange;
Begin
  If FUpdateCount <> 0 Then
    FUpdateAffected := True
  Else If Assigned(OnChange) Then
    OnChange(Self);
End;

Procedure TCalendarSettings.Prepare(LocaleID: Cardinal);
Begin
  Prepare(LCIDToLocaleName(LocaleID));
End;

Procedure TCalendarSettings.Prepare(Const Locale: String);
Begin
  FLocale := Locale;
  BeginUpdate;
  Try
    PrepareSettings(Locale, CalendarClass.CalendarID);
  Finally
    EndUpdate;
  End;
End;

Procedure TCalendarSettings.LoadFromStream(Stream: TStream);
Var
  Reader: TReader;
  SourceClassName: String;
Begin
  BeginUpdate;
  Try
    Reader := TReader.Create(Stream, 1024);
    Try
      SourceClassName := Reader.ReadString;
      If Not SameText(SourceClassName, CalendarClass.ClassName) Then
        EStreamError.CreateRes(@SStreamReadError);
      FLocale := Reader.ReadString;
      ReadSettings(Reader);
    Finally
      Reader.Free;
    End;
  Finally
    EndUpdate;
  End;
End;

Procedure TCalendarSettings.SaveToStream(Stream: TStream);
Var
  Writer: TWriter;
Begin
  Writer := TWriter.Create(Stream, 1024);
  Try
    Writer.WriteString(CalendarClass.ClassName);
    Writer.WriteString(Locale);
    WriteSettings(Writer);
    Writer.FlushBuffer;
  Finally
    Writer.Free;
  End;
End;

Procedure TCalendarSettings.BeginUpdate;
Begin
  If FUpdateCount = 0 Then
    FUpdateAffected := False;
  Inc(FUpdateCount);
End;

Procedure TCalendarSettings.EndUpdate;
Begin
  Dec(FUpdateCount);
  If (FUpdateCount = 0) And FUpdateAffected Then
    DoChange;
End;

Procedure TCalendarSettings.Assign(Source: TPersistent);
Var
  I: Integer;
  Day: TDayOfWeek;
Begin
  If (Source Is TCalendarSettings) And (CalendarClass = TCalendarSettings(Source).CalendarClass) Then
  Begin
    BeginUpdate;
    Try
      WeekRule := TCalendarSettings(Source).WeekRule;
      FirstDayOfWeek := TCalendarSettings(Source).FirstDayOfWeek;
      TwoDigitYearMax := TCalendarSettings(Source).TwoDigitYearMax;
      CalendarName := TCalendarSettings(Source).CalendarName;
      For I := 1 To EraCount Do
      Begin
        EraNames[I] := TCalendarSettings(Source).EraNames[I];
        ShortEraNames[I] := TCalendarSettings(Source).ShortEraNames[I];
      End;
      For I := 1 To MonthCount Do
      Begin
        MonthNames[I] := TCalendarSettings(Source).MonthNames[I];
        ShortMonthNames[I] := TCalendarSettings(Source).ShortMonthNames[I];
        GenitiveMonthNames[I] := TCalendarSettings(Source).GenitiveMonthNames[I];
      End;
      For Day := Low(TDayOfWeek) To High(TDayOfWeek) Do
      Begin
        DayNames[Day] := TCalendarSettings(Source).DayNames[Day];
        ShortDayNames[Day] := TCalendarSettings(Source).ShortDayNames[Day];
        ShortestDayNames[Day] := TCalendarSettings(Source).ShortestDayNames[Day];
      End;
      LongDateFormat := TCalendarSettings(Source).LongDateFormat;
      ShortDateFormat := TCalendarSettings(Source).ShortDateFormat;
      YearMonthFormat := TCalendarSettings(Source).YearMonthFormat;
      MonthDayFormat := TCalendarSettings(Source).MonthDayFormat;
      YearFormat := TCalendarSettings(Source).YearFormat;
      ShortTimeFormat := TCalendarSettings(Source).ShortTimeFormat;
      LongTimeFormat := TCalendarSettings(Source).LongTimeFormat;
      DateSeparator := TCalendarSettings(Source).DateSeparator;
      TimeSeparator := TCalendarSettings(Source).TimeSeparator;
      TimeAMString := TCalendarSettings(Source).TimeAMString;
      TimePMString := TCalendarSettings(Source).TimePMString;
    Finally
      EndUpdate;
    End;
  End
  Else
    Inherited Assign(Source);
End;

Procedure TCalendarSettings.PrepareEraName(Era: Integer; Const Locale: String; CalendarID: Cardinal);
Var
  Temp: String;
Begin
  Temp := GetCalendarStr(Locale, CalendarID, CAL_SERASTRING, '');
  SetLength(Temp, StrLen(PChar(Temp)));
  EraNames[Era] := Temp;
  Temp := GetCalendarStr(Locale, CalendarID, CAL_SABBREVERASTRING, '');
  SetLength(Temp, StrLen(PChar(Temp)));
  ShortEraNames[Era] := Temp;
End;

Procedure TCalendarSettings.PrepareMonthNames(Const Locale: String; CalendarID: Cardinal);
Var
  M, I: Integer;
Begin
  For M := 1 To MonthCount Do
  Begin
    I := GetMonthInfoIndex(M);
    MonthNames[M] := SafeGetInfoStr(Locale, CalendarID, LOCALE_SMONTHNAME1 + I, CAL_SMONTHNAME1 + I, '');
    ShortMonthNames[M] := SafeGetInfoStr(Locale, CalendarID, LOCALE_SABBREVMONTHNAME1 + I,
      CAL_SABBREVMONTHNAME1 + I, '');
    GenitiveMonthNames[M] := SafeGetInfoStr(Locale, CalendarID, LOCALE_RETURN_GENITIVE_NAMES Or CAL_SMONTHNAME1
      + I, CAL_RETURN_GENITIVE_NAMES Or CAL_SMONTHNAME1 + I, MonthNames[M]);
  End;
End;

Procedure TCalendarSettings.PrepareDayNames(Const Locale: String; CalendarID: Cardinal);
Var
  Day: TDayOfWeek;
  I: Integer;
Begin
  For Day := Low(TDayOfWeek) To High(TDayOfWeek) Do
  Begin
    I := Ord(Day) - Ord(Monday);
    DayNames[Day] := SafeGetInfoStr(Locale, CalendarID, LOCALE_SDAYNAME1 + I, CAL_SDAYNAME1 + I, '');
    ShortDayNames[Day] := SafeGetInfoStr(Locale, CalendarID, LOCALE_SABBREVDAYNAME1 + I,
      CAL_SABBREVDAYNAME1 + I, '');
    ShortestDayNames[Day] := SafeGetInfoStr(Locale, CalendarID, LOCALE_SSHORTESTDAYNAME1 + I,
      CAL_SSHORTESTDAYNAME1 + I, ShortDayNames[Day]);
  End;
End;

Procedure TCalendarSettings.PrepareDateFormats(Const Locale: String; CalendarID: Cardinal);
Const
  ArabicComma = #$60C;
Var
  TranslateComma: Boolean;
  Format: String;
Begin
  TranslateComma := (Pos('Arab;', GetLocaleScripts(Locale)) <> 0);
  Format := SafeGetInfoStr(Locale, CalendarID, LOCALE_SLONGDATE, CAL_SLONGDATE, 'dddd, MMMM d, y');
  If TranslateComma Then
    LongDateFormat := StringReplace(Format, ',', ArabicComma, [RfReplaceAll])
  Else
    LongDateFormat := Format;
  Format := SafeGetInfoStr(Locale, CalendarID, LOCALE_SSHORTDATE, CAL_SSHORTDATE, 'dd/MM/yyyy');
  If TranslateComma Then
    ShortDateFormat := StringReplace(Format, ',', ArabicComma, [RfReplaceAll])
  Else
    ShortDateFormat := Format;
  Format := SafeGetInfoStr(Locale, CalendarID, LOCALE_SYEARMONTH, CAL_SYEARMONTH, 'MMMM, y');
  If TranslateComma Then
    YearMonthFormat := StringReplace(Format, ',', ArabicComma, [RfReplaceAll])
  Else
    YearMonthFormat := Format;
  Format := SafeGetInfoStr(Locale, CalendarID, LOCALE_SMONTHDAY, CAL_SMONTHDAY, 'd MMMM');
  If TranslateComma Then
    MonthDayFormat := StringReplace(Format, ',', ArabicComma, [RfReplaceAll])
  Else
    MonthDayFormat := Format;
  YearFormat := 'y';
End;

Procedure TCalendarSettings.PrepareTimeFormats(Const Locale: String);
Var
  TimePrefix, TimePostfix: String;
  HourFmt: String;
Begin
  TimePrefix := '';
  TimePostfix := '';
  If GetLocaleInt(Locale, LOCALE_ITIME, 0) = 0 Then
  Begin
    If GetLocaleInt(Locale, LOCALE_ITIMEMARKPOSN, 0) = 0 Then
      TimePostfix := ' AMPM'
    Else
      TimePrefix := 'AMPM ';
  End;
  If GetLocaleInt(Locale, LOCALE_ITLZERO, 0) = 0 Then
    HourFmt := 'h'
  Else
    HourFmt := 'hh';
  DateSeparator := GetLocaleStr(Locale, LOCALE_SDATE, '/');
  TimeSeparator := GetLocaleStr(Locale, LOCALE_STIME, ':');
  TimeAMString := GetLocaleStr(Locale, LOCALE_S1159, 'am');
  TimePMString := GetLocaleStr(Locale, LOCALE_S2359, 'pm');
  ShortTimeFormat := GetLocaleStr(Locale, LOCALE_SSHORTTIME, TimePrefix + HourFmt + ':nn' + TimePostfix);
  ShortTimeFormat := StringReplace(ShortTimeFormat, 'tt', 'ampm', [RfReplaceAll]);
  LongTimeFormat := GetLocaleStr(Locale, LOCALE_STIMEFORMAT, TimePrefix + HourFmt + ':nn:ss' + TimePostfix);
  LongTimeFormat := StringReplace(LongTimeFormat, 'tt', 'ampm', [RfReplaceAll]);
End;

Procedure TCalendarSettings.PrepareWeekParams(Const Locale: String);
Const
  FullWeekLocales: Array [1 .. 2] Of String = ('et-EE', 'ur-PK');
  FirstFourWeekLocales: Array [1 .. 34] Of String = ('de-AT', 'fr-BE', 'nl-BE', 'cs-CZ', 'da-DK', 'et-EE',
    'fo-FO', 'fi-FI', 'sv-FI', 'mk-MK', 'fr-FR', 'de-DE', 'el-GR', 'is-IS', 'en-IE', 'it-IT', 'lv-LV', 'de-LI',
    'de-LU', 'fr-LU', 'nl-NL', 'nb-NO', 'nn-NO', 'pl-PL', 'pt-PT', 'ca-ES', 'es-ES', 'es-ES_tradnl', 'eu-ES',
    'sv-SE', 'de-CH', 'it-CH', 'fr-CH', 'en-GB');

  Function IsMemberOf(Const Locales: Array Of String): Boolean;
  Var
    I: Integer;
  Begin
    For I := Low(Locales) To High(Locales) Do
      If SameText(Locale, Locales[I]) Then
      Begin
        Result := True;
        Exit;
      End;
    Result := False;
  End;

Begin
  FirstDayOfWeek := TDayOfWeek(GetLocaleInt(Locale, LOCALE_IFIRSTDAYOFWEEK, 0) + 1);
  If IsMemberOf(FullWeekLocales) Then
    WeekRule := WrFullWeek
  Else If IsMemberOf(FirstFourWeekLocales) Then
    WeekRule := WrFirstFourDayWeek
  Else
    WeekRule := WrFirstDayWeek;
End;

Procedure TCalendarSettings.PrepareSettings(Const Locale: String; CalendarID: Cardinal);
Var
  Era: Integer;
Begin
  CalendarName := GetCalendarStr(Locale, CalendarID, CAL_SCALNAME, CalendarClass.CalendarName);
  For Era := 1 To EraCount Do
    PrepareEraName(Era, Locale, CalendarID);
  PrepareMonthNames(Locale, CalendarID);
  PrepareDayNames(Locale, CalendarID);
  PrepareDateFormats(Locale, CalendarID);
  PrepareDateFormats(Locale, CalendarID);
  PrepareTimeFormats(Locale);
  PrepareWeekParams(Locale);
  TwoDigitYearMax := GetCalendarInt(Locale, CalendarID, CAL_ITWODIGITYEARMAX, 99);
End;

Procedure TCalendarSettings.ReadSettings(Reader: TReader);
Var
  I: Integer;
  Day: TDayOfWeek;
Begin
  WeekRule := TWeekRule(Reader.ReadInteger);
  FirstDayOfWeek := TDayOfWeek(Reader.ReadInteger);
  TwoDigitYearMax := Reader.ReadInteger;
  CalendarName := Reader.ReadString;
  For I := 1 To EraCount Do
  Begin
    EraNames[I] := Reader.ReadString;
    ShortEraNames[I] := Reader.ReadString;
  End;
  For I := 1 To MonthCount Do
  Begin
    MonthNames[I] := Reader.ReadString;
    ShortMonthNames[I] := Reader.ReadString;
    GenitiveMonthNames[I] := Reader.ReadString;
  End;
  For Day := Low(TDayOfWeek) To High(TDayOfWeek) Do
  Begin
    DayNames[Day] := Reader.ReadString;
    ShortDayNames[Day] := Reader.ReadString;
    ShortestDayNames[Day] := Reader.ReadString;
  End;
  LongDateFormat := Reader.ReadString;
  ShortDateFormat := Reader.ReadString;
  YearMonthFormat := Reader.ReadString;
  MonthDayFormat := Reader.ReadString;
  YearFormat := Reader.ReadString;
  ShortTimeFormat := Reader.ReadString;
  LongTimeFormat := Reader.ReadString;
  DateSeparator := Reader.ReadString;
  TimeSeparator := Reader.ReadString;
  TimeAMString := Reader.ReadString;
  TimePMString := Reader.ReadString;
End;

Procedure TCalendarSettings.WriteSettings(Writer: TWriter);
Var
  I: Integer;
  Day: TDayOfWeek;
Begin
  Writer.WriteInteger(Ord(WeekRule));
  Writer.WriteInteger(Ord(FirstDayOfWeek));
  Writer.WriteInteger(TwoDigitYearMax);
  Writer.WriteString(CalendarName);
  For I := 1 To EraCount Do
  Begin
    Writer.WriteString(EraNames[I]);
    Writer.WriteString(ShortEraNames[I]);
  End;
  For I := 1 To MonthCount Do
  Begin
    Writer.WriteString(MonthNames[I]);
    Writer.WriteString(ShortMonthNames[I]);
    Writer.WriteString(GenitiveMonthNames[I]);
  End;
  For Day := Low(TDayOfWeek) To High(TDayOfWeek) Do
  Begin
    Writer.WriteString(DayNames[Day]);
    Writer.WriteString(ShortDayNames[Day]);
    Writer.WriteString(ShortestDayNames[Day]);
  End;
  Writer.WriteString(LongDateFormat);
  Writer.WriteString(ShortDateFormat);
  Writer.WriteString(YearMonthFormat);
  Writer.WriteString(MonthDayFormat);
  Writer.WriteString(YearFormat);
  Writer.WriteString(ShortTimeFormat);
  Writer.WriteString(LongTimeFormat);
  Writer.WriteString(DateSeparator);
  Writer.WriteString(TimeSeparator);
  Writer.WriteString(TimeAMString);
  Writer.WriteString(TimePMString);
End;

{ TCalendar }

Constructor TCalendar.Create;
Begin
  FSettings := SettingsClass.Create;
  FSettings.DateSeparator := '/';
  FSettings.OnChange := SettingsChanged;
  FDefaultEra := MaxEra;
End;

Constructor TCalendar.Create(LocaleID: Integer);
Begin
  Create;
  Settings.Prepare(LocaleID);
End;

Constructor TCalendar.Create(Const Locale: String);
Begin
  Create;
  Settings.Prepare(Locale);
End;

Destructor TCalendar.Destroy;
Begin
  FSettings.Free;
  Inherited Destroy;
End;

Procedure TCalendar.SettingsChanged(Sender: TObject);
Begin
  DoChange;
End;

Procedure TCalendar.Assign(Source: TPersistent);
Begin
  If Source Is TCalendar Then
  Begin
    DefaultEra := TCalendar(Source).DefaultEra;
    Settings.Assign(TCalendar(Source).Settings);
  End
  Else If Source Is TCalendarSettings Then
    Settings.Assign(Source)
  Else
    Inherited Assign(Source);
End;

Procedure TCalendar.AssignTo(Dest: TPersistent);
Begin
  If Dest Is TCalendar Then
    TCalendar(Dest).Assign(Self)
  Else If Dest Is TCalendarSettings Then
    TCalendarSettings(Dest).Assign(Self.Settings)
  Else
    Inherited AssignTo(Dest);
End;

Function TCalendar.Clone: TCalendar;
Begin
  Result := TCalendarClass(ClassType).Create;
  Result.Assign(Self);
End;

Procedure TCalendar.SetSettings(Value: TCalendarSettings);
Begin
  Settings.Assign(Value);
End;

Procedure TCalendar.SetDefaultEra(Value: Integer);
Begin
  If DefaultEra <> Value Then
  Begin
    If Not IsValidEra(Value) Then
      EraError(Value);
    FDefaultEra := Value;
    DoChange;
  End;
End;

Class Function TCalendar.SettingsClass: TCalendarSettingsClass;
Begin
  Result := TCalendarSettings;
End;

Procedure TCalendar.DoChange;
Begin
  If Assigned(OnChange) Then
    OnChange(Self);
End;

Class Procedure TCalendar.EraError(Era: Integer);
Begin
  Raise EEncodingError.CreateResFmt(@SInvalidEraError, [Era, CalendarName]);
End;

Class Procedure TCalendar.YearError(Era, Year: Integer);
Begin
  Raise EEncodingError.CreateResFmt(@SInvalidYearError, [Year, CalendarName, Era]);
End;

Class Procedure TCalendar.MonthError(Era, Year, Month: Integer);
Begin
  Raise EEncodingError.CreateResFmt(@SInvalidMonthError, [Year, Month, CalendarName, Era]);
End;

Class Procedure TCalendar.DayError(Era, Year, Month, Day: Integer);
Begin
  Raise EEncodingError.CreateResFmt(@SInvalidDayError, [Year, Month, Day, CalendarName, Era]);
End;

Class Procedure TCalendar.DayNumError(DayNum: Integer);
Begin
  Raise EConvertError.CreateResFmt(@SInvalidDayNumError, [DayNum]);
End;

Class Procedure TCalendar.DateError(Const DateTime: TDateTime);
Begin
  Raise EConvertError.CreateResFmt(@SInvalidDateError, [DateToStr(DateTime), CalendarName]);
End;

Class Procedure TCalendar.DateDayError(Era, Year, DayOfYear: Integer);
Begin
  Raise EEncodingError.CreateResFmt(@SInvalidDateDayError, [Year, DayOfYear, CalendarName, Era]);
End;

Class Procedure TCalendar.DateWeekError(Era, Year, Week: Integer; DoW: TDayOfWeek);
Begin
  Raise EEncodingError.CreateResFmt(@SInvalidDateWeekError, [Year, Week, Ord(DoW), CalendarName, Era]);
End;

Class Procedure TCalendar.ConvertEraError(Year, FromEra, ToEra: Integer);
Begin
  Raise EComponentError.CreateResFmt(@SInvalidConvertEraError, [Year, FromEra, ToEra, CalendarName]);
End;

Class Procedure TCalendar.Divide(Numerator, Denominator: Integer; Out Quotient, Reminder: Integer);
Begin
  Quotient := Floor(Numerator / Denominator);
  Reminder := Numerator - (Denominator * Quotient);
End;

Class Function TCalendar.Modulus(Numerator, Denominator: Integer): Integer;
Begin
  Result := Numerator - (Denominator * Floor(Numerator / Denominator));
End;

Function TCalendar.ToZeroBase(Era, Year: Integer): Integer;
Begin
  Result := Year;
  If (Year < 1) And Not CanHaveZeroYear(Era) Then
    Inc(Result)
End;

Function TCalendar.FromZeroBase(Era, Year: Integer): Integer;
Begin
  Result := Year;
  If (Year < 1) And Not CanHaveZeroYear(Era) Then
    Dec(Result)
End;

Class Function TCalendar.CalendarID: Cardinal;
Begin
  Result := 0;
End;

Class Function TCalendar.CalendarKind: TCalendarKind;
Begin
  Result := CkUnknown;
End;

Class Function TCalendar.CalendarName: String;
Var
  I: Integer;
  NeedsSpace: Boolean;
  NeedsCloseParenthesis: Boolean;
Begin
  Result := String(ClassName);
  Delete(Result, 1, 1);
  NeedsSpace := False;
  NeedsCloseParenthesis := False;
  I := 1;
  While I <= Length(Result) Do
  Begin
    If (Result[I] = '_') And Not NeedsCloseParenthesis Then
    Begin
      Result[I] := ' ';
      Inc(I);
      Insert('(', Result, I);
      NeedsSpace := False;
      NeedsCloseParenthesis := True;
    End
    Else If Not CharInSet(Result[I], ['A' .. 'Z', '0' .. '9']) Then
      NeedsSpace := True
    Else If NeedsSpace Then
    Begin
      Insert(' ', Result, I);
      Inc(I);
      NeedsSpace := False;
    End;
    Inc(I);
  End;
  If NeedsCloseParenthesis Then
    Insert(')', Result, I);
End;

Class Function TCalendar.MaxEra: Integer;
Begin
  Result := 1;
End;

Class Function TCalendar.MaxMonthsPerYear: Integer;
Begin
  Result := 12;
End;

Class Function TCalendar.MinSupportedDateTime: TDateTime;
Begin
  Result := SysUtils.MinDateTime;
End;

Class Function TCalendar.MaxSupportedDateTime: TDateTime;
Begin
  Result := SysUtils.MaxDateTime;
End;

Class Function TCalendar.Today: TDate;
Begin
  Result := SysUtils.Date;
End;

Class Function TCalendar.Yesterday: TDate;
Begin
  Result := PrevDay(Today);
End;

Class Function TCalendar.Tomorrow: TDate;
Begin
  Result := NextDay(Today);
End;

Class Function TCalendar.FirstMonthOfYear: Integer;
Begin
  Result := 1;
End;

Function TCalendar.DaysToMonth(Era, Year, Month: Integer): Integer;
Begin
  If Month <> FirstMonthOfYear Then
    Result := DaysBetween(StartOfMonth(Era, Year, Month), StartOfYear(Era, Year))
  Else
    Result := 0;
End;

Function TCalendar.DaysToMonth(Year, Month: Integer): Integer;
Begin
  Result := DaysToMonth(DefaultEra, Year, Month);
End;

Function TCalendar.DaysToMonth(Const DateTime: TDateTime): Integer;
Var
  Era, Year, Month, Day: Integer;
Begin
  DecodeDate(DateTime, Era, Year, Month, Day);
  Result := DaysToMonth(Era, Year, Month);
End;

Function TCalendar.DaysInMonth(Era, Year, Month: Integer): Integer;
Begin
  Result := DaysBetween(EndOfMonth(Era, Year, Month), StartOfMonth(Era, Year, Month)) + 1;
End;

Function TCalendar.DaysInMonth(Year, Month: Integer): Integer;
Begin
  Result := DaysInMonth(DefaultEra, Year, Month);
End;

Function TCalendar.DaysInMonth(Const DateTime: TDateTime): Integer;
Var
  Era, Year, Month, Day: Integer;
Begin
  DecodeDate(DateTime, Era, Year, Month, Day);
  Result := DaysInMonth(Era, Year, Month);
End;

Function TCalendar.DaysInYear(Era, Year: Integer): Integer;
Begin
  Result := DaysBetween(EndOfYear(Era, Year), StartOfYear(Era, Year)) + 1;
End;

Function TCalendar.DaysInYear(Year: Integer): Integer;
Begin
  Result := DaysInYear(DefaultEra, Year);
End;

Function TCalendar.DaysInYear(Const DateTime: TDateTime): Integer;
Var
  Era, Year, Month, Day: Integer;
Begin
  DecodeDate(DateTime, Era, Year, Month, Day);
  Result := DaysInYear(Era, Year);
End;

Function TCalendar.MonthsInYear(Era, Year: Integer): Integer;
Begin
  Result := MaxMonthsPerYear;
End;

Function TCalendar.MonthsInYear(Year: Integer): Integer;
Begin
  Result := MonthsInYear(DefaultEra, Year);
End;

Function TCalendar.MonthsInYear(Const DateTime: TDateTime): Integer;
Var
  Era, Year, Month, Day: Integer;
Begin
  DecodeDate(DateTime, Era, Year, Month, Day);
  Result := MonthsInYear(Era, Year);
End;

Function TCalendar.WeeksInYear(Era, Year: Integer): Integer;
Var
  LastWeekDate: TDate;
Begin
  LastWeekDate := StartOfWeek(EndOfYear(Era, Year));
  Result := WeekOfYear(LastWeekDate);
  If Result = 1 Then
  Begin
    DecWeek(LastWeekDate);
    Result := WeekOfYear(LastWeekDate);
  End;
End;

Function TCalendar.WeeksInYear(Year: Integer): Integer;
Begin
  Result := WeeksInYear(DefaultEra, Year);
End;

Function TCalendar.WeeksInYear(Const DateTime: TDateTime): Integer;
Var
  Era, Year, Month, Day: Integer;
Begin
  DecodeDate(DateTime, Era, Year, Month, Day);
  Result := WeeksInYear(Era, Year);
End;

Function TCalendar.WeeksInMonth(Era, Year, Month: Integer): Integer;
Begin
  Result := ((DaysInMonth(Era, Year, Month) - 1) Div DaysPerWeek) + 1;
End;

Function TCalendar.WeeksInMonth(Year, Month: Integer): Integer;
Begin
  Result := WeeksInMonth(DefaultEra, Year, Month);
End;

Function TCalendar.WeeksInMonth(Const DateTime: TDateTime): Integer;
Var
  Era, Year, Month, Day: Integer;
Begin
  DecodeDate(DateTime, Era, Year, Month, Day);
  Result := WeeksInMonth(Era, Year, Month);
End;

Function TCalendar.DayOfYearToDayOfMonth(Era, Year, DayOfYear: Integer; Var Month, Day: Integer): Boolean;
Var
  FirstMonth: Integer;
  MonthsPerYear: Integer;
  DaysPerMonth: Integer;
Begin
  FirstMonth := FirstMonthOfYear;
  Day := DayOfYear;
  Month := FirstMonth;
  MonthsPerYear := MonthsInYear(Era, Year);
  DaysPerMonth := DaysInMonth(Era, Year, Month);
  While Day > DaysPerMonth Do
  Begin
    Dec(Day, DaysPerMonth);
    Inc(Month);
    If Month > MonthsPerYear Then
    Begin
      Month := 1;
      If Month = FirstMonth Then
      Begin
        Result := False;
        Exit;
      End;
    End;
    DaysPerMonth := DaysInMonth(Era, Year, Month);
  End;
  Result := (Day >= 1);
End;

Function TCalendar.DayOfYearToDayOfMonth(Year, DayOfYear: Integer; Var Month, Day: Integer): Boolean;
Begin
  Result := DayOfYearToDayOfMonth(DefaultEra, Year, DayOfYear, Month, Day);
End;

Function TCalendar.DayOfMonthToDayOfYear(Era, Year, Month, Day: Integer; Out DayOfYear: Integer): Boolean;
Begin
  If IsValidDay(Era, Year, Month, Day) Then
  Begin
    If Month = FirstMonthOfYear Then
      DayOfYear := Day
    Else
      DayOfYear := DaysToMonth(Era, Year, Month) + Day;
    Result := True;
  End
  Else
    Result := False;
End;

Function TCalendar.DayOfMonthToDayOfYear(Year, Month, Day: Integer; Out DayOfYear: Integer): Boolean;
Begin
  Result := DayOfMonthToDayOfYear(DefaultEra, Year, Month, Day, DayOfYear);
End;

Function TCalendar.ConvertYear(Year, FromEra, ToEra: Integer): Integer;
Begin
  If FromEra <> ToEra Then
    ConvertEraError(Year, FromEra, ToEra);
  If Not IsValidYear(FromEra, Year) Then
    YearError(FromEra, Year);
  Result := Year;
End;

Function TCalendar.AdjustJulianDay(JD: Extended; Backward: Boolean): Extended;
Begin
  Result := JD;
End;

Function TCalendar.BaseEra: Integer;
Begin
  Result := 1;
End;

Function TCalendar.BestEraOfBaseEraYear(Year: Integer): Integer;
Begin
  Result := DefaultEra;
End;

Function TCalendar.ToFourDigitYear(Year: Integer): Integer;
Begin
  Result := Year;
  If (Year > 0) And (Year < 100) Then
  Begin
    Inc(Result, 100 * (Settings.TwoDigitYearMax Div 100));
    If Result > Settings.TwoDigitYearMax Then
      Dec(Result, 100);
  End;
End;

Function TCalendar.OffsetYear(Year, NumberOfYears, FromEra, ToEra: Integer): Integer;
Begin
  // In most calendars the year zero is meaningless and we should take care of it.
  Result := FromZeroBase(ToEra, ToZeroBase(FromEra, Year) + NumberOfYears);
End;

Function TCalendar.OffsetYear(Era, Year, NumberOfYears: Integer): Integer;
Begin
  Result := OffsetYear(Year, NumberOfYears, Era, Era);
End;

Function TCalendar.OffsetYear(Year, NumberOfYears: Integer): Integer;
Begin
  Result := OffsetYear(DefaultEra, Year, NumberOfYears);
End;

Class Function TCalendar.OffsetDateTime(Const DateTime: TDateTime; Const Delta: Double): TDateTime;
Begin
  // We should take in account that in TDateTime type
  // 1. The fraction part is always a positive value even when the whole value is negative
  // 2. There is no value in range (-1, 0)
  Result := JulianDayToDateTime(DateTimeToJulianDay(DateTime) + Delta);
End;

Class Function TCalendar.DaysBetween(DateTime1, DateTime2: TDateTime): Integer;
Begin
  Result := Abs(Trunc(DateTime1) - Trunc(DateTime2));
End;

Class Function TCalendar.DaysOfWeekBetween(DoW1, DoW2: TDayOfWeek): Integer;
Begin
  Result := (Ord(DoW1) - Ord(DoW2) + DaysPerWeek) Mod DaysPerWeek;
End;

Function TCalendar.DayOfWeekToDayNum(DoW: TDayOfWeek): Integer;
Begin
  Result := ((DaysPerWeek - Ord(Settings.FirstDayOfWeek) + Ord(Dow)) Mod DaysPerWeek) + 1;
End;

Function TCalendar.DayNumToDayOfWeek(DayNum: Integer): TDayOfWeek;
Begin
  If DayNum In [1 .. 7] Then
    DayNumError(DayNum);
  Result := TDayOfWeek(((Ord(Settings.FirstDayOfWeek) + DayNum - 2) Mod DaysPerWeek) + 1);
End;

Function TCalendar.CanHaveZeroYear(Era: Integer): Boolean;
Begin
  Result := False;
End;

Class Function TCalendar.IsNoDate(Const DateTime: TDateTime): Boolean;
Begin
  Result := (DateTime > -1) And (DateTime < 0);
End;

Function TCalendar.IsValidDate(Const DateTime: TDateTime): Boolean;
Begin
  Result := IsInRange(DateTime, MinSupportedDateTime, MaxSupportedDateTime);
End;

Function TCalendar.IsValidDateDay(Era, Year, DayOfYear: Integer): Boolean;
Begin
  Result := IsValidYear(Era, Year) And (DayOfYear >= 1) And (DayOfYear <= DaysInYear(Era, Year));
End;

Function TCalendar.IsValidDateDay(Year, DayOfYear: Integer): Boolean;
Begin
  Result := IsValidDateDay(DefaultEra, Year);
End;

Function TCalendar.IsValidDateWeek(Era, Year, Week: Integer; DoW: TDayOfWeek): Boolean;
Begin
  Result := IsValidYear(Era, Year) And (Week >= 1) And (Week <= WeeksInYear(Era, Year)) And
    (Ord(DoW) In [Ord( Low(TDayOfWeek)) .. Ord( High(TDayOfWeek))]);
End;

Function TCalendar.IsValidDateWeek(Year, Week: Integer; DoW: TDayOfWeek): Boolean;
Begin
  Result := IsValidDateWeek(DefaultEra, Year, DoW);
End;

Function TCalendar.IsValidEra(Era: Integer): Boolean;
Begin
  Result := (Era = DefaultEra) Or ((Era >= 1) And (Era <= MaxEra));
End;

Function TCalendar.IsValidYear(Era, Year: Integer): Boolean;
Begin
  Result := IsValidEra(Era) And ((Year <> 0) Or CanHaveZeroYear(Era));
End;

Function TCalendar.IsValidYear(Year: Integer): Boolean;
Begin
  Result := IsValidYear(DefaultEra, Year);
End;

Function TCalendar.IsValidMonth(Era, Year, Month: Integer): Boolean;
Begin
  Result := IsValidYear(Era, Year) And
    ((Month = FirstMonthOfYear) Or ((Month >= 1) And (Month <= MonthsInYear(Era, Year))));
End;

Function TCalendar.IsValidMonth(Year, Month: Integer): Boolean;
Begin
  Result := IsValidMonth(DefaultEra, Year, Month);
End;

Function TCalendar.IsValidDay(Era, Year, Month, Day: Integer): Boolean;
Begin
  Result := IsValidMonth(Era, Year, Month) And
    ((Day = 1) Or ((Day >= 1) And (Day <= DaysInMonth(Era, Year, Month))));
End;

Function TCalendar.IsValidDay(Year, Month, Day: Integer): Boolean;
Begin
  Result := IsValidDay(DefaultEra, Year, Month)
End;

Function TCalendar.IsLeapYear(Era, Year: Integer): Boolean;
Begin
  If Not IsValidYear(Era, Year) Then
    YearError(Era, Year);
  Result := False;
End;

Function TCalendar.IsLeapYear(Year: Integer): Boolean;
Begin
  Result := IsLeapYear(DefaultEra, Year);
End;

Function TCalendar.IsInLeapYear(Const DateTime: TDateTime): Boolean;
Var
  Era, Year, Month, Day: Integer;
Begin
  DecodeDate(DateTime, Era, Year, Month, Day);
  Result := IsLeapYear(Era, Year);
End;

Class Function TCalendar.IsInRange(Const DateTime, RangeStart, RangeEnd: TDateTime): Boolean;
Begin
  Result := Not IsNoDate(DateTime) And (Compare(DateTime, RangeStart) >= 0) And
    (Compare(DateTime, RangeEnd) <= 0);
End;

Function TCalendar.IsThisEra(Const DateTime: TDateTime): Boolean;
Begin
  Result := IsSameEra(DateTime, Today);
End;

Function TCalendar.IsThisEra(Era: Integer): Boolean;
Begin
  Result := (EraOf(Today) = Era);
End;

Function TCalendar.IsThisYear(Const DateTime: TDateTime): Boolean;
Begin
  Result := IsSameYear(DateTime, Today);
End;

Function TCalendar.IsThisYear(Era, Year: Integer): Boolean;
Begin
  Result := IsThisYear(ConvertYear(Year, Era, DefaultEra));
End;

Function TCalendar.IsThisYear(Year: Integer): Boolean;
Begin
  Result := (YearOf(Today) = Year);
End;

Function TCalendar.IsThisMonth(Const DateTime: TDateTime): Boolean;
Begin
  Result := IsSameMonth(DateTime, Today);
End;

Function TCalendar.IsThisMonth(Era, Year, Month: Integer): Boolean;
Begin
  Result := IsThisMonth(ConvertYear(Year, Era, DefaultEra), Month);
End;

Function TCalendar.IsThisMonth(Year, Month: Integer): Boolean;
Var
  TodayYear, TodayMonth, TodayDay: Integer;
Begin
  DecodeDate(Today, TodayYear, TodayMonth, TodayDay);
  Result := (TodayYear = Year) And (TodayMonth = Month);
End;

Function TCalendar.IsThisMonth(Month: Integer): Boolean;
Begin
  Result := (MonthOfYear(Today) = Month);
End;

Function TCalendar.IsThisWeek(Const DateTime: TDateTime): Boolean;
Begin
  Result := IsSameWeek(DateTime, Today);
End;

Class Function TCalendar.IsToday(Const DateTime: TDateTime): Boolean;
Begin
  Result := IsSameDay(DateTime, Today);
End;

Class Function TCalendar.IsToday(DoW: TDayOfWeek): Boolean;
Begin
  Result := (DayOfWeek(Today) = DoW);
End;

Class Function TCalendar.IsPM(Const DateTime: TDateTime): Boolean;
Begin
  Result := TimeOf(DateTime) >= 0.5;
End;

Class Function TCalendar.IsSame(Const DateTime1, DateTime2: TDateTime): Boolean;
Begin
  Result := (Abs(DateTime1 - DateTime2) < OneMilliSecond);
End;

Function TCalendar.IsSameEra(Const DateTime1, DateTime2: TDateTime): Boolean;
Begin
  Result := (EraOf(DateTime1) = EraOf(DateTime2));
End;

Function TCalendar.IsSameYear(Const DateTime1, DateTime2: TDateTime): Boolean;
Var
  Era1, Year1, Month1, Day1: Integer;
  Era2, Year2, Month2, Day2: Integer;
Begin
  DecodeDate(DateTime1, Era1, Year1, Month1, Day1);
  DecodeDate(DateTime2, Era2, Year2, Month2, Day2);
  Result := (Era1 = Era2) And (Year1 = Year2);
End;

Function TCalendar.IsSameMonth(Const DateTime1, DateTime2: TDateTime): Boolean;
Var
  Era1, Year1, Month1, Day1: Integer;
  Era2, Year2, Month2, Day2: Integer;
Begin
  DecodeDate(DateTime1, Era1, Year1, Month1, Day1);
  DecodeDate(DateTime2, Era2, Year2, Month2, Day2);
  Result := (Era1 = Era2) And (Year1 = Year2) And (Month1 = Month2);
End;

Function TCalendar.IsSameWeek(Const DateTime1, DateTime2: TDateTime): Boolean;
Begin
  Result := IsInRange(DateTime1, StartOfWeek(DateTime2), EndOfWeek(DateTime2));
End;

Class Function TCalendar.IsSameDay(Const DateTime1, DateTime2: TDateTime): Boolean;
Begin
  Result := (Trunc(DateTime1) = Trunc(DateTime2));
End;

Class Function TCalendar.Compare(Const DateTime1, DateTime2: TDateTime): Integer;
Begin
  If IsSame(DateTime1, DateTime2) Then
    Result := 0
  Else If DateTime1 < DateTime2 Then
    Result := -1
  Else
    Result := +1;
End;

Class Function TCalendar.CompareDate(Const DateTime1, DateTime2: TDateTime): Integer;
Begin
  Result := Compare(DateOf(DateTime1), DateOf(DateTime2));
End;

Class Function TCalendar.CompareTime(Const DateTime1, DateTime2: TDateTime): Integer;
Begin
  Result := Compare(TimeOf(DateTime1), TimeOf(DateTime2));
End;

Function TCalendar.TryEncodeDateWeek(Era, Year, Week: Integer; DoW: TDayOfWeek;
  Out DateTime: TDateTime): Boolean;
Var
  YearStart: TDateTime;
  DayOfYear: Integer;
  Offset: Integer;
Begin
  Result := False;
  If IsValidDateWeek(Era, Year, Week, DoW) Then
  Begin
    YearStart := StartOfYear(Era, Year);
    DayOfYear := ((Week - 1) * DaysPerWeek) + DaysOfWeekBetween(DoW, Settings.FirstDayOfWeek);
    Offset := DaysOfWeekBetween(DayOfWeek(YearStart), Settings.FirstDayOfWeek);
    Case Settings.WeekRule Of
      WrFirstFourDayWeek:
        If Offset >= 4 Then
          Dec(Offset, DaysPerWeek);
      WrFullWeek:
        If Offset <> 0 Then
          Dec(Offset, DaysPerWeek);
    End;
    DateTime := YearStart + DayOfYear - Offset;
    Result := True;
  End;
End;

Function TCalendar.TryEncodeDateWeek(Year, Week: Integer; DoW: TDayOfWeek; Out DateTime: TDateTime): Boolean;
Begin
  Result := TryEncodeDateWeek(DefaultEra, Year, Week, DoW, DateTime);
End;

Function TCalendar.EncodeDateWeek(Era, Year, Week: Integer; DoW: TDayOfWeek): TDateTime;
Begin
  If Not TryEncodeDateWeek(Era, Year, Week, DoW, Result) Then
    DateWeekError(Era, Year, Week, DoW);
End;

Function TCalendar.EncodeDateWeek(Year, Week: Integer; DoW: TDayOfWeek): TDateTime;
Begin
  Result := EncodeDateWeek(DefaultEra, Year, Week, DoW);
End;

Function TCalendar.TryDecodeDateWeek(Const DateTime: TDateTime; Out Era, Year, Week: Integer;
  Out DoW: TDayOfWeek): Boolean;

  Function DoDecode(Const Date: TDateTime): Boolean;
  Var
    DayOfYear: Integer;
    FirstOffset, Offset: Integer;
  Begin
    Result := TryDecodeDateDay(Date, Era, Year, DayOfYear);
    If Result Then
    Begin
      FirstOffset := Ord(DayOfWeek(Date)) - ((DayOfYear - 1) Mod DaysPerWeek);
      If Settings.WeekRule = WrFirstDayWeek Then
      Begin
        Offset := (FirstOffset - Ord(Settings.FirstDayOfWeek) + 2 * DaysPerWeek) Mod DaysPerWeek;
        Week := ((DayOfYear + Offset - 1) Div DaysPerWeek) + 1;
      End
      Else
      Begin
        Offset := (Ord(Settings.FirstDayOfWeek) - FirstOffset + 2 * DaysPerWeek) Mod DaysPerWeek;
        If (Settings.WeekRule = WrFirstFourDayWeek) And (Offset >= 4) Then
          Dec(Offset, DaysPerWeek);
        If DayOfYear > Offset Then
          Week := ((DayOfYear - Offset - 1) Div DaysPerWeek) + 1
        Else
          Result := DoDecode(Date - DayOfYear);
      End;
    End;
  End;

Begin
  DoW := DayOfWeek(DateTime);
  If (Settings.WeekRule = WrFirstFourDayWeek) And (DaysOfWeekBetween(DoW, Settings.FirstDayOfWeek) <= 2) Then
    Result := DoDecode(DateTime + 3)
  Else
    Result := DoDecode(DateTime);
End;

Function TCalendar.TryDecodeDateWeek(Const DateTime: TDateTime; Out Year, Week: Integer;
  Out DoW: TDayOfWeek): Boolean;
Var
  Era: Integer;
Begin
  Result := TryDecodeDateWeek(DateTime, Era, Year, Week, DoW);
  If Result And (Era <> DefaultEra) Then
    Year := ConvertYear(Year, Era, DefaultEra);
End;

Procedure TCalendar.DecodeDateWeek(Const DateTime: TDateTime; Out Era, Year, Week: Integer;
  Out DoW: TDayOfWeek);
Begin
  If Not TryDecodeDateWeek(DateTime, Era, Year, Week, DoW) Then
    DateError(DateTime);
End;

Procedure TCalendar.DecodeDateWeek(Const DateTime: TDateTime; Out Year, Week: Integer; Out DoW: TDayOfWeek);
Begin
  If Not TryDecodeDateWeek(DateTime, Year, Week, DoW) Then
    DateError(DateTime);
End;

Function TCalendar.TryEncodeDateDay(Era, Year, DayOfYear: Integer; Out DateTime: TDateTime): Boolean;
Begin
  If IsValidDateDay(Era, Year, DayOfYear) Then
  Begin
    DateTime := StartOfYear(Era, Year) + (DayOfYear - 1);
    Result := True;
  End
  Else
    Result := False;
End;

Function TCalendar.TryEncodeDateDay(Year, DayOfYear: Integer; Out DateTime: TDateTime): Boolean;
Begin
  Result := TryEncodeDateDay(DefaultEra, Year, DayOfYear, DateTime);
End;

Function TCalendar.EncodeDateDay(Era, Year, DayOfYear: Integer): TDateTime;
Begin
  If Not TryEncodeDateDay(Era, Year, DayOfYear, Result) Then
    DateDayError(Era, Year, DayOfYear);
End;

Function TCalendar.EncodeDateDay(Year, DayOfYear: Integer): TDateTime;
Begin
  Result := EncodeDateDay(DefaultEra, Year, DayOfYear);
End;

Function TCalendar.TryDecodeDateDay(Const DateTime: TDateTime; Out Era, Year, DayOfYear: Integer): Boolean;
Var
  Month, Day: Integer;
Begin
  Result := TryDecodeDate(DateTime, Era, Year, Month, Day) And DayOfMonthToDayOfYear(Era, Year, Month, Day,
    DayOfYear);
End;

Function TCalendar.TryDecodeDateDay(Const DateTime: TDateTime; Out Year, DayOfYear: Integer): Boolean;
Var
  Era: Integer;
Begin
  Result := TryDecodeDateDay(DateTime, Era, Year, DayOfYear);
  If Result And (Era <> DefaultEra) Then
    Year := ConvertYear(Year, Era, DefaultEra);
End;

Procedure TCalendar.DecodeDateDay(Const DateTime: TDateTime; Out Era, Year, DayOfYear: Integer);
Begin
  If Not TryDecodeDateDay(DateTime, Era, Year, DayOfYear) Then
    DateError(DateTime);
End;

Procedure TCalendar.DecodeDateDay(Const DateTime: TDateTime; Out Year, DayOfYear: Integer);
Begin
  If Not TryDecodeDateDay(DateTime, Year, DayOfYear) Then
    DateError(DateTime);
End;

Function TCalendar.TryEncodeDate(Era, Year, Month, Day: Integer; Out DateTime: TDateTime): Boolean;
Var
  JD: Extended;
Begin
  Result := False;
  If IsValidDay(Era, Year, Month, Day) Then
  Begin
    Year := ConvertYear(Year, Era, BaseEra);
    JD := ToJulianDay(Year, Month, Day);
    JD := AdjustJulianDay(JD, True);
    DateTime := JulianDayToDateTime(JD);
    Result := True;
  End;
End;

Function TCalendar.TryEncodeDate(Year, Month, Day: Integer; Out DateTime: TDateTime): Boolean;
Begin
  Result := TryEncodeDate(DefaultEra, Year, Month, Day, DateTime);
End;

Function TCalendar.EncodeDate(Era, Year, Month, Day: Integer): TDateTime;
Begin
  If Not TryEncodeDate(Era, Year, Month, Day, Result) Then
    DayError(Era, Year, Month, Day);
End;

Function TCalendar.EncodeDate(Year, Month, Day: Integer): TDateTime;
Begin
  Result := EncodeDate(DefaultEra, Year, Month, Day);
End;

Function TCalendar.TryDecodeDate(Const DateTime: TDateTime; Out Era, Year, Month, Day: Integer): Boolean;
Var
  JD: Extended;
Begin
  JD := DateTimeToJulianDay(DateTime);
  JD := AdjustJulianDay(JD, False);
  Result := FromJulianDay(JD, Year, Month, Day);
  If Result Then
  Begin
    Era := BestEraOfBaseEraYear(Year);
    If Era <> BaseEra Then
      Year := ConvertYear(Year, BaseEra, Era);
  End;
End;

Function TCalendar.TryDecodeDate(Const DateTime: TDateTime; Out Year, Month, Day: Integer): Boolean;
Var
  JD: Extended;
Begin
  JD := DateTimeToJulianDay(DateTime);
  JD := AdjustJulianDay(JD, False);
  Result := FromJulianDay(JD, Year, Month, Day);
  If Result And (DefaultEra <> BaseEra) Then
    Year := ConvertYear(Year, BaseEra, DefaultEra);
End;

Procedure TCalendar.DecodeDate(Const DateTime: TDateTime; Out Era, Year, Month, Day: Integer);
Begin
  If Not TryDecodeDate(DateTime, Era, Year, Month, Day) Then
    DateError(DateTime);
End;

Procedure TCalendar.DecodeDate(Const DateTime: TDateTime; Out Year, Month, Day: Integer);
Begin
  If Not TryDecodeDate(DateTime, Year, Month, Day) Then
    DateError(DateTime);
End;

Class Function TCalendar.DateTimeToJulianDay(Const DateTime: TDateTime): Extended;
Begin
  Result := DateTime + DATETIME_EPOCH;
  // We should take in account that in TDateTime type
  // 1. The fraction part is always a positive value even when the whole value is negative
  // 2. There is no value in range (-1, 0)
  If DateTime < 0 Then
    Result := Result - 2 * Frac(DateTime);
End;

Class Function TCalendar.JulianDayToDateTime(Const JulianDate: Extended): TDateTime;
Var
  F: Double;
Begin
  Result := JulianDate - DATETIME_EPOCH;
  // We should take in account that in TDateTime type
  // 1. The fraction part is always a positive value even when the whole value is negative
  // 2. There is no value in range (-1, 0)
  If Result < 0 Then
  Begin
    F := Frac(Result);
    If F <> 0 Then
      Result := Result - 2 * (F + 1);
  End;
End;

Class Function TCalendar.CombineDateTime(Const Date: TDate; Const Time: TTime): TDateTime;
Begin
  If Date < 0 Then
    Result := Date - Time
  Else
    Result := Date + Time;
End;

Class Function TCalendar.LocalToUniversalDateTime(Const DateTime: TDateTime): TDateTime;
Var
  LocalSystemTime, UniversalSystemTime: TSystemTime;
  LocalFileTime, UniversalFileTime: TFileTime;
Begin
  DateTimeToSystemTime(DateTime, LocalSystemTime);
  SystemTimeToFileTime(LocalSystemTime, LocalFileTime);
  LocalFileTimeToFileTime(LocalFileTime, UniversalFileTime);
  FileTimeToSystemTime(UniversalFileTime, UniversalSystemTime);
  Result := SystemTimeToDateTime(UniversalSystemTime);
End;

Class Function TCalendar.UniversalToLocalDateTime(Const DateTime: TDateTime): TDateTime;
Var
  LocalTime, UniversalTime: TSystemTime;
Begin
  DateTimeToSystemTime(DateTime, UniversalTime);
  SystemTimeToTzSpecificLocalTime(Nil, UniversalTime, LocalTime);
  Result := SystemTimeToDateTime(LocalTime);
End;

Class Function TCalendar.DateOf(Const DateTime: TDateTime): TDate;
Begin
  Result := Trunc(DateTime);
End;

Function TCalendar.EraOf(Const DateTime: TDateTime): Integer;
Var
  Year, Month, Day: Integer;
Begin
  DecodeDate(DateTime, Result, Year, Month, Day);
End;

Function TCalendar.YearOf(Const DateTime: TDateTime): Integer;
Var
  Month, Day: Integer;
Begin
  DecodeDate(DateTime, Result, Month, Day);
End;

Function TCalendar.MonthOfYear(Const DateTime: TDateTime): Integer;
Var
  Year, Day: Integer;
Begin
  DecodeDate(DateTime, Year, Result, Day);
End;

Function TCalendar.WeekOfYear(Const DateTime: TDateTime): Integer;
Var
  Year: Integer;
  DoW: TDayOfWeek;
Begin
  DecodeDateWeek(DateTime, Year, Result, DoW);
End;

Function TCalendar.WeekOfMonth(Const DateTime: TDateTime): Integer;
Begin
  Result := ((DayOfMonth(DateTime) - 1) Div DaysPerWeek) + 1;
End;

Function TCalendar.DayOfYear(Const DateTime: TDateTime): Integer;
Begin
  Result := DaysBetween(DateTime, StartOfYear(DateTime)) + 1;
End;

Function TCalendar.DayOfMonth(Const DateTime: TDateTime): Integer;
Var
  Year, Month: Integer;
Begin
  DecodeDate(DateTime, Year, Month, Result);
End;

Class Function TCalendar.DayOfWeek(Const DateTime: TDateTime): TDayOfWeek;
Begin
  Result := TDayOfWeek((Trunc(DateTimeToJulianDay(DateTime) + 0.5) Mod DaysPerWeek) + 1);
End;

Class Function TCalendar.TimeOf(Const DateTime: TDateTime): TTime;
Begin
  Result := Abs(Frac(DateTime));
End;

Class Function TCalendar.HourOfDay(Const DateTime: TDateTime): Integer;
Begin
  Result := Trunc(TimeOf(DateTime) * HoursPerDay);
End;

Class Function TCalendar.MinuteOfDay(Const DateTime: TDateTime): Integer;
Begin
  Result := Trunc(TimeOf(DateTime) * MinsPerDay);
End;

Class Function TCalendar.MinuteOfHour(Const DateTime: TDateTime): Integer;
Begin
  Result := MinuteOfDay(DateTime) Mod MinsPerHour;
End;

Class Function TCalendar.SecondOfDay(Const DateTime: TDateTime): Integer;
Begin
  Result := Trunc(TimeOf(DateTime) * SecsPerDay);
End;

Class Function TCalendar.SecondOfMinute(Const DateTime: TDateTime): Integer;
Begin
  Result := SecondOfDay(DateTime) Mod SecsPerMin;
End;

Class Function TCalendar.MillisecondOfDay(Const DateTime: TDateTime): Integer;
Begin
  Result := Trunc(TimeOf(DateTime) * MSecsPerDay);
End;

Class Function TCalendar.MillisecondOfSecond(Const DateTime: TDateTime): Integer;
Begin
  Result := MillisecondOfDay(DateTime) Mod MSecsPerSec;
End;

Function TCalendar.StartOfEra(Era: Integer): TDateTime;
Begin
  If Not IsValidEra(Era) Then
    EraError(Era);
  If CanHaveZeroYear(Era) Then
    Result := StartOfYear(Era, 0)
  Else
    Result := StartOfYear(Era, 1);
End;

Function TCalendar.StartOfYear(Era, Year: Integer): TDateTime;
Begin
  Result := EncodeDate(Era, Year, FirstMonthOfYear, 1);
End;

Function TCalendar.StartOfYear(Year: Integer): TDateTime;
Begin
  Result := StartOfYear(DefaultEra, Year);
End;

Function TCalendar.StartOfYear(Const DateTime: TDateTime): TDateTime;
Begin
  Result := StartOfYear(YearOf(DateTime));
End;

Function TCalendar.StartOfMonth(Era, Year, Month: Integer): TDateTime;
Begin
  Result := EncodeDate(Era, Year, Month, 1);
End;

Function TCalendar.StartOfMonth(Year, Month: Integer): TDateTime;
Begin
  Result := StartOfMonth(DefaultEra, Year, Month);
End;

Function TCalendar.StartOfMonth(Const DateTime: TDateTime): TDateTime;
Var
  Year, Month, Day: Integer;
Begin
  DecodeDate(DateTime, Year, Month, Day);
  Result := DateOf(DateTime) - (Day - 1);
End;

Function TCalendar.StartOfWeek(Const DateTime: TDateTime): TDateTime;
Begin
  Result := DateOf(DateTime) - DaysOfWeekBetween(DayOfWeek(DateTime), Settings.FirstDayOfWeek);
End;

Class Function TCalendar.StartOfDay(Const DateTime: TDateTime): TDateTime;
Begin
  Result := DateOf(DateTime);
End;

Class Function TCalendar.StartOfHour(Const DateTime: TDateTime): TDateTime;
Begin
  Result := Trunc(DateTime / OneHour) * OneHour;
End;

Class Function TCalendar.StartOfMinute(Const DateTime: TDateTime): TDateTime;
Begin
  Result := Trunc(DateTime / OneMinute) * OneMinute;
End;

Class Function TCalendar.StartOfSecond(Const DateTime: TDateTime): TDateTime;
Begin
  Result := Trunc(DateTime / OneSecond) * OneSecond;
End;

Function TCalendar.EndOfEra(Era: Integer): TDateTime;
Begin
  If Not IsValidEra(Era) Then
    EraError(Era);
  If Era < MaxEra Then
    Result := PrevMillisecond(StartOfEra(Era + 1))
  Else
    Result := MaxSupportedDateTime;
End;

Function TCalendar.EndOfYear(Era, Year: Integer): TDateTime;
Var
  NextYear: Integer;
Begin
  NextYear := OffsetYear(Era, Year, +1);
  Result := PrevMillisecond(StartOfYear(Era, NextYear));
End;

Function TCalendar.EndOfYear(Year: Integer): TDateTime;
Begin
  Result := EndOfYear(DefaultEra, Year);
End;

Function TCalendar.EndOfYear(Const DateTime: TDateTime): TDateTime;
Begin
  Result := EndOfYear(YearOf(DateTime));
End;

Function TCalendar.EndOfMonth(Era, Year, Month: Integer): TDateTime;
Var
  NextMonth: Integer;
Begin
  NextMonth := (Month Mod MonthsInYear(Era, Year)) + 1;
  If NextMonth = FirstMonthOfYear Then
    Result := EndOfYear(Era, Year)
  Else
    Result := PrevMillisecond(StartOfMonth(Era, Year, NextMonth));
End;

Function TCalendar.EndOfMonth(Year, Month: Integer): TDateTime;
Begin
  Result := EndOfMonth(DefaultEra, Year, Month);
End;

Function TCalendar.EndOfMonth(Const DateTime: TDateTime): TDateTime;
Var
  Era, Year, Month, Day: Integer;
Begin
  DecodeDate(DateTime, Era, Year, Month, Day);
  Result := EndOfMonth(Era, Year, Month);
End;

Function TCalendar.EndOfWeek(Const DateTime: TDateTime): TDateTime;
Begin
  Result := PrevMillisecond(StartOfWeek(DateTime) + OneWeek);
End;

Class Function TCalendar.EndOfDay(Const DateTime: TDateTime): TDateTime;
Begin
  Result := PrevMillisecond(StartOfDay(DateTime) + OneDay);
End;

Class Function TCalendar.EndOfHour(Const DateTime: TDateTime): TDateTime;
Begin
  Result := PrevMillisecond(StartOfHour(DateTime) + OneHour);
End;

Class Function TCalendar.EndOfMinute(Const DateTime: TDateTime): TDateTime;
Begin
  Result := PrevMillisecond(StartOfMinute(DateTime) + OneMinute);
End;

Class Function TCalendar.EndOfSecond(Const DateTime: TDateTime): TDateTime;
Begin
  Result := PrevMillisecond(StartOfSecond(DateTime) + OneSecond);
End;

Function TCalendar.PrevYear(Const DateTime: TDateTime; NumberOfYears: Integer): TDateTime;
Begin
  Result := NextYear(DateTime, -NumberOfYears);
End;

Function TCalendar.PrevMonth(Const DateTime: TDateTime; NumberOfMonths: Integer): TDateTime;
Begin
  Result := NextMonth(DateTime, -NumberOfMonths)
End;

Class Function TCalendar.PrevWeek(Const DateTime: TDateTime; NumberOfWeeks: Integer): TDateTime;
Begin
  Result := OffsetDateTime(DateTime, -NumberOfWeeks * OneWeek);
End;

Class Function TCalendar.PrevDay(Const DateTime: TDateTime; NumberOfDays: Integer): TDateTime;
Begin
  Result := OffsetDateTime(DateTime, -NumberOfDays * OneDay);
End;

Function TCalendar.PrevDayOfWeek(Const DateTime: TDateTime; DoW: TDayOfWeek): TDateTime;
Var
  Days: Integer;
Begin
  Days := DaysPerWeek - DaysOfWeekBetween(DayOfWeek(DateTime), DoW);
  Result := PrevDay(DateTime, Days);
End;

Class Function TCalendar.PrevHour(Const DateTime: TDateTime; NumberOfHours: Integer): TDateTime;
Begin
  Result := OffsetDateTime(DateTime, -NumberOfHours * OneHour);
End;

Class Function TCalendar.PrevMinute(Const DateTime: TDateTime; NumberOfMinutes: Integer): TDateTime;
Begin
  Result := OffsetDateTime(DateTime, -NumberOfMinutes * OneMinute);
End;

Class Function TCalendar.PrevSecond(Const DateTime: TDateTime; NumberOfSeconds: Integer): TDateTime;
Begin
  Result := OffsetDateTime(DateTime, -NumberOfSeconds * OneSecond);
End;

Class Function TCalendar.PrevMillisecond(Const DateTime: TDateTime; NumberOfMilliseconds: Integer): TDateTime;
Begin
  Result := OffsetDateTime(DateTime, -NumberOfMilliseconds * OneMillisecond);
End;

Function TCalendar.NextYear(Const DateTime: TDateTime; NumberOfYears: Integer): TDateTime;
Var
  Era, Year, Month, Day: Integer;
Begin
  If NumberOfYears = 0 Then
  Begin
    Result := DateTime;
    Exit;
  End;
  DecodeDate(DateTime, Era, Year, Month, Day);
  Year := OffsetYear(Era, Year, NumberOfYears);
  Month := Min(Month, MonthsInYear(Era, Year));
  Day := Min(Day, DaysInMonth(Era, Year, Month));
  Result := CombineDateTime(EncodeDate(Era, Year, Month, Day), Frac(DateTime));
End;

Function TCalendar.NextMonth(Const DateTime: TDateTime; NumberOfMonths: Integer): TDateTime;
Var
  Era, Year, Month, Day: Integer;
  MonthsToStartOfYear, MonthsOfThisYear: Integer;
  FirstMonth: Integer;
Begin
  If NumberOfMonths = 0 Then
  Begin
    Result := DateTime;
    Exit;
  End;
  DecodeDate(DateTime, Era, Year, Month, Day);
  MonthsOfThisYear := MonthsInYear(Era, Year);
  FirstMonth := FirstMonthOfYear;
  If NumberOfMonths < 0 Then
  Begin
    If Month < FirstMonth Then
      MonthsToStartOfYear := MonthsOfThisYear - (FirstMonth - Month)
    Else
      MonthsToStartOfYear := Month - FirstMonth;
    If -NumberOfMonths > MonthsToStartOfYear Then
    Begin
      Inc(NumberOfMonths, MonthsToStartOfYear);
      Month := FirstMonth;
      Repeat
        Year := OffsetYear(Era, Year, -1);
        MonthsOfThisYear := MonthsInYear(Era, Year);
        Inc(NumberOfMonths, MonthsOfThisYear);
      Until NumberOfMonths >= 0;
    End;
  End
  Else
  Begin
    If Month >= FirstMonth Then
      MonthsToStartOfYear := MonthsOfThisYear - (Month - FirstMonth)
    Else
      MonthsToStartOfYear := FirstMonth - Month;
    If NumberOfMonths >= MonthsToStartOfYear Then
    Begin
      Month := FirstMonth;
      Repeat
        Dec(NumberOfMonths, MonthsToStartOfYear);
        Year := OffsetYear(Era, Year, +1);
        MonthsToStartOfYear := MonthsInYear(Era, Year);
      Until NumberOfMonths < MonthsToStartOfYear;
      MonthsOfThisYear := MonthsToStartOfYear;
    End;
  End;
  Month := ((Month + NumberOfMonths + MonthsOfThisYear - 1) Mod MonthsOfThisYear) + 1;
  Day := Min(Day, DaysInMonth(Era, Year, Month));
  Result := CombineDateTime(EncodeDate(Era, Year, Month, Day), Frac(DateTime));
End;

Class Function TCalendar.NextWeek(Const DateTime: TDateTime; NumberOfWeeks: Integer): TDateTime;
Begin
  Result := OffsetDateTime(DateTime, NumberOfWeeks * OneWeek);
End;

Class Function TCalendar.NextDay(Const DateTime: TDateTime; NumberOfDays: Integer): TDateTime;
Begin
  Result := OffsetDateTime(DateTime, NumberOfDays * OneDay);
End;

Function TCalendar.NextDayOfWeek(Const DateTime: TDateTime; DoW: TDayOfWeek): TDateTime;
Var
  Days: Integer;
Begin
  Days := DaysOfWeekBetween(DayOfWeek(DateTime), DoW);
  If Days = 0 Then
    Days := DaysPerWeek;
  Result := NextDay(DateTime, Days);
End;

Class Function TCalendar.NextHour(Const DateTime: TDateTime; NumberOfHours: Integer): TDateTime;
Begin
  Result := OffsetDateTime(DateTime, NumberOfHours * OneHour);
End;

Class Function TCalendar.NextMinute(Const DateTime: TDateTime; NumberOfMinutes: Integer): TDateTime;
Begin
  Result := OffsetDateTime(DateTime, NumberOfMinutes * OneDay);
End;

Class Function TCalendar.NextSecond(Const DateTime: TDateTime; NumberOfSeconds: Integer): TDateTime;
Begin
  Result := OffsetDateTime(DateTime, NumberOfSeconds * OneSecond);
End;

Class Function TCalendar.NextMillisecond(Const DateTime: TDateTime; NumberOfMilliseconds: Integer): TDateTime;
Begin
  Result := OffsetDateTime(DateTime, NumberOfMilliseconds * OneSecond);
End;

Procedure TCalendar.IncYear(Var DateTime: TDateTime; NumberOfYears: Integer);
Begin
  DateTime := NextYear(DateTime, NumberOfYears);
End;

Procedure TCalendar.IncYear(Var Date: TDate; NumberOfYears: Integer);
Begin
  IncYear(TDateTime(Date), NumberOfYears);
End;

Procedure TCalendar.IncMonth(Var DateTime: TDateTime; NumberOfMonths: Integer);
Begin
  DateTime := NextMonth(DateTime, NumberOfMonths);
End;

Procedure TCalendar.IncMonth(Var Date: TDate; NumberOfMonths: Integer);
Begin
  IncMonth(TDateTime(Date), NumberOfMonths);
End;

Class Procedure TCalendar.IncWeek(Var DateTime: TDateTime; NumberOfWeeks: Integer);
Begin
  IncDateTime(DateTime, NumberOfWeeks * OneWeek);
End;

Class Procedure TCalendar.IncWeek(Var Date: TDate; NumberOfWeeks: Integer);
Begin
  IncWeek(TDateTime(Date), NumberOfWeeks);
End;

Class Procedure TCalendar.IncDay(Var DateTime: TDateTime; NumberOfDays: Integer);
Begin
  IncDateTime(DateTime, NumberOfDays * OneDay);
End;

Class Procedure TCalendar.IncDay(Var Date: TDate; NumberOfDays: Integer);
Begin
  IncDay(TDateTime(Date), NumberOfDays);
End;

Class Procedure TCalendar.IncHour(Var DateTime: TDateTime; NumberOfHours: Integer);
Begin
  IncDateTime(DateTime, NumberOfHours * OneHour);
End;

Class Procedure TCalendar.IncHour(Var Time: TTime; NumberOfHours: Integer);
Begin
  IncHour(TDateTime(Time), NumberOfHours);
End;

Class Procedure TCalendar.IncMinute(Var DateTime: TDateTime; NumberOfMinutes: Integer);
Begin
  IncDateTime(DateTime, NumberOfMinutes * OneMinute);
End;

Class Procedure TCalendar.IncMinute(Var Time: TTime; NumberOfMinutes: Integer);
Begin
  IncMinute(TDateTime(Time), NumberOfMinutes);
End;

Class Procedure TCalendar.IncSecond(Var DateTime: TDateTime; NumberOfSeconds: Integer);
Begin
  IncDateTime(DateTime, NumberOfSeconds * OneSecond);
End;

Class Procedure TCalendar.IncSecond(Var Time: TTime; NumberOfSeconds: Integer);
Begin
  IncSecond(TDateTime(Time), NumberOfSeconds);
End;

Class Procedure TCalendar.IncMillisecond(Var DateTime: TDateTime; NumberOfMilliseconds: Integer);
Begin
  IncDateTime(DateTime, NumberOfMilliseconds * OneMillisecond);
End;

Class Procedure TCalendar.IncMillisecond(Var Time: TTime; NumberOfMilliseconds: Integer);
Begin
  IncMillisecond(TDateTime(Time), NumberOfMilliseconds);
End;

Class Procedure TCalendar.IncDateTime(Var DateTime: TDateTime; Const Delta: Double);
Begin
  DateTime := OffsetDateTime(DateTime, Delta);
End;

Procedure TCalendar.DecYear(Var DateTime: TDateTime; NumberOfYears: Integer);
Begin
  DateTime := PrevYear(DateTime, NumberOfYears);
End;

Procedure TCalendar.DecYear(Var Date: TDate; NumberOfYears: Integer);
Begin
  DecYear(TDateTime(Date), NumberOfYears);
End;

Procedure TCalendar.DecMonth(Var DateTime: TDateTime; NumberOfMonths: Integer);
Begin
  DateTime := PrevMonth(DateTime, NumberOfMonths);
End;

Procedure TCalendar.DecMonth(Var Date: TDate; NumberOfMonths: Integer);
Begin
  DecMonth(TDateTime(Date), NumberOfMonths);
End;

Class Procedure TCalendar.DecWeek(Var DateTime: TDateTime; NumberOfWeeks: Integer);
Begin
  DecDateTime(DateTime, NumberOfWeeks * OneWeek);
End;

Class Procedure TCalendar.DecWeek(Var Date: TDate; NumberOfWeeks: Integer);
Begin
  DecWeek(TDateTime(Date), NumberOfWeeks);
End;

Class Procedure TCalendar.DecDay(Var DateTime: TDateTime; NumberOfDays: Integer);
Begin
  DecDateTime(DateTime, NumberOfDays * OneDay);
End;

Class Procedure TCalendar.DecDay(Var Date: TDate; NumberOfDays: Integer);
Begin
  DecDay(TDateTime(Date), NumberOfDays);
End;

Class Procedure TCalendar.DecHour(Var DateTime: TDateTime; NumberOfHours: Integer);
Begin
  DecDateTime(DateTime, NumberOfHours * OneHour);
End;

Class Procedure TCalendar.DecHour(Var Time: TTime; NumberOfHours: Integer);
Begin
  DecHour(TDateTime(Time), NumberOfHours);
End;

Class Procedure TCalendar.DecMinute(Var DateTime: TDateTime; NumberOfMinutes: Integer);
Begin
  DecDateTime(DateTime, NumberOfMinutes * OneMinute);
End;

Class Procedure TCalendar.DecMinute(Var Time: TTime; NumberOfMinutes: Integer);
Begin
  DecMinute(TDateTime(Time), NumberOfMinutes);
End;

Class Procedure TCalendar.DecSecond(Var DateTime: TDateTime; NumberOfSeconds: Integer);
Begin
  DecDateTime(DateTime, NumberOfSeconds * OneSecond);
End;

Class Procedure TCalendar.DecSecond(Var Time: TTime; NumberOfSeconds: Integer);
Begin
  DecSecond(TDateTime(Time), NumberOfSeconds);
End;

Class Procedure TCalendar.DecMillisecond(Var DateTime: TDateTime; NumberOfMilliseconds: Integer);
Begin
  DecDateTime(DateTime, NumberOfMilliseconds * OneMillisecond);
End;

Class Procedure TCalendar.DecMillisecond(Var Time: TTime; NumberOfMilliseconds: Integer);
Begin
  DecMillisecond(TDateTime(Time), NumberOfMilliseconds);
End;

Class Procedure TCalendar.DecDateTime(Var DateTime: TDateTime; Const Delta: Double);
Begin
  DateTime := OffsetDateTime(DateTime, -Delta);
End;

Function TCalendar.Format(Const FmtStr: String; Const DateTime: TDateTime): String;
Var
  BufPos, AppendLevel: Integer;
  Buffer: Array [0 .. $3FF] Of Char;
  Era, Year, Month, Day: Integer;
  Hour, Min, Sec, MSec, H: Word;
  WeekDecoded, DateDecoded, TimeDecoded: Boolean;
  WeekYear, WeekNum: Integer;
  DoW: TDayOfWeek;

  Procedure AppendChars(P: PChar; Count: Integer);
  Var
    N: Integer;
  Begin
    N := (SizeOf(Buffer) Div SizeOf(Char)) - BufPos;
    If N > Count Then
      N := Count;
    If N <> 0 Then
    Begin
      Move(P^, Buffer[BufPos], N * SizeOf(Char));
      Inc(BufPos, N);
    End;
  End;

  Procedure AppendString(Const S: String); Inline;
  Begin
    AppendChars(PChar(S), Length(S));
  End;

  Procedure AppendNumber(Number, Digits: Integer); Inline;
  Begin
    AppendString(SysUtils.Format('%.*d', [Digits, Number]));
  End;

  Procedure AppendFormat(Fmt: PChar; TranslateSeparators: Boolean = False);
  Var
    Starter, Token, LastToken: Char;
    Use12HourClock: Boolean;
    BetweenQuotes: Boolean;
    Count: Integer;
    P: PChar;

    Procedure GetCount;
    Var
      P: PChar;
    Begin
      P := Fmt;
      While Fmt^ = Starter Do
        Inc(Fmt);
      Count := Fmt - P + 1;
    End;

    Procedure GetDateWeek;
    Begin
      If Not WeekDecoded Then
      Begin
        DecodeDateWeek(DateTime, WeekYear, WeekNum, DoW);
        WeekDecoded := True;
      End;
    End;

    Procedure GetDate;
    Begin
      If Not DateDecoded Then
      Begin
        DecodeDate(DateTime, Era, Year, Month, Day);
        DateDecoded := True;
      End;
    End;

    Procedure GetTime;
    Begin
      If Not TimeDecoded Then
      Begin
        DecodeTime(DateTime, Hour, Min, Sec, MSec);
        TimeDecoded := True;
      End;
    End;

  Begin
    If (Fmt <> Nil) And (AppendLevel < 2) Then
    Begin
      Inc(AppendLevel);
      LastToken := ' ';
      Use12HourClock := False;
      While Fmt^ <> #0 Do
      Begin
        Starter := Fmt^;
        Inc(Fmt);
        Token := UpCase(Starter);
        If CharInSet(Token, ['A' .. 'Z']) Then
        Begin
          If (Token = 'M') And (LastToken = 'H') Then
            Token := 'N'
          Else If Token = 'E' Then
            Token := 'Y';
          LastToken := Token;
        End;
        Case Token Of
          'G':
            Begin
              GetCount;
              GetDate;
              If Count = 1 Then
                AppendString(Settings.ShortEraNames[Era])
              Else
                AppendString(Settings.EraNames[Era]);
            End;
          'Y':
            Begin
              GetCount;
              GetDate;
              Case Count Of
                1:
                  AppendNumber(Year, 1);
                2:
                  AppendNumber(Year Mod 100, 2);
              Else
                AppendNumber(Year, 4);
              End;
            End;
          'M':
            Begin
              GetCount;
              GetDate;
              Case Count Of
                1, 2:
                  AppendNumber(Month, Count);
                3:
                  AppendString(Settings.ShortMonthNames[Month]);
              Else
                If (LastToken = 'Y') Or (LastToken = 'D') Then
                  AppendString(Settings.GenitiveMonthNames[Month])
                Else
                  AppendString(Settings.MonthNames[Month]);
              End;
            End;
          'D':
            Begin
              GetCount;
              If Count <= 2 Then
                GetDate;
              Case Count Of
                1, 2:
                  AppendNumber(Day, Count);
                3:
                  AppendString(Settings.ShortDayNames[DayOfWeek(DateTime)]);
                4:
                  AppendString(Settings.DayNames[DayOfWeek(DateTime)]);
                5:
                  AppendFormat(PChar(Settings.ShortDateFormat));
              Else
                AppendFormat(PChar(Settings.LongDateFormat));
              End;
            End;
          'I':
            Begin
              GetCount;
              GetDate;
              AppendNumber(DayOfYear(DateTime), Count);
            End;
          'O':
            Begin
              GetCount;
              AppendNumber(DayOfWeekToDayNum(DayOfWeek(DateTime)), 1);
            End;
          'W':
            Begin
              GetCount;
              GetDateWeek;
              AppendNumber(WeekNum, Count)
            End;
          'R':
            Begin
              GetCount;
              GetDateWeek;
              Case Count Of
                1:
                  AppendNumber(WeekYear, 1);
                2:
                  AppendNumber(WeekYear Mod 100, 2);
              Else
                AppendNumber(WeekYear, 4);
              End;
            End;
          'H':
            Begin
              GetCount;
              GetTime;
              BetweenQuotes := False;
              P := Fmt;
              While P^ <> #0 Do
              Begin
                Case P^ Of
                  'A', 'a':
                    If Not BetweenQuotes Then
                    Begin
                      If ((AnsiStrLIComp(P, 'AM/PM', 5) = 0) Or (AnsiStrLIComp(P, 'A/P', 3) = 0) Or
                        (AnsiStrLIComp(P, 'AMPM', 4) = 0)) Then
                        Use12HourClock := True;
                      Break;
                    End;
                  'H', 'h':
                    Break;
                  '''', '"':
                    BetweenQuotes := Not BetweenQuotes;
                End;
                Inc(P);
              End;
              H := Hour;
              If Use12HourClock Then
                If H = 0 Then
                  H := 12
                Else If H > 12 Then
                  Dec(H, 12);
              If Count > 2 Then
                Count := 2;
              AppendNumber(H, Count);
            End;
          'N':
            Begin
              GetCount;
              GetTime;
              If Count > 2 Then
                Count := 2;
              AppendNumber(Min, Count);
            End;
          'S':
            Begin
              GetCount;
              GetTime;
              If Count > 2 Then
                Count := 2;
              AppendNumber(Sec, Count);
            End;
          'Z':
            Begin
              GetCount;
              GetTime;
              If Count > 3 Then
                Count := 3;
              AppendNumber(MSec, Count);
            End;
          'A':
            Begin
              GetTime;
              P := Fmt - 1;
              If AnsiStrLIComp(P, 'AM/PM', 5) = 0 Then
              Begin
                If Hour >= 12 Then
                  Inc(P, 3);
                AppendChars(P, 2);
                Inc(Fmt, 4);
                Use12HourClock := True;
              End
              Else If AnsiStrLIComp(P, 'A/P', 3) = 0 Then
              Begin
                If Hour >= 12 Then
                  Inc(P, 2);
                AppendChars(P, 1);
                Inc(Fmt, 2);
                Use12HourClock := True;
              End
              Else If AnsiStrLIComp(P, 'AMPM', 4) = 0 Then
              Begin
                If Hour < 12 Then
                  AppendString(Settings.TimeAMString)
                Else
                  AppendString(Settings.TimePMString);
                Inc(Fmt, 3);
                Use12HourClock := True;
              End
              Else
                AppendChars(@Starter, 1);
            End;
          'C':
            Begin
              GetCount;
              AppendFormat(PChar(Settings.ShortDateFormat));
              GetTime;
              If (Hour <> 0) Or (Min <> 0) Or (Sec <> 0) Then
              Begin
                AppendChars(' ', 1);
                AppendFormat(PChar(Settings.LongTimeFormat));
              End;
            End;
          'T':
            Begin
              GetCount;
              If Count = 1 Then
                AppendFormat(PChar(Settings.ShortTimeFormat))
              Else
                AppendFormat(PChar(Settings.LongTimeFormat));
            End;
          'Q':
            Begin
              GetCount;
              AppendFormat('YYYY-MM-DD HH:NN:SS');
            End;
          'U':
            Begin
              GetCount;
              AppendString(Format('Q', LocalToUniversalDateTime(DateTime)));
            End;
          '/':
            Begin
              If TranslateSeparators Then
                AppendString(Settings.DateSeparator)
              Else
                AppendChars(@Starter, 1);
            End;
          ':':
            Begin
              If TranslateSeparators Then
                AppendString(Settings.TimeSeparator)
              Else
                AppendChars(@Starter, 1);
            End;
          '"', '''':
            Begin
              P := Fmt;
              While (Fmt^ <> #0) And (Fmt^ <> Starter) Do
                Inc(Fmt);
              AppendChars(P, Fmt - P);
              If Fmt^ <> #0 Then
                Inc(Fmt);
            End;
        Else
          AppendChars(@Starter, 1);
        End;
      End;
      Dec(AppendLevel);
    End;
  End;

Begin
  If IsNoDate(DateTime) Then
    Result := ''
  Else
  Begin
    BufPos := 0;
    AppendLevel := 0;
    WeekDecoded := False;
    DateDecoded := False;
    TimeDecoded := False;
    If FmtStr <> '' Then
      AppendFormat(PChar(FmtStr), True)
    Else
      AppendFormat('C');
    SetString(Result, Buffer, BufPos);
  End;
End;

Function TCalendar.Parse(Const FmtStr: String; Const DateTimeStr: String): TDateTime;
Var
  Era, Year, Month, Day: Integer;
  Hour, Min, Sec, MSec: Word;
  WeekYear, WeekNum, DoW, DoY: Integer;
  WeekScanned, DateScanned, TimeScanned: Boolean;
  Is12Hours, IsHourPM: Boolean;
  IsUTC: Boolean;
  ScanLevel: Integer;
  S: PChar;

  Function ScanNumber: Integer;
  Var
    Negate: Boolean;
  Begin
    Result := 0;
    If (S^ = UCC_NODS) Or (S^ = UCC_NADS) Then
      Inc(S);
    Negate := (S^ = '-');
    If Negate Then
      Inc(S);
    While CharInSet(S^, ['0' .. '9']) Do
    Begin
      Result := Result * 10 + (Ord(S^) - Ord('0'));
      Inc(S);
    End;
    If Negate Then
      Result := -Result;
  End;

  Function ScanString(Terminator: Char): String;
  Var
    Start: PChar;
  Begin
    Start := S;
    While (S^ <> #0) And (S^ <> Terminator) Do
      Inc(S);
    SetString(Result, Start, S - Start);
  End;

  Procedure ScanFormat(Fmt: PChar; TranslateSeparators: Boolean = False);
  Var
    Starter, Token, LastToken: Char;
    Count: Integer;
    P: PChar;

    Procedure GetCount;
    Var
      P: PChar;
    Begin
      P := Fmt;
      While Fmt^ = Starter Do
        Inc(Fmt);
      Count := Fmt - P + 1;
    End;

  Begin
    If (Fmt <> Nil) And (ScanLevel < 2) Then
    Begin
      Inc(ScanLevel);
      LastToken := ' ';
      While (Fmt^ <> #0) And (S^ <> #0) Do
      Begin
        Starter := Fmt^;
        Inc(Fmt);
        Token := UpCase(Starter);
        If CharInSet(Token, ['A' .. 'Z']) Then
        Begin
          If (Token = 'M') And (LastToken = 'H') Then
            Token := 'N';
          If Token = 'E' Then
            Token := 'Y';
          LastToken := Token;
        End;
        Case Token Of
          'G':
            Begin
              GetCount;
              If Count = 1 Then
                Era := Settings.ScanShortEraNames(S)
              Else
                Era := Settings.ScanEraNames(S)
            End;
          'Y':
            Begin
              GetCount;
              Year := ScanNumber;
              If Count <= 2 Then
                Year := ToFourDigitYear(Year);
              DateScanned := True;
              If Not WeekScanned Then
                WeekYear := Year;
            End;
          'M':
            Begin
              GetCount;
              Case Count Of
                1, 2:
                  Month := ScanNumber;
                3:
                  Month := Settings.ScanShortMonthNames(S);
              Else
                Month := Settings.ScanMonthNames(S);
                If Month = 0 Then
                  Month := Settings.ScanGenitiveMonthNames(S);
              End;
              DateScanned := True;
            End;
          'D':
            Begin
              GetCount;
              Case Count Of
                1, 2:
                  Begin
                    Day := ScanNumber;
                    DateScanned := True;
                  End;
                3:
                  DoW := Settings.ScanShortDayNames(S);
                4:
                  DoW := Settings.ScanDayNames(S);
                5:
                  ScanFormat(PChar(Settings.ShortDateFormat));
              Else
                ScanFormat(PChar(Settings.LongDateFormat));
              End;
            End;
          'I':
            Begin
              GetCount;
              DoY := ScanNumber;
              DateScanned := True;
            End;
          'O':
            Begin
              GetCount;
              DoW := Ord(DayNumToDayOfWeek(ScanNumber));
              WeekScanned := True;
            End;
          'W':
            Begin
              GetCount;
              WeekNum := ScanNumber;
              WeekScanned := True;
            End;
          'R':
            Begin
              GetCount;
              WeekYear := ScanNumber;
              If Count <= 2 Then
                WeekYear := ToFourDigitYear(WeekYear);
              WeekScanned := True;
            End;
          'H':
            Begin
              GetCount;
              Hour := ScanNumber;
              TimeScanned := True;
            End;
          'N':
            Begin
              GetCount;
              Min := ScanNumber;
              TimeScanned := True;
            End;
          'S':
            Begin
              GetCount;
              Sec := ScanNumber;
              TimeScanned := True;
            End;
          'Z':
            Begin
              GetCount;
              MSec := ScanNumber;
              TimeScanned := True;
            End;
          'A':
            Begin
              P := Fmt - 1;
              If AnsiStrLIComp(P, 'AM/PM', 5) = 0 Then
              Begin
                Inc(Fmt, 4);
                Is12Hours := True;
                IsHourPM := (AnsiStrLIComp(S, 'PM', 2) = 0);
                Inc(S, 2);
              End
              Else If AnsiStrLIComp(P, 'A/P', 3) = 0 Then
              Begin
                Inc(Fmt, 2);
                Is12Hours := True;
                IsHourPM := (AnsiStrLIComp(S, 'P', 1) = 0);
                Inc(S, 1);
              End
              Else If AnsiStrLIComp(P, 'AMPM', 4) = 0 Then
              Begin
                Inc(Fmt, 3);
                Is12Hours := True;
                IsHourPM := (Settings.ScanAMPM(S) = 2);
              End;
            End;
          'C':
            Begin
              GetCount;
              ScanFormat(PChar(Settings.ShortDateFormat + ' ' + Settings.LongTimeFormat));
            End;
          'T':
            Begin
              GetCount;
              If Count = 1 Then
                ScanFormat(PChar(Settings.ShortTimeFormat))
              Else
                ScanFormat(PChar(Settings.LongTimeFormat));
            End;
          'Q', 'U':
            Begin
              GetCount;
              ScanFormat('YYYY-MM-DD HH:NN:SS');
              IsUTC := (Token = 'U');
            End;
          '/':
            Begin
              If TranslateSeparators Then
              Begin
                P := PChar(Settings.DateSeparator);
                While (P^ <> #0) And (S^ <> #0) Do
                Begin
                  Inc(P);
                  Inc(S);
                End;
              End
              Else
                Inc(S);
            End;
          ':':
            Begin
              If TranslateSeparators Then
              Begin
                P := PChar(Settings.TimeSeparator);
                While (P^ <> #0) And (S^ <> #0) Do
                Begin
                  Inc(P);
                  Inc(S);
                End;
              End
              Else
                Inc(S);
            End;
          '"', '''':
            Begin
              While (Fmt^ <> #0) And (Fmt^ <> Starter) And (S^ <> #0) Do
              Begin
                Inc(Fmt);
                Inc(S);
              End;
              If Fmt^ <> #0 Then
                Inc(Fmt);
            End;
        Else
          Inc(S);
        End;
      End;
      Dec(ScanLevel);
    End;
  End;

Var
  Date: TDate;
  Time: TTime;
Begin
  Era := DefaultEra;
  Year := 0;
  Month := 0;
  Day := 0;
  Hour := 0;
  Min := 0;
  Sec := 0;
  MSec := 0;
  WeekYear := 0;
  WeekNum := 0;
  DoW := 0;
  WeekScanned := False;
  DateScanned := False;
  TimeScanned := False;
  Is12Hours := False;
  IsHourPM := False;
  IsUTC := False;
  S := PChar(DateTimeStr);
  ScanLevel := 0;
  If FmtStr <> '' Then
    ScanFormat(PChar(FmtStr), True)
  Else
    ScanFormat('C');
  If TimeScanned Then
  Begin
    If Is12Hours Then
    Begin
      If IsHourPM And (Hour < 12) Then
        Hour := (Hour + 12) Mod 24
      Else If Not IsHourPM And (Hour = 12) Then
        Hour := 0;
    End;
    Time := EncodeTime(Hour, Min, Sec, MSec);
  End
  Else
    Time := 0;
  If DateScanned Then
  Begin
    If (DoY <> 0) And ((Month = 0) Or (Day = 0)) Then
      Date := EncodeDateDay(Era, Year, DoY)
    Else
      Date := EncodeDate(Era, Year, Month, Day);
  End
  Else If WeekScanned Then
    Date := EncodeDateWeek(Era, WeekYear, WeekNum, TDayOfWeek(DoW))
  Else
    Date := 0;
  Result := CombineDateTime(Date, Time);
  If IsUTC Then
    Result := UniversalToLocalDateTime(Result);
End;

Function TCalendar.TryParse(Const FmtStr: String; Const DateTimeStr: String; Var DateTime: TDateTime): Boolean;
Begin
  Try
    DateTime := Parse(FmtStr, DateTimeStr);
    Result := True;
  Except
    Result := False;
  End;
End;

{ TCalendarClassList }

Function TCalendarClassList.GetItems(Index: Integer): TCalendarClass;
Begin
  Result := TCalendarClass(Get(Index));
End;

Procedure TCalendarClassList.SetItems(Index: Integer; Value: TCalendarClass);
Begin
  Put(Index, Value);
End;

Function TCalendarClassList.First: TCalendarClass;
Begin
  Result := TCalendarClass( Inherited First);
End;

Function TCalendarClassList.Last: TCalendarClass;
Begin
  Result := TCalendarClass( Inherited Last);
End;

Function TCalendarClassList.Add(Item: TCalendarClass): Integer;
Begin
  Result := Inherited Add(Item);
End;

Procedure TCalendarClassList.Insert(Index: Integer; Item: TCalendarClass);
Begin
  Inherited Insert(Index, Item);
End;

Function TCalendarClassList.Remove(Item: TCalendarClass): Integer;
Begin
  Result := Inherited Remove(Item);
End;

{$IFDEF COMPILER2010_UP}

Function TCalendarClassList.RemoveItem(Item: TCalendarClass; Direction: TList.TDirection): Integer;
Begin
  Result := Inherited RemoveItem(Item, Direction);
End;
{$ENDIF}

Function TCalendarClassList.Extract(Item: TCalendarClass): TCalendarClass;
Begin
  Result := TCalendarClass( Inherited Extract(Item));
End;

{$IFDEF COMPILER2010_UP}

Function TCalendarClassList.ExtractItem(Item: TCalendarClass; Direction: TList.TDirection): TCalendarClass;
Begin
  Result := TCalendarClass( Inherited ExtractItem(Item, Direction));
End;
{$ENDIF}

Function TCalendarClassList.IndexOf(Item: TCalendarClass): Integer;
Begin
  Result := Inherited IndexOf(Item);
End;

{$IFDEF COMPILER2010_UP}

Function TCalendarClassList.IndexOfItem(Item: TCalendarClass; Direction: TList.TDirection): Integer;
Begin
  Result := Inherited IndexOfItem(Item, Direction);
End;
{$ENDIF}
{ CalendarTypes }

Class Procedure CalendarTypes.Cleanup;
Begin
  If Assigned(CalendarList) Then
    FreeAndNil(CalendarList);
End;

Class Procedure CalendarTypes.Register(CalendarClass: TCalendarClass);
Var
  Index: Integer;
  RegInfo: TCalendarRegInfo;
Begin
  If Not Assigned(CalendarList) Then
  Begin
    CalendarList := TStringList.Create;
    CalendarList.Sorted := True;
    CalendarList.CaseSensitive := False;
    CalendarList.Duplicates := DupIgnore;
    CalendarList.OwnsObjects := True;
  End;
  Index := CalendarList.Add(CalendarClass.CalendarName);
  RegInfo := TCalendarRegInfo(CalendarList.Objects[Index]);
  If Not Assigned(RegInfo) Then
  Begin
    RegInfo := TCalendarRegInfo.Create(CalendarClass);
    CalendarList.Objects[Index] := RegInfo;
  End;
End;

Class Procedure CalendarTypes.Register(CalendarClass: TCalendarClass; Const Locales: Array Of String);
Var
  Index: Integer;
Begin
  Register(CalendarClass);
  Index := CalendarList.IndexOf(CalendarClass.CalendarName);
  TCalendarRegInfo(CalendarList.Objects[Index]).Associate(Locales);
End;

Class Procedure CalendarTypes.Unregister(CalendarClass: TCalendarClass);
Var
  Index: Integer;
Begin
  If Assigned(CalendarList) Then
  Begin
    Index := CalendarList.IndexOf(CalendarClass.CalendarName);
    If Index >= 0 Then
      CalendarList.Delete(Index);
  End;
End;

Class Procedure CalendarTypes.Unregister(CalendarClass: TCalendarClass; Const Locales: Array Of String);
Var
  Index: Integer;
  RegInfo: TCalendarRegInfo;
Begin
  If Assigned(CalendarList) Then
  Begin
    Index := CalendarList.IndexOf(CalendarClass.CalendarName);
    If Index >= 0 Then
    Begin
      RegInfo := TCalendarRegInfo(CalendarList.Objects[Index]);
      RegInfo.Dissociate(Locales);
    End;
  End;
End;

Class Function CalendarTypes.MinSupportedDateTime: TDateTime;
Var
  I: Integer;
  CalendarClass: TCalendarClass;
Begin
  Result := SysUtils.MinDateTime;
  If Assigned(CalendarList) Then
    For I := CalendarList.Count - 1 Downto 0 Do
    Begin
      CalendarClass := TCalendarRegInfo(CalendarList.Objects[I]).CalendarClass;
      If TCalendar.CompareDate(CalendarClass.MinSupportedDateTime, Result) > 0 Then
        Result := CalendarClass.MinSupportedDateTime;
    End;
End;

Class Function CalendarTypes.MaxSupportedDateTime: TDateTime;
Var
  I: Integer;
  CalendarClass: TCalendarClass;
Begin
  Result := SysUtils.MaxDateTime;
  If Assigned(CalendarList) Then
    For I := CalendarList.Count - 1 Downto 0 Do
    Begin
      CalendarClass := TCalendarRegInfo(CalendarList.Objects[I]).CalendarClass;
      If TCalendar.CompareDate(CalendarClass.MaxSupportedDateTime, Result) < 0 Then
        Result := CalendarClass.MaxSupportedDateTime;
    End;
End;

Class Function CalendarTypes.Count: Integer;
Begin
  If Assigned(CalendarList) Then
    Result := CalendarList.Count
  Else
    Result := 0;
End;

Class Function CalendarTypes.ByIndex(Index: Integer): TCalendarClass;
Begin
  If Assigned(CalendarList) And (Index >= 0) And (Index < CalendarList.Count) Then
    Result := TCalendarRegInfo(CalendarList.Objects[Index]).CalendarClass
  Else
    Result := Nil;
End;

Class Function CalendarTypes.ByID(ID: Cardinal): TCalendarClass;
Var
  I: Integer;
Begin
  If Assigned(CalendarList) Then
    For I := CalendarList.Count - 1 Downto 0 Do
    Begin
      Result := TCalendarRegInfo(CalendarList.Objects[I]).CalendarClass;
      If Result.CalendarID = ID Then
        Exit;
    End;
  Result := Nil;
End;

Class Function CalendarTypes.ByName(Const Name: String): TCalendarClass;
Var
  I: Integer;
Begin
  If Assigned(CalendarList) And CalendarList.Find(Name, I) Then
    Result := TCalendarRegInfo(CalendarList.Objects[I]).CalendarClass
  Else
    Result := Nil;
End;

Class Function CalendarTypes.ByLocale(Const Locale: String): TCalendarClass;
Var
  CalendarID: Cardinal;
  Index, I: Integer;
  RegInfo: TCalendarRegInfo;
Begin
  CalendarID := GetLocaleInt(Locale, LOCALE_ICALENDARTYPE, CAL_GREGORIAN);
  Result := ByID(CalendarID);
  // For Index := 0 To Count - 1 Do
  // Begin
  // RegInfo := TCalendarRegInfo(CalendarList.Objects[Index]);
  // If Assigned(RegInfo.AssociatedLocales) Then
  // Begin
  // For I := 0 To RegInfo.AssociatedLocales.Count - 1 Do
  // If SameLanguage(Locale, RegInfo.AssociatedLocales[I]) Then
  // Begin
  // Result := RegInfo.CalendarClass;
  // If SameText(Locale, RegInfo.AssociatedLocales[I]) Then
  // Exit;
  // End;
  // End;
  // End;
End;

Class Procedure CalendarTypes.GetCalendersOfLocale(Const Locale: String; CalendarClasses: TCalendarClassList);
Var
  Index, I: Integer;
  RegInfo: TCalendarRegInfo;
Begin
  // For Index := 0 To Count - 1 Do
  // Begin
  // RegInfo := TCalendarRegInfo(CalendarList.Objects[Index]);
  // If Not Assigned(RegInfo.AssociatedLocales) Then
  // CalendarClasses.Add(RegInfo.CalendarClass)
  // Else
  // Begin
  // For I := 0 To RegInfo.AssociatedLocales.Count - 1 Do
  // If SameLanguage(Locale, RegInfo.AssociatedLocales[I]) Then
  // Begin
  // CalendarClasses.Add(RegInfo.CalendarClass);
  // Break;
  // End;
  // End;
  // End;
End;

{ CalendarTypes.TCalendarRegInfo }

Constructor CalendarTypes.TCalendarRegInfo.Create(ACalendarClass: TCalendarClass);
Begin
  FCalendarClass := ACalendarClass;
End;

Destructor CalendarTypes.TCalendarRegInfo.Destroy;
Begin
  If Assigned(FLocales) Then
    FLocales.Free;
  Inherited Destroy;
End;

Procedure CalendarTypes.TCalendarRegInfo.Associate(Const Locales: Array Of String);
Var
  I: Integer;
Begin
  If Length(Locales) = 0 Then
    Exit;
  If Not Assigned(FLocales) Then
  Begin
    FLocales := TStringList.Create;
    FLocales.Sorted := True;
    FLocales.CaseSensitive := False;
    FLocales.Duplicates := DupIgnore;
  End;
  For I := Low(Locales) To High(Locales) Do
    FLocales.Add(Locales[I]);
End;

Procedure CalendarTypes.TCalendarRegInfo.Dissociate(Const Locales: Array Of String);
Var
  I, Index: Integer;
Begin
  If Assigned(FLocales) Then
    For I := Low(Locales) To High(Locales) Do
      If FLocales.Find(Locales[I], Index) Then
        FLocales.Delete(Index);
End;

Initialization

Finalization

CalendarTypes.Cleanup;

End.
