Unit NCtrls;

{$I DELPHIAREA.INC}

Interface

Uses
  Windows, Messages, Classes, Contnrs, Graphics, Controls, StdCtrls, ImgList,
{$IFDEF COMPILER_XE3_UP} UITypes, {$ENDIF}
  NHashList;

Type

{$IFNDEF COMPILER_XE2_UP}
  TListBoxItemData = TCustomLongData;
{$ENDIF}
  TPartDrawState = (PdsNormal, PdsHot, PdsPressed, PdsDisabled);
  TTransparentPixel = (TpNone, TpTopLeft, TpTopRight, TpBottomRight, TpBottomLeft);
Function GetCheckBoxSize: TSize;
Function DrawCheckBox(DC: HDC; Const Rect: TRect; State: TCheckBoxState; DrawState: TPartDrawState;
  Flat, Themed: Boolean): TRect;
Function GetDropDownSize: TSize;
Procedure DrawDropDown(DC: HDC; Const Rect: TRect; DrawState: TPartDrawState; Themed: Boolean);

Implementation

Uses
  UxTheme, SysUtils, Types, CommCtrl, Themes, RTLConsts, CheckLst;

// {$R NFlags_16x11.res}

Const
  HorzSpacing = 2;
  VertSpacing = 1;

{$IFNDEF COMPILER_XE2_UP}

Type
  TCustomStyleServices = Class(TThemeServices)
  Public
    Function Enabled: Boolean;
  End;

Function TCustomStyleServices.Enabled: Boolean;
Begin
  Result := ThemesEnabled;
End;

Function StyleServices: TCustomStyleServices;
Begin
  Result := TCustomStyleServices(ThemeServices);
End;
{$ENDIF}
{ Helper Functions }

Var
  CheckBoxSize: TSize;

Function GetCheckBoxSize: TSize;
Begin
  Result := CheckBoxSize;
End;

Function CalcCheckBoxSize: TSize;
Begin
  With TBitmap.Create Do
    Try
      Handle := LoadBitmap(0, PChar(OBM_CHECKBOXES));
      Result.Cx := Width Div 4;
      Result.Cy := Height Div 3;
    Finally
      Free;
    End;
End;

Function DrawCheckBox(DC: HDC; Const Rect: TRect; State: TCheckBoxState; DrawState: TPartDrawState;
  Flat, Themed: Boolean): TRect;
Var
  DrawRect: TRect Absolute Result;
  DrawDetail: TThemedButton;
  DrawFlag: Integer;
  OldRgn, Rgn: HRgn;
  Brush: HBRUSH;
  ElementDetails: TThemedElementDetails;
Begin
  OldRgn := 0;
  DrawRect.Left := (Rect.Left + Rect.Right - CheckBoxSize.Cx) Div 2;
  DrawRect.Top := (Rect.Top + Rect.Bottom - CheckBoxSize.Cy) Div 2;
  DrawRect.Right := DrawRect.Left + CheckBoxSize.Cx;
  DrawRect.Bottom := DrawRect.Top + CheckBoxSize.Cy;
  If Not RectVisible(DC, DrawRect) Then
    Exit;
  If Flat Then
  Begin
    OldRgn := CreateRectRgn(0, 0, 0, 0);
    GetClipRgn(DC, OldRgn);
    With DrawRect Do
      Rgn := CreateRectRgn(Left + 2, Top + 2, Right - 2, Bottom - 2);
    SelectClipRgn(DC, Rgn);
    DeleteObject(Rgn);
  End;
  If Themed And StyleServices.Enabled Then
  Begin
    Case State Of
      CbChecked:
        Case DrawState Of
          PdsNormal:
            DrawDetail := TbCheckBoxCheckedNormal;
          PdsHot:
            DrawDetail := TbCheckBoxCheckedHot;
          PdsPressed:
            DrawDetail := TbCheckBoxCheckedPressed;
        Else // pdsDisabled
          DrawDetail := TbCheckBoxCheckedDisabled;
        End;
      CbUnchecked:
        Case DrawState Of
          PdsNormal:
            DrawDetail := TbCheckBoxUncheckedNormal;
          PdsHot:
            DrawDetail := TbCheckBoxUncheckedHot;
          PdsPressed:
            DrawDetail := TbCheckBoxUncheckedPressed;
        Else // pdsDisabled
          DrawDetail := TbCheckBoxUncheckedDisabled;
        End;
    Else // cbGrayed
      Case DrawState Of
        PdsNormal:
          DrawDetail := TbCheckBoxMixedNormal;
        PdsHot:
          DrawDetail := TbCheckBoxMixedHot;
        PdsPressed:
          DrawDetail := TbCheckBoxMixedPressed;
      Else // pdsDisabled
        DrawDetail := TbCheckBoxMixedDisabled;
      End;
    End;
    ElementDetails := StyleServices.GetElementDetails(DrawDetail);
    StyleServices.DrawElement(DC, ElementDetails, DrawRect);
  End
  Else
  Begin
    Case State Of
      CbChecked:
        DrawFlag := DFCS_BUTTONCHECK Or DFCS_CHECKED;
      CbUnchecked:
        DrawFlag := DFCS_BUTTONCHECK;
    Else // cbGrayed
      DrawFlag := DFCS_BUTTON3STATE Or DFCS_CHECKED;
    End;
    Case DrawState Of
      PdsDisabled:
        DrawFlag := DrawFlag Or DFCS_INACTIVE;
      PdsHot:
        DrawFlag := DrawFlag Or DFCS_HOT;
      PdsPressed:
        DrawFlag := DrawFlag Or DFCS_PUSHED;
    End;
    DrawFrameControl(DC, DrawRect, DFC_BUTTON, DrawFlag);
  End;
  If Flat Then
  Begin
    InflateRect(DrawRect, -1, -1);
    SelectClipRgn(DC, OldRgn);
    DeleteObject(OldRgn);
    Brush := CreateSolidBrush(ColorToRGB(ClBtnShadow));
    FrameRect(DC, DrawRect, Brush);
    DeleteObject(Brush);
  End;
End;

Var
  DropDownSize: TSize;

Function GetDropDownSize: TSize;
Begin
  Result := DropDownSize;
End;

Function CalcDropDownSize: TSize;
Begin
  Result.Cx := GetSystemMetrics(SM_CXVSCROLL);
  Result.Cy := GetSystemMetrics(SM_CYVSCROLL);
End;

Procedure DrawDropDown(DC: HDC; Const Rect: TRect; DrawState: TPartDrawState; Themed: Boolean);
Var
  DrawDetail: TThemedComboBox;
  DrawFlag: Integer;
  ElementDetails: TThemedElementDetails;
Begin
  If Not RectVisible(DC, Rect) Then
    Exit;
  If Themed And StyleServices.Enabled Then
  Begin
    Case DrawState Of
      PdsNormal:
        DrawDetail := TcDropDownButtonNormal;
      PdsHot:
        DrawDetail := TcDropDownButtonHot;
      PdsPressed:
        DrawDetail := TcDropDownButtonPressed;
    Else // pdsDisabled
      DrawDetail := TcDropDownButtonDisabled;
    End;
    ElementDetails := StyleServices.GetElementDetails(DrawDetail);
    StyleServices.DrawElement(DC, ElementDetails, Rect);
  End
  Else
  Begin
    DrawFlag := DFCS_SCROLLCOMBOBOX;
    Case DrawState Of
      PdsDisabled:
        DrawFlag := DrawFlag Or DFCS_INACTIVE;
      PdsHot:
        DrawFlag := DrawFlag Or DFCS_HOT;
      PdsPressed:
        DrawFlag := DrawFlag Or DFCS_PUSHED;
    End;
    DrawFrameControl(DC, Rect, DFC_SCROLL, DrawFlag);
  End;
End;

{$IFDEF COMPILER2010_UP}

Procedure FillGlassRect(Canvas: TCanvas; Rect: TRect);
Var
  MemDC: HDC;
  PaintBuffer: HPAINTBUFFER;
Begin
  PaintBuffer := BeginBufferedPaint(Canvas.Handle, Rect, BPBF_TOPDOWNDIB, Nil, MemDC);
  Try
    FillRect(MemDC, Rect, Canvas.Brush.Handle);
    BufferedPaintMakeOpaque(PaintBuffer, Rect);
  Finally
    EndBufferedPaint(PaintBuffer, True);
  End;
End;
{$ENDIF}

Initialization

CheckBoxSize := CalcCheckBoxSize;
DropDownSize := CalcDropDownSize;

End.
